import java.io.*;

public class Mapper {

    final NESCore nes;
    final int num_8k_ROM_banks;
    final int num_1k_VROM_banks;
    final int mapper_no;
    int patch;

    public Mapper(final NESCore _nes) {
        nes = _nes;
        num_8k_ROM_banks = nes.num_16k_ROM_banks << 1;
        num_1k_VROM_banks = nes.num_8k_VROM_banks << 3;
        mapper_no = nes.mapper_no;
    }

    final void set_CPU_bank(final int page, final int bank) {
        nes.set_cpu_bank(page, bank);
    }

    final void set_CPU_bank4(int bank) {
        set_CPU_bank(4, bank);
    }

    final void set_CPU_bank5(int bank) {
        set_CPU_bank(5, bank);
    }

    final void set_CPU_bank6(int bank) {
        set_CPU_bank(6, bank);
    }

    final void set_CPU_bank7(int bank) {
        set_CPU_bank(7, bank);
    }

    final void set_PPU_bank(int page, int bank) {
        nes.set_ppu_bank(page, bank);
    }

    final void set_PPU_dummy_bank(int page) {
        nes.set_dummy_ppu_bank(page);
    }

    final void set_PPU_bank0(int bank) {
        set_PPU_bank(0, bank);
    }

    final void set_PPU_bank1(int bank) {
        set_PPU_bank(1, bank);
    }

    final void set_PPU_bank2(int bank) {
        set_PPU_bank(2, bank);
    }

    final void set_PPU_bank3(int bank) {
        set_PPU_bank(3, bank);
    }

    final void set_PPU_bank4(int bank) {
        set_PPU_bank(4, bank);
    }

    final void set_PPU_bank5(int bank) {
        set_PPU_bank(5, bank);
    }

    final void set_PPU_bank6(int bank) {
        set_PPU_bank(6, bank);
    }

    final void set_PPU_bank7(int bank) {
        set_PPU_bank(7, bank);
    }

    final void set_VRAM_bank(int page, int bank) {
        nes.set_vram_bank(page, bank);
    }

    final void set_CPU_banks(
        final int b4,
        final int b5,
        final int b6,
        final int b7) {
        set_CPU_bank4(b4);
        set_CPU_bank5(b5);
        set_CPU_bank6(b6);
        set_CPU_bank7(b7);
    }

    final void set_PPU_banks(
        final int b0,
        final int b1,
        final int b2,
        final int b3,
        final int b4,
        final int b5,
        final int b6,
        final int b7) {
        set_PPU_bank0(b0);
        set_PPU_bank1(b1);
        set_PPU_bank2(b2);
        set_PPU_bank3(b3);
        set_PPU_bank4(b4);
        set_PPU_bank5(b5);
        set_PPU_bank6(b6);
        set_PPU_bank7(b7);
    }

    final void set_mirroring(int nt0, int nt1, int nt2, int nt3) {
        nes.set_mirroring(nt0, nt1, nt2, nt3);
    }

    final void set_mirroring(int mirroring) {
        nes.set_mirroring(mirroring);
    }

    public void Reset() {
    }

    public int MemoryReadLow(final int addr) {
        return addr >> 8;
    }

    public void MemoryReadSaveRAM(int addr) {
    }

    public void MemoryWriteLow(int addr, int data) {
    }

    public void MemoryWriteSaveRAM(int addr, int data) {
    }

    public void MemoryWrite(int addr, int data) {
    }

    public void HSync(int scanline) {
    }

    public void VSync() {
    }

    public void Clock(int cyc) {
    }

    public void PPU_Latch_FDFE(int addr) {
    }

    public int PPU_Latch_RenderScreen(int mode, int addr) {
        return 0;
    }

    public byte GetDiskSideNum() {
        return 0;
    }

    public byte GetDiskSide() {
        return 0;
    }

    public void SetDiskSide(byte side) {
    }

    public byte GetDiskData(int pt) {
        return 0;
    }

    public void SetDiskData(int pt, byte data) {
    }

    public byte DiskAccessed() {
        return 0;
    }
    
    public void StateLoad(DataInputStream dis) throws IOException {
    }

    public void StateSave(DataOutputStream dos) throws IOException {
    }

    public void PreStateSave() {
    }

    public static Mapper createMapper(final int mapperNo, final NESCore nes) {
        switch (mapperNo) {
            case 0 :
            case 2 :
            case 3 :
            case 7 :
            case 34 :
            case 40 :
            case 70 :
            case 72 :
            case 77 :
            case 78 :
            case 80 :
            case 86 :
            case 87 :
            case 89 :
            case 92 :
            case 93 :
            case 94 :
            case 96 :
            case 101 :
            case 102 :
            case 122 :
            case 180 :
            case 184 :
            case 185 :
            case 188 :
                return new MapperOther(nes);
            case 1 :
                return new Mapper001(nes);
            case 4 :
                return new Mapper004(nes);
            case 5 :
                return new Mapper005(nes);
            case 9 :
                return new Mapper009(nes);
            case 10 :
                return new Mapper010(nes);
            case 16 :
                return new Mapper016(nes);
            case 17 :
                return new Mapper017(nes);
            case 18 :
                return new Mapper018(nes);
            case 19 :
                return new Mapper019(nes);
            case 20 :
                return new Mapper020(nes);
            case 21 :
                return new Mapper021(nes);
            case 22 :
                return new Mapper022(nes);
            case 23 :
                return new Mapper023(nes);
            case 24 :
                return new Mapper024(nes);
            case 25 :
                return new Mapper025(nes);
            case 26 :
                return new Mapper026(nes);
            case 32 :
                return new Mapper032(nes);
            case 33 :
                return new Mapper033(nes);
            case 64 :
                return new Mapper064(nes);
            case 65 :
                return new Mapper065(nes);
            case 66 :
            case 140 :
                return new Mapper066(nes);
            case 67 :
                return new Mapper067(nes);
            case 68 :
                return new Mapper068(nes);
            case 69 :
                return new Mapper069(nes);
            case 73 :
                return new Mapper073(nes);
            case 75 :
                return new Mapper075(nes);
            case 76 :
                return new Mapper076(nes);
            case 82 :
                return new Mapper082(nes);
            case 85 :
                return new Mapper085(nes);
            case 88 :
                return new Mapper088(nes);
            case 90 :
                return new Mapper090(nes);
            case 95 :
                return new Mapper095(nes);
            case 97 :
                return new Mapper097(nes);
            case 118 :
                return new Mapper118(nes);
            default :
                return null;
        }
    }
}
