import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public final class Mapper004 extends Mapper {

    private static final int[] security_data =
        {
            0xff,
            0xbf,
            0xb7,
            0x97,
            0x97,
            0x17,
            0x57,
            0x4f,
            0x6f,
            0x6b,
            0xeb,
            0xa9,
            0xb1,
            0x90,
            0x94,
            0x14,
            0x56,
            0x4e,
            0x6f,
            0x6b,
            0xeb,
            0xa9,
            0xb1,
            0x90,
            0xd4,
            0x5c,
            0x3e,
            0x26,
            0x87,
            0x83,
            0x13,
            0x00 };

    private int patch;
    private int[] regs;

    private int prg0, prg1;
    private int chr01, chr23, chr4, chr5, chr6, chr7;

    private final int chr_swap() {
        return regs[0] & 0x80;
    }
    private final int prg_swap() {
        return regs[0] & 0x40;
    }

    private int irq_enabled; // IRQs enabled
    private int irq_counter; // IRQ scanline counter, decreasing
    private int irq_latch; // IRQ scanline counter latch
    //private int irq_request;
	private int irq_preset;
    private int irq_preset_vbl;

    private int vs_index; // VS Atari RBI Baseball and VS TKO Boxing

    public Mapper004(NESCore nes) {
        super(nes);

        regs = new int[8];
    }

    /////////////////////////////////////////////////////////////////////
    // Mapper 4
    // much of this is based on the DarcNES source. thanks, nyef :)
    public final void Reset() {
        patch = 0;

        if (nes.crc == 0xdebea5a6 || nes.crc == 0xc5fea9f2) {
            // Ninja Ryukenden 2 - Ankoku no Jashin Ken ||
            // Dai2Ji - Super Robot Taisen
            patch = 1;
        }

        if (nes.crc == 0xd7a97b38) {
            // Chou Jinrou Senki - Warwolf
            patch = 2;
        }

        if (nes.crc == 0xeb2dba63) {
            // VS TKO Boxing
            patch = 3;
            vs_index = 0;
        }

        if (nes.crc == 0x135adf7c) {
            // VS Atari RBI Baseball
            patch = 4;
            vs_index = 0;
        }

        // clear registers FIRST!!!
        for (int i = 0; i < 8; i++) {
            regs[i] = 0x00;
        }

        // set CPU bank pointers
        prg0 = 0;
        prg1 = 1;
        MMC3_set_CPU_banks();

        // set VROM banks
        if (num_1k_VROM_banks != 0) {
            chr01 = 0;
            chr23 = 2;
            chr4 = 4;
            chr5 = 5;
            chr6 = 6;
            chr7 = 7;
            MMC3_set_PPU_banks();
        } else {
            chr01 = chr23 = chr4 = chr5 = chr6 = chr7 = 0;
        }

        irq_enabled    = 0;
        irq_counter    = 0;
        irq_latch      = 0xFF;
        //irq_request    = 0;
        irq_preset     = 0;
        irq_preset_vbl = 0;

    }

    public final void StateLoad(DataInputStream dis) throws IOException {
        patch         = dis.readInt();
        for (int i=0; i < 8; i++)
            regs[i]   = dis.readInt();
        prg0          = dis.readInt();
        prg1          = dis.readInt();
        chr01         = dis.readInt();
        chr23         = dis.readInt();
        chr4          = dis.readInt();
        chr5          = dis.readInt();
        chr6          = dis.readInt();
        chr7          = dis.readInt();
        irq_enabled   = dis.readInt();
        irq_counter   = dis.readInt();
        irq_latch     = dis.readInt();
        vs_index      = dis.readInt();
    }

    public final void StateSave(DataOutputStream dos) throws IOException {
        dos.writeInt(patch);
        for (int i=0; i < 8; i++)
        	dos.writeInt(regs[i]);
        dos.writeInt(prg0);
        dos.writeInt(prg1);
        dos.writeInt(chr01);
        dos.writeInt(chr23);
        dos.writeInt(chr4);
        dos.writeInt(chr5);
        dos.writeInt(chr6);
        dos.writeInt(chr7);
        dos.writeInt(irq_enabled);
        dos.writeInt(irq_counter);
        dos.writeInt(irq_latch);
        dos.writeInt(vs_index);
    }

    public final int MemoryReadLow(int addr) {
        if (patch == 3) {
            // VS TKO Boxing security
            if (addr == 0x5E00) {
                vs_index = 0;
                return 0x00;
            } else if (addr == 0x5E01) {
                return security_data[(vs_index++) & 0x1F];
            }
        } else if (patch == 4) {
            // VS Atari RBI Baseball security
            if (addr == 0x5E00) {
                vs_index = 0;
                return 0x00;
            } else if (addr == 0x5E01) {
                switch (vs_index++) {
                    case 0x09 :
                        return 0x6F;
                    default :
                        return 0xB4;
                }
            }
        }
        return (addr >> 8);
    }

    public final void MemoryWrite(int addr, int data) {
        switch (addr & 0xE001) {
            case 0x8000 :
                regs[0] = data;
                MMC3_set_PPU_banks();
                MMC3_set_CPU_banks();
                break;

            case 0x8001 :
                int bank_num;

                regs[1] = data;
                bank_num = regs[1];

                switch (regs[0] & 0x07) {
                    case 0x00 :
                        // if(num_1k_VROM_banks) {
                        bank_num &= 0xfe;
                        chr01 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x01 :
                        // if(num_1k_VROM_banks) {
                        bank_num &= 0xfe;
                        chr23 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x02 :
                        // if(num_1k_VROM_banks) {
                        chr4 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x03 :
                        // if(num_1k_VROM_banks) {
                        chr5 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x04 :
                        // if(num_1k_VROM_banks) {
                        chr6 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x05 :
                        // if(num_1k_VROM_banks) {
                        chr7 = bank_num;
                        MMC3_set_PPU_banks();
                        // }
                        break;

                    case 0x06 :
                        prg0 = bank_num;
                        MMC3_set_CPU_banks();
                        break;

                    case 0x07 :
                        prg1 = bank_num;
                        MMC3_set_CPU_banks();
                        break;
                }
                break;

            case 0xA000 :
                regs[2] = data;

                // if ((data & 0x40) != 0) {
                // LOG("MAP4 MIRRORING: 0x40 ???" << endl);
                // }

                if (nes.mirroring != NESCore.MIRROR_FOUR_SCREEN) {
                    if ((data & 0x01) != 0) {
                        set_mirroring(NESCore.MIRROR_HORIZ);
                    } else {
                        set_mirroring(NESCore.MIRROR_VERT);
                    }
                }
                break;

            case 0xA001 :
                regs[3] = data;

                // if ((data & 0x80) != 0) {
                // enable save RAM $6000-$7FFF
                // } else {
                // disable save RAM $6000-$7FFF
                // }
                break;

            case 0xC000 :
                regs[4] = data;
                irq_latch = data;
                break;

            case 0xC001 :
                regs[5] = data;
                if(!nes.is_blank_line()) {
					irq_counter |= 0x80;
					irq_preset = 0xFF;
				} else {
					irq_counter |= 0x80;
					irq_preset_vbl = 0xFF;
					irq_preset = 0;
				}
                break;
            case 0xE000 :
                regs[6] = data;
                irq_enabled = 0;
                //irq_request = 0;
                nes.clear_mapper_irq();
                break;

            case 0xE001 :
                regs[7] = data;
                irq_enabled = 1;
                //irq_request = 0;
                break;

            default :
                // LOG("MAP4: UNKNOWN: " << HEX(addr,4) << " = " << HEX(data) << endl);
                break;

        }
    }

    public final void HSync(final int scanline) {
        if ((scanline >= 0) && (scanline <= 239)) {
    		if( irq_preset_vbl != 0) {
    			irq_counter = irq_latch;
    			irq_preset_vbl = 0;
    		}
    		if( irq_preset != 0) {
    			irq_counter = irq_latch;
    			irq_preset = 0;
    		} else if( irq_counter > 0 ) {
    			irq_counter--;
    		}
    		if( irq_counter == 0 ) {
    			if( irq_enabled != 0 ) {
    				nes.set_mapper_irq();
    			}
    			irq_preset = 0xFF;
    		}
    	}
    }

    private final void MMC3_set_CPU_banks() {
        if (prg_swap() != 0) {
            set_CPU_banks(
                num_8k_ROM_banks - 2,
                prg1,
                prg0,
                num_8k_ROM_banks - 1);
        } else {
            set_CPU_banks(
                prg0,
                prg1,
                num_8k_ROM_banks - 2,
                num_8k_ROM_banks - 1);
        }
    }

    private final void MMC3_set_PPU_banks() {
        if (num_1k_VROM_banks != 0) {
            if (chr_swap() != 0) {
                set_PPU_banks(
                    chr4,
                    chr5,
                    chr6,
                    chr7,
                    chr01,
                    chr01 + 1,
                    chr23,
                    chr23 + 1);
            } else {
                set_PPU_banks(
                    chr01,
                    chr01 + 1,
                    chr23,
                    chr23 + 1,
                    chr4,
                    chr5,
                    chr6,
                    chr7);
            }
        } else {
            if (chr_swap() != 0) {
                set_VRAM_bank(0, chr4);
                set_VRAM_bank(1, chr5);
                set_VRAM_bank(2, chr6);
                set_VRAM_bank(3, chr7);
                set_VRAM_bank(4, chr01 + 0);
                set_VRAM_bank(5, chr01 + 1);
                set_VRAM_bank(6, chr23 + 0);
                set_VRAM_bank(7, chr23 + 1);
            } else {
                set_VRAM_bank(0, chr01 + 0);
                set_VRAM_bank(1, chr01 + 1);
                set_VRAM_bank(2, chr23 + 0);
                set_VRAM_bank(3, chr23 + 1);
                set_VRAM_bank(4, chr4);
                set_VRAM_bank(5, chr5);
                set_VRAM_bank(6, chr6);
                set_VRAM_bank(7, chr7);
            }
        }
    }
    /////////////////////////////////////////////////////////////////////

    // state_save, state_load
}
