import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public final class Mapper005 extends Mapper {

    private int[] wb;
    private byte[][] wram;
    private int sram_size;

    private int chr_mode;
    private int[][] chr_page;

    private int irq_enabled;
    private int irq_status;
    private int irq_scanline;
    private int irq_line;
    private int irq_clear;

    private int irq_type;

    private int mult_a;
    private int mult_b;

    private int sram_we_a;
    private int sram_we_b;
    private int prg_size;
    private int chr_size;
    private int gfx_mode;
    private int ntb_mode;
    private int ntb_type[];
    
    private int fill_chr;
    private int fill_pal;
    private int split_control;
    private int split_scroll;
    private int split_page;
    private int cur_ex_bank;
    private int cur_nrm_bank;

    public Mapper005(NESCore nes) {
        super(nes);

        wb = new int[8];
        wram = new byte[8][0x2002];
        chr_page = new int[2][8];
        ntb_type = new int[4];

        nes.exgfx_mapper = false;
    }

    /////////////////////////////////////////////////////////////////////
    // Mapper 5
    public final void Reset() {
        sram_size = 1;
        irq_type = 0;

        if (nes.crc == 0x2b548d75 || // Bandit Kings of Ancient China (J)
            nes.crc == 0xf4cd4998 || // Dai Koukai Jidai (J)
            nes.crc == 0x8fa95456 || // Ishin no Arashi (J)
            nes.crc == 0x98c8e090 || // Nobunaga no Yabou - Sengoku Gunyuu Den (J)
            nes.crc == 0x57e3218b || // L'Empereur (J)
            nes.crc == 0x2f50bd38 || // L'Empereur (U)
            nes.crc == 0x8e9a5e2f || // L'Empereur (Alt)(U)
            nes.crc == 0xb56958d1 || // Nobunaga's Ambition 2 (J)
            nes.crc == 0xe6c28c5f || // Suikoden - Tenmei no Chikai (J)
            nes.crc == 0xcd35e2e9)   // Uncharted Waters (J)
        {
            sram_size = 1;
        }

        if (nes.crc == 0xf4120e58 || // Aoki Ookami to Shiroki Mejika - Genchou Hishi (J)
            nes.crc == 0x286613d8 || // Nobunaga no Yabou - Bushou Fuuun Roku (J)
            nes.crc == 0x11eaad26 || // Romance of the 3 Kingdoms 2 (J)
            nes.crc == 0x95ba5733)   // Sangokushi 2 (J)
        {
            sram_size = 2;
        }

        if( nes.crc == 0xcd9acf43 ) { // Metal Slader Glory(J)
            irq_type = 1;
        }

        // set SaveRAM
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 0x2000; j++) {
                wram[i][j] = nes.SRAM[0x2000*i+j];
            }
        }
        for (int i = 0; i < 8; i++) wb[i] = 8; 
        MMC5_setBank_SRAM(3, 0);

        // Init ExSound
        //parent_NES->apu->SelectExSound(8);

        // set CPU bank pointers
        set_CPU_bank(4, num_8k_ROM_banks - 1);
        set_CPU_bank(5, num_8k_ROM_banks - 1);
        set_CPU_bank(6, num_8k_ROM_banks - 1);
        set_CPU_bank(7, num_8k_ROM_banks - 1);

        // set PPU bank pointers
        set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
        cur_ex_bank = cur_nrm_bank = 0xFF;

        chr_mode = 0;
        for (int i = 0; i < 8; i++) {
            chr_page[0][i] = i;
            chr_page[1][i] = (i & 0x03) + 4;
        }

        prg_size = 3;
        sram_we_a = 0x02;
        sram_we_b = 0x01;
        chr_size = 3;
        gfx_mode = 0;
        ntb_mode = 0;
        for (int i = 0; i < 4; i++) ntb_type[i] = 0; 
        
        fill_chr = 0;
        fill_pal = 0;
        
        irq_enabled = 0;
        irq_status = 0;
        irq_scanline = 0;
        irq_line = 0;
        irq_clear = 0;

        split_control = 0;
        split_scroll = 0;
        split_page = 0;
    }

    public void StateLoad(DataInputStream dis) throws IOException {
        for (int i = 0; i < 8; i++) {
            wb[i] = dis.readInt();
        }
        for (int i = 0; i < 8; i++) {
            chr_page[0][i] = dis.readInt();
            chr_page[1][i] = dis.readInt();
        }
        irq_enabled = dis.readInt();
        irq_status = dis.readInt();
        irq_scanline = dis.readInt();
        irq_line = dis.readInt();
        mult_a = dis.readInt();
        mult_b = dis.readInt();
        sram_we_a = dis.readInt();
        sram_we_b = dis.readInt();
        prg_size = dis.readInt();
        chr_size = dis.readInt();
        gfx_mode = dis.readInt();        

        if(gfx_mode == 1)
            nes.exgfx_mapper = true;
        else
            nes.exgfx_mapper = false;
        cur_ex_bank = cur_nrm_bank = 0xFF;

        // set SaveRAM
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 0x2000; j++) {
                wram[i][j] = nes.SRAM[0x2000*i+j];
            }
        }
        for (int i = 3; i < 8; i++) {
            if(wb[i] != 8) {
                nes.cpu_map[i] = wram[wb[i]];
                if(i > 4) {
                    try {
                        nes.cpu_map[i - 1][0x2000] = nes.cpu_map[i][0x0000];
                        nes.cpu_map[i - 1][0x2001] = nes.cpu_map[i][0x0001];
                    } catch (Exception e) {}
                }
                if (i < 7) {
                    try {
                        nes.cpu_map[i][0x2000] = nes.cpu_map[i + 1][0x0000];
                        nes.cpu_map[i][0x2001] = nes.cpu_map[i + 1][0x0001];
                    } catch (Exception e) {}
                }
            }
        }
    }

    public void PreStateSave() {
        // set SeveRAM
        for (int i = 0; i < 8; i++) {
            for (int j = 0; j < 0x2000; j++) {
                nes.SRAM[0x2000*i+j] = wram[i][j];
            }
        }
    }

    public void StateSave(DataOutputStream dos) throws IOException {
        for (int i = 0; i < 8; i++) {
            dos.writeInt(wb[i]);
        }
        for (int i = 0; i < 8; i++) {
            dos.writeInt(chr_page[0][i]);
            dos.writeInt(chr_page[1][i]);
        }
        dos.writeInt(irq_enabled);
        dos.writeInt(irq_status);
        dos.writeInt(irq_scanline);
        dos.writeInt(irq_line);
        dos.writeInt(mult_a);
        dos.writeInt(mult_b);
        dos.writeInt(sram_we_a);
        dos.writeInt(sram_we_b);
        dos.writeInt(prg_size);
        dos.writeInt(chr_size);
        dos.writeInt(gfx_mode);
    }

    public final int MemoryReadLow(int addr) {
        int ret = (int) (addr >> 8);
        switch(addr) {
        case 0x5015:
            //ret = nes.ExRead( addr );
            break;
        case 0x5204:
            ret = irq_status;
            irq_status &= ~0x80;
            nes.clear_mapper_irq();
            break;
        case 0x5205:
            ret = (int) ((mult_a * mult_b) & 0x00FF);
            break;
        case 0x5206:
            ret = (int) (((mult_a * mult_b) & 0xFF00) >> 8);
            break;
        default:
            break;
        }
        if (addr >= 0x5C00 && addr <= 0x5FFF) {
            if (gfx_mode == 2 || gfx_mode == 3) {
                ret = nes.VRAM[10][addr & 0x3FF] & 0xFF;
            }
        }
        return ret;
    }

    public final void MemoryWriteLow(int addr, int data) {
        int i;
        byte[] nametable3;

        switch (addr) {
            case 0x5100 :
                prg_size = data & 0x03;
                break;

            case 0x5101 :
                chr_size = data & 0x03;
                cur_nrm_bank = 0xFF;
                break;

            case 0x5102 :
                sram_we_a = data & 0x03;
                break;

            case 0x5103 :
                sram_we_b = data & 0x03;
                break;

            case 0x5104 :
                gfx_mode = data & 0x03;
                if(gfx_mode == 1)
                    nes.exgfx_mapper = true;
                else
                    nes.exgfx_mapper = false;
                break;

            case 0x5105 :
                ntb_mode = data & 0x03;
                for (i = 0; i < 4; i++) {
                    ntb_type[i] = data & 0x03;
                    set_VRAM_bank(8 + i, ntb_type[i]);
                    data >>= 2;
                }
                break;

            case 0x5106 :
                fill_chr = data;
                /**/
                nametable3 = nes.VRAM[11];
                for (i = 0; i < 0x3C0; i++) {
                    nametable3[i] = (byte) data;
                }
                /**/
                break;

            case 0x5107 :
                fill_pal = data & 0x03;
                /**/
                nametable3 = nes.VRAM[11];
                data &= 0x03;
                data = data | (data << 2) | (data << 4) | (data << 6);
                for (i = 0x3C0; i < 0x400; i++) {
                    nametable3[i] = (byte) data;
                }
                /**/
                break;

            case 0x5113 :
                MMC5_setBank_SRAM(3, data & 0x07);
                break;

            case 0x5114 :
            case 0x5115 :
            case 0x5116 :
            case 0x5117 :
                MMC5_setBank_CPU(addr, data);
                break;

            case 0x5120 :
            case 0x5121 :
            case 0x5122 :
            case 0x5123 :
            case 0x5124 :
            case 0x5125 :
            case 0x5126 :
            case 0x5127 :
                chr_mode = 0;
                chr_page[0][addr & 0x07] = data;
                cur_ex_bank = 0xFF;
                cur_nrm_bank = 0;
                MMC5_setBank_PPU(0);
                break;

            case 0x5128 :
            case 0x5129 :
            case 0x512A :
            case 0x512B :
                chr_mode = 1;
                chr_page[1][(addr & 0x03) + 0] = data;
                chr_page[1][(addr & 0x03) + 4] = data;
                cur_nrm_bank = 0xFF;
                MMC5_setBank_PPU(1);
                break;

            case 0x5200 :
                split_control = data;
                break;

            case 0x5201 :
                split_scroll = data;
                break;

            case 0x5202 :
                split_page = data & 0x3F;
                break;

            case 0x5203 :
                irq_line = data;
                nes.clear_mapper_irq();
                break;

            case 0x5204 :
                irq_enabled = data;
                nes.clear_mapper_irq();
                break;

            case 0x5205 :
                mult_a = data;
                break;

            case 0x5206 :
                mult_b = data;
                break;

            default :
                /*
                if (addr >= 0x5000 && addr <= 0x5015) {
                    parent_NES->apu->ExWrite(addr, data);
                } else
                */
                if( addr >= 0x5C00 && addr <= 0x5FFF ) {
                    if( gfx_mode == 2 ) {
                        nes.VRAM[10][addr&0x3FF] = (byte)data;
                    } else if( gfx_mode != 3 ) {
                        if( (irq_status & 0x40) != 0) {
                            nes.VRAM[10][addr&0x3FF] = (byte)data;
                        } else {
                            nes.VRAM[10][addr&0x3FF] = 0;
                        }
                    }
                }
                break;
        }
    }

    public final void MemoryWriteSaveRAM(int addr, int data) {
        if (sram_we_a == 0x02 && sram_we_b == 0x01) {
            if (wb[3] != 8) {
                wram[wb[3]][(addr & 0x1FFF)] = (byte) data;
            }
        }
    }

    public final void MemoryWrite(int addr, int data) {
        if (sram_we_a == 0x02 && sram_we_b == 0x01) {
            if (addr >= 0x8000 && addr <= 0x9FFF) {
                if (wb[4] != 8) {
                    wram[wb[4]][(addr & 0x1FFF)] = (byte) data;
                }
            } else if (addr >= 0xA000 && addr <= 0xBFFF) {
                if (wb[5] != 8) {
                    wram[wb[5]][(addr & 0x1FFF)] = (byte) data;
                }
            } else if (addr >= 0xC000 && addr <= 0xDFFF) {
                if (wb[6] != 8) {
                    wram[wb[6]][(addr & 0x1FFF)] = (byte) data;
                }
            }
        }
    }

    private final void MMC5_setBank_CPU(int addr, int bank) {
        if ((bank & 0x80) != 0) {
            switch(addr & 0x07){
            case 4:
                if (prg_size == 3) {
                    set_CPU_bank(4, bank & 0x7F);
                    wb[4] = 8;
                }
                break;
            case 5:
                if (prg_size == 1 || prg_size == 2) {
                    set_CPU_bank(4, (bank & 0x7E) + 0);
                    set_CPU_bank(5, (bank & 0x7E) + 1);
                    wb[4] = wb[5] = 8;
                } else if(prg_size == 3) {
                    set_CPU_bank(5, bank & 0x7F);
                    wb[5] = 8;
                }
                break;
            case 6:
                if (prg_size == 2 || prg_size == 3) {
                    set_CPU_bank(6, bank & 0x7F);
                    wb[6] = 8;
                }
                break;
            case 7:
                if (prg_size == 0) {
                    set_CPU_bank(4, (bank & 0x7C) + 0);
                    set_CPU_bank(5, (bank & 0x7C) + 1);
                    set_CPU_bank(6, (bank & 0x7C) + 2);
                    set_CPU_bank(7, (bank & 0x7C) + 3);
                    wb[4] = wb[5] = wb[6] = 8;
                } else if (prg_size == 1) {
                    set_CPU_bank(6, (bank & 0x7E) + 0);
                    set_CPU_bank(7, (bank & 0x7E) + 1);
                    wb[6] = 8;
                } else if (prg_size == 2 || prg_size == 3) {
                    set_CPU_bank(7, bank & 0x7F);
                }
                break;
            }
        } else {
            switch(addr & 0x07){
            case 4:
                if (prg_size == 3) {
                    MMC5_setBank_SRAM(4, bank & 0x07);
                }
                break;
            case 5:
                if (prg_size == 1 || prg_size == 2) {
                    MMC5_setBank_SRAM(4, (bank & 0x06) + 0);
                    MMC5_setBank_SRAM(5, (bank & 0x06) + 1);
                } else if(prg_size == 3) {
                    MMC5_setBank_SRAM(5, bank & 0x07);
                }
                break;
            case 6:
                if (prg_size == 2 || prg_size == 3) {
                    MMC5_setBank_SRAM(6, bank & 0x07);
                }
                break;
            }
        }
    }

    private final void MMC5_setBank_SRAM(int page, int bank) {
        if (bank != 8) {
            switch (sram_size) {
                case 0 :
                    bank = (bank > 3) ? 8 : 0;
                    break;
                case 1 :
                    bank = (bank > 3) ? 1 : 0;
                    break;
                case 2 :
                    bank = (bank > 3) ? 8 : bank;
                    break;
                case 3 :
                    bank = (bank > 3) ? 4 : bank;
                    break;
                default :
                    break;
            }
        }
        wb[page] = bank;

        if (bank != 8) {
            nes.cpu_map[page] = wram[bank];
            if(page > 4) {
                try {
                    nes.cpu_map[page - 1][0x2000] = nes.cpu_map[page][0x0000];
                    nes.cpu_map[page - 1][0x2001] = nes.cpu_map[page][0x0001];
                } catch (Exception e) {}
            }
            if (page < 7) {
                try {
                    nes.cpu_map[page][0x2000] = nes.cpu_map[page + 1][0x0000];
                    nes.cpu_map[page][0x2001] = nes.cpu_map[page + 1][0x0001];
                } catch (Exception e) {}
            }
        }
    }

    private final void MMC5_setBank_PPU(int mode) {
        int bank;
        if(chr_mode == 0) {
            switch (chr_size) {
            case 0:
                bank = chr_page[mode][7] << 3;
                nes.set_ppu_bank(0, bank + 0);
                nes.set_ppu_bank(1, bank + 1);
                nes.set_ppu_bank(2, bank + 2);
                nes.set_ppu_bank(3, bank + 3);
                nes.set_ppu_bank(4, bank + 4);
                nes.set_ppu_bank(5, bank + 5);
                nes.set_ppu_bank(6, bank + 6);
                nes.set_ppu_bank(7, bank + 7);
                break;
            case 1:
                bank = chr_page[mode][3] << 2;
                nes.set_ppu_bank(0, bank + 0);
                nes.set_ppu_bank(1, bank + 1);
                nes.set_ppu_bank(2, bank + 2);
                nes.set_ppu_bank(3, bank + 3);
                bank = chr_page[mode][7] << 2;
                nes.set_ppu_bank(4, bank + 0);
                nes.set_ppu_bank(5, bank + 1);
                nes.set_ppu_bank(6, bank + 2);
                nes.set_ppu_bank(7, bank + 3);
                break;
            case 2:
                bank = chr_page[mode][1] << 1;
                nes.set_ppu_bank(0, bank + 0);
                nes.set_ppu_bank(1, bank + 1);
                bank = chr_page[mode][3] << 1;
                nes.set_ppu_bank(2, bank + 0);
                nes.set_ppu_bank(3, bank + 1);
                bank = chr_page[mode][5] << 1;
                nes.set_ppu_bank(4, bank + 0);
                nes.set_ppu_bank(5, bank + 1);
                bank = chr_page[mode][7] << 1;
                nes.set_ppu_bank(6, bank + 0);
                nes.set_ppu_bank(7, bank + 1);
                break;
            case 3:
                nes.set_ppu_bank(0, chr_page[mode][0]);
                nes.set_ppu_bank(1, chr_page[mode][1]);
                nes.set_ppu_bank(2, chr_page[mode][2]);
                nes.set_ppu_bank(3, chr_page[mode][3]);
                nes.set_ppu_bank(4, chr_page[mode][4]);
                nes.set_ppu_bank(5, chr_page[mode][5]);
                nes.set_ppu_bank(6, chr_page[mode][6]);
                nes.set_ppu_bank(7, chr_page[mode][7]);
                break;
            }
        } else if( chr_mode == 1 ) {
                // PPU BG Bank
                /*
                switch( chr_size ) {
                case 0:
                    for( int i = 0; i < 8; i++ ) {
                        BG_MEM_BANK[i] = VROM+0x2000*(chr_page[1][7]%VROM_8K_SIZE)+0x0400*i;
                        BG_MEM_PAGE[i] = (chr_page[1][7]%VROM_8K_SIZE)*8+i;
                    }
                    break;
                case 1:
                    for( int i = 0; i < 4; i++ ) {
                        BG_MEM_BANK[i+0] = VROM+0x1000*(chr_page[1][3]%VROM_4K_SIZE)+0x0400*i;
                        BG_MEM_BANK[i+4] = VROM+0x1000*(chr_page[1][7]%VROM_4K_SIZE)+0x0400*i;
                        BG_MEM_PAGE[i+0] = (chr_page[1][3]%VROM_4K_SIZE)*4+i;
                        BG_MEM_PAGE[i+4] = (chr_page[1][7]%VROM_4K_SIZE)*4+i;
                }
                break;
            case 2:
                for( int i = 0; i < 2; i++ ) {
                    BG_MEM_BANK[i+0] = VROM+0x0800*(chr_page[1][1]%VROM_2K_SIZE)+0x0400*i;
                    BG_MEM_BANK[i+2] = VROM+0x0800*(chr_page[1][3]%VROM_2K_SIZE)+0x0400*i;
                    BG_MEM_BANK[i+4] = VROM+0x0800*(chr_page[1][5]%VROM_2K_SIZE)+0x0400*i;
                    BG_MEM_BANK[i+6] = VROM+0x0800*(chr_page[1][7]%VROM_2K_SIZE)+0x0400*i;
                    BG_MEM_PAGE[i+0] = (chr_page[1][1]%VROM_2K_SIZE)*2+i;
                    BG_MEM_PAGE[i+2] = (chr_page[1][3]%VROM_2K_SIZE)*2+i;
                    BG_MEM_PAGE[i+4] = (chr_page[1][5]%VROM_2K_SIZE)*2+i;
                    BG_MEM_PAGE[i+6] = (chr_page[1][7]%VROM_2K_SIZE)*2+i;
                }
                break;
            case 3:
                for( int i = 0; i < 8; i++ ) {
                    BG_MEM_BANK[i] = VROM+0x0400*(chr_page[1][i]%VROM_1K_SIZE);
                    BG_MEM_PAGE[i] = (chr_page[1][i]%VROM_1K_SIZE)+i;
                }
                break;
            }
            */
        }
    }

    public final void HSync(int scanline) {
        if( irq_type == 1 ) {
            if( irq_scanline == irq_line ) {
                irq_status |= 0x80;
            }
        }
        if (scanline < 240) {
            irq_scanline++;
            irq_status |= 0x40;
            irq_clear = 0;
        } else if( irq_type == 1 ) {
            irq_scanline = 0;
            irq_status &= ~0x80;
            irq_status &= ~0x40;
        }
        if( irq_type != 1 ) {
            if (irq_scanline == irq_line) {
                irq_status |= 0x80;
            }
        }
        if( ++irq_clear > 2 ) {
            irq_scanline = 0;
            irq_status &= ~0x80;
            irq_status &= ~0x40;
            nes.clear_mapper_irq();
        }
        if((irq_enabled & 0x80) != 0 && (irq_status & 0x80) != 0 && (irq_status & 0x40) != 0 ) {
            nes.set_mapper_irq();
        }
    }

    public final int PPU_Latch_RenderScreen(int mode, int addr) {
        if (mode == 1) {
            // ex gfx mode
            int bank = (nes.VRAM[10][addr] & 0x3F) << 2;
            if (bank != cur_ex_bank) {
                nes.set_ppu_bank(0, bank    );
                nes.set_ppu_bank(1, bank + 1);
                nes.set_ppu_bank(2, bank + 2);
                nes.set_ppu_bank(3, bank + 3);
                nes.set_ppu_bank(4, bank    );
                nes.set_ppu_bank(5, bank + 1);
                nes.set_ppu_bank(6, bank + 2);
                nes.set_ppu_bank(7, bank + 3);
                cur_ex_bank = bank;
                cur_nrm_bank = 0xFF;
            }
            return  (((nes.VRAM[10][addr] & 0xC0) >> 4) | 0x01);
        }
        // normal
        if(mode != cur_nrm_bank) {
            MMC5_setBank_PPU(mode);
            cur_ex_bank = 0xFF;
            cur_nrm_bank = mode;
        }
        return 0;
    }
}
