public final class Mapper018 extends Mapper {

    private int patch;
    private int[] regs;
    private int irq_enabled;
    private int irq_latch;
    private int irq_counter;

    public Mapper018(final NESCore nes) {
        super(nes);
        regs = new int[11];
    }

    public final void Reset() {
        if (nes.crc == 0x424ec0a6) {
            patch = 1;
        }
        set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
        regs[0] = 0;
        regs[1] = 1;
        regs[2] = num_8k_ROM_banks - 2;
        regs[3] = num_8k_ROM_banks - 1;
        regs[4] = 0;
        regs[5] = 0;
        regs[6] = 0;
        regs[7] = 0;
        regs[8] = 0;
        regs[9] = 0;
        regs[10] = 0;
        irq_enabled = 0;
        irq_latch = 0;
        irq_counter = 0;
    }

    public final void MemoryWrite(int addr, int data) {
        switch (addr) {
            case 0x8000 :
                regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
                set_CPU_bank4(regs[0]);
                break;
            case 0x8001 :
                regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
                set_CPU_bank4(regs[0]);
                break;
            case 0x8002 :
                regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
                set_CPU_bank5(regs[1]);
                break;
            case 0x8003 :
                regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
                set_CPU_bank5(regs[1]);
                break;
            case 0x9000 :
                regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
                set_CPU_bank6(regs[2]);
                break;
            case 0x9001 :
                regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
                set_CPU_bank6(regs[2]);
                break;
            case 0xA000 :
                regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
                set_PPU_bank0(regs[3]);
                break;
            case 0xA001 :
                regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank0(regs[3]);
                break;
            case 0xA002 :
                regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
                set_PPU_bank1(regs[4]);
                break;
            case 0xA003 :
                regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank1(regs[4]);
                break;
            case 0xB000 :
                regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
                set_PPU_bank2(regs[5]);
                break;
            case 0xB001 :
                regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank2(regs[5]);
                break;
            case 0xB002 :
                regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
                set_PPU_bank3(regs[6]);
                break;
            case 0xB003 :
                regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank3(regs[6]);
                break;
            case 0xC000 :
                regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
                set_PPU_bank4(regs[7]);
                break;
            case 0xC001 :
                regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank4(regs[7]);
                break;
            case 0xC002 :
                regs[8] = (regs[8] & 0xF0) | (data & 0x0F);
                set_PPU_bank5(regs[8]);
                break;
            case 0xC003 :
                regs[8] = (regs[8] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank5(regs[8]);
                break;
            case 0xD000 :
                regs[9] = (regs[9] & 0xF0) | (data & 0x0F);
                set_PPU_bank6(regs[9]);
                break;
            case 0xD001 :
                regs[9] = (regs[9] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank6(regs[9]);
                break;
            case 0xD002 :
                regs[10] = (regs[10] & 0xF0) | (data & 0x0F);
                set_PPU_bank7(regs[10]);
                break;
            case 0xD003 :
                regs[10] = (regs[10] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank7(regs[10]);
                break;
            case 0xE000 :
                irq_latch = (irq_latch & 0xFFF0) | (data & 0x0F);
                break;
            case 0xE001 :
                irq_latch = (irq_latch & 0xFF0F) | ((data & 0x0F) << 4);
                break;
            case 0xE002 :
                irq_latch = (irq_latch & 0xF0FF) | ((data & 0x0F) << 8);
                break;
            case 0xE003 :
                irq_latch = (irq_latch & 0x0FFF) | ((data & 0x0F) << 12);
                break;
            case 0xF000 :
                irq_counter = irq_latch;
                break;
            case 0xF001 :
                irq_enabled = data & 0x01;
                break;
            case 0xF002 :
                data &= 0x03;
                if (data == 0) {
                    set_mirroring(NESCore.MIRROR_HORIZ);
                } else if (data == 1) {
                    set_mirroring(NESCore.MIRROR_VERT);
                } else {
                    set_mirroring(0, 0, 0, 0);
                }
                break;
        }
    }

    public final void HSync(int scanline) {
        if (irq_enabled != 0) {
            if (irq_counter <= 113) {
                nes.do_irq();
                irq_counter = (patch == 1) ? 114 : 0;
                irq_enabled = 0;
            } else {
                irq_counter -= 113;
            }
        }
    }
}
