import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

public final class Mapper025 extends Mapper {

    private int[] regs;
    private int irq_enabled;
    private int irq_counter;
    private int irq_latch;
    private int irq_occur;
    private int irq_clock;

    public Mapper025(NESCore nes) {
        super(nes);

        regs = new int[11];
        nes.exclk_mapper = true;
    }

    /////////////////////////////////////////////////////////////////////
    // Mapper 25
    public final void Reset() {

        // set CPU bank pointers
        set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);

        // set PPU bank pointers
        if (num_1k_VROM_banks != 0) {
            set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
        }

        for (int i = 0; i < 11; i++)
            regs[i] = 0;
        regs[9] = num_8k_ROM_banks - 2;

        irq_enabled = 0;
        irq_counter = 0;
        irq_latch = 0;
        irq_occur = 0;
        irq_clock = 0;
    }

    public final void StateLoad(DataInputStream dis) throws IOException {
        for (int i=0; i < 11; i++)
            regs[i]   = dis.readInt();
        irq_enabled   = dis.readInt();
        irq_counter   = dis.readInt();
        irq_latch     = dis.readInt();
        irq_occur     = dis.readInt();
        irq_clock     = dis.readInt();

    }

    public final void StateSave(DataOutputStream dos) throws IOException {
        for (int i=0; i < 11; i++)
            dos.writeInt(regs[i]);
        dos.writeInt(irq_enabled);
        dos.writeInt(irq_counter);
        dos.writeInt(irq_latch);
        dos.writeInt(irq_occur);
        dos.writeInt(irq_clock);
    }
    
    public final void MemoryWrite(int addr, int data) {
        // regs[0] ... 1K VROM bank at PPU $0000
        // regs[1] ... 1K VROM bank at PPU $0400
        // regs[2] ... 1K VROM bank at PPU $0800
        // regs[3] ... 1K VROM bank at PPU $0C00
        // regs[4] ... 1K VROM bank at PPU $1000
        // regs[5] ... 1K VROM bank at PPU $1400
        // regs[6] ... 1K VROM bank at PPU $1800
        // regs[7] ... 1K VROM bank at PPU $1C00
        // regs[8] ... 8K ROM bank at CPU $8000
        // regs[9] ... 8K ROM bank at CPU $C000
        // regs[10] .. ROM Swap flag

        switch (addr & 0xF000) {
            case 0x8000 :
                if ((regs[10] & 0x02) != 0) {
                    regs[9] = data;
                    set_CPU_bank6(data);
                } else {
                    regs[8] = data;
                    set_CPU_bank4(data);
                }
                break;

            case 0xA000 :
                set_CPU_bank5(data);
                break;
        }

        switch (addr & 0xF00F) {
            case 0x9000 :
                data &= 0x03;
                if (data == 0) {
                    set_mirroring(NESCore.MIRROR_VERT);
                } else if (data == 1) {
                    set_mirroring(NESCore.MIRROR_HORIZ);
                } else if (data == 2) {
                    set_mirroring(0, 0, 0, 0);
                } else {
                    set_mirroring(1, 1, 1, 1);
                }
                break;

            case 0x9001 :
            case 0x9004 :
                if ((regs[10] & 0x02) != (data & 0x02)) {
                    int swap = regs[8];
                    regs[8] = regs[9];
                    regs[9] = swap;
                    set_CPU_bank4(regs[8]);
                    set_CPU_bank6(regs[9]);
                }
                regs[10] = data;
                break;

            case 0xB000 :
                regs[0] = (regs[0] & 0xF0) | (data & 0x0F);
                set_PPU_bank0(regs[0]);
                break;

            case 0xB001 :
            case 0xB004 :
                regs[1] = (regs[1] & 0xF0) | (data & 0x0F);
                set_PPU_bank1(regs[1]);
                break;

            case 0xB002 :
            case 0xB008 :
                regs[0] = (regs[0] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank0(regs[0]);
                break;

            case 0xB003 :
            case 0xB00C :
                regs[1] = (regs[1] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank1(regs[1]);
                break;

            case 0xC000 :
                regs[2] = (regs[2] & 0xF0) | (data & 0x0F);
                set_PPU_bank2(regs[2]);
                break;

            case 0xC001 :
            case 0xC004 :
                regs[3] = (regs[3] & 0xF0) | (data & 0x0F);
                set_PPU_bank3(regs[3]);
                break;

            case 0xC002 :
            case 0xC008 :
                regs[2] = (regs[2] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank2(regs[2]);
                break;

            case 0xC003 :
            case 0xC00C :
                regs[3] = (regs[3] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank3(regs[3]);
                break;

            case 0xD000 :
                regs[4] = (regs[4] & 0xF0) | (data & 0x0F);
                set_PPU_bank4(regs[4]);
                break;

            case 0xD001 :
            case 0xD004 :
                regs[5] = (regs[5] & 0xF0) | (data & 0x0F);
                set_PPU_bank5(regs[5]);
                break;

            case 0xD002 :
            case 0xD008 :
                regs[4] = (regs[4] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank4(regs[4]);
                break;

            case 0xD003 :
            case 0xD00C :
                regs[5] = (regs[5] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank5(regs[5]);
                break;

            case 0xE000 :
                regs[6] = (regs[6] & 0xF0) | (data & 0x0F);
                set_PPU_bank6(regs[6]);
                break;

            case 0xE001 :
            case 0xE004 :
                regs[7] = (regs[7] & 0xF0) | (data & 0x0F);
                set_PPU_bank7(regs[7]);
                break;

            case 0xE002 :
            case 0xE008 :
                regs[6] = (regs[6] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank6(regs[6]);
                break;

            case 0xE003 :
            case 0xE00C :
                regs[7] = (regs[7] & 0x0F) | ((data & 0x0F) << 4);
                set_PPU_bank7(regs[7]);
                break;

            case 0xF000 :
                irq_latch = (irq_latch & 0xF0) | (data & 0x0F);
                nes.clear_mapper_irq();
                break;

            case 0xF001 :
            case 0xF004 :
                irq_enabled = data & 0x03;
                irq_counter = irq_latch;
                irq_clock = 0;
                nes.clear_mapper_irq();
                break;

            case 0xF002 :
            case 0xF008 :
                irq_latch = (irq_latch & 0x0F) | ((data & 0x0F) << 4);
                nes.clear_mapper_irq();
                break;

            case 0xF003 :
            case 0xF00C :
                irq_enabled = (irq_enabled & 0x01) * 3;
                nes.clear_mapper_irq();
                break;
        }
    }

    public final void Clock(int cyc){
        if((irq_enabled & 0x02) != 0) {
            irq_clock += cyc;
            while( irq_clock >= 1364 ) {
                irq_clock -= 1364;
                if(++irq_counter >= 256) {
                    irq_counter = irq_latch;
                    nes.set_mapper_irq();
                }
            }
        }
    }
    /*
    #define MAP25_ROM(ptr)  (((ptr)-parent_NES->ROM->get_ROM_banks())  >> 13)
    #define MAP25_VROM(ptr) (((ptr)-parent_NES->ROM->get_VROM_banks()) >> 10)
    
    void NES_mapper25::SNSS_fixup()
    {
      NES_6502::Context context;
      parent_NES->cpu->GetContext(&context);
      regs[0] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[0]);
      regs[1] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[1]);
      regs[2] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[2]);
      regs[3] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[3]);
      regs[4] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[4]);
      regs[5] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[5]);
      regs[6] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[6]);
      regs[7] = MAP25_VROM(parent_NES->ppu->PPU_VRAM_banks[7]);
      regs[8] = MAP25_ROM(context.mem_page[4]);
      regs[9] = MAP25_ROM(context.mem_page[6]);
    }
    */
    /////////////////////////////////////////////////////////////////////

}
