public final class Mapper065 extends Mapper {

    private int patch;
    private int irq_enabled;
    private int irq_counter;
    private int irq_latch;

    public Mapper065(final NESCore nes) {
        super(nes);
    }

    public final void Reset() {
        patch = 0;
        if (nes.crc == 0xe30b7f64) {
            patch = 1;
        }
        set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
        if (num_1k_VROM_banks != 0) {
            set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
        }
        irq_enabled = 0;
        irq_counter = 0;
    }

    public final void MemoryWrite(int addr, int data) {
        switch (addr) {
            case 0x8000 :
                set_CPU_bank4(data);
                break;
            case 0x9000 :
                if (patch == 0) {
                    if ((data & 0x40) != 0) {
                        set_mirroring(NESCore.MIRROR_VERT);
                    } else {
                        set_mirroring(NESCore.MIRROR_HORIZ);
                    }
                }
                break;
            case 0x9001 :
                if (patch != 0) {
                    if ((data & 0x80) != 0) {
                        set_mirroring(NESCore.MIRROR_HORIZ);
                    } else {
                        set_mirroring(NESCore.MIRROR_VERT);
                    }
                }
                break;
            case 0x9003 :
                if (patch == 0) {
                    irq_enabled = data & 0x80;
                }
                break;
            case 0x9004 :
                if (patch == 0) {
                    irq_counter = irq_latch;
                }
                break;
            case 0x9005 :
                if (patch != 0) {
                    irq_counter = (data << 1);
                    irq_enabled = data;
                } else {
                    irq_latch = (irq_latch & 0x00FF) | (data << 8);
                }
                break;
            case 0x9006 :
                if (patch != 0) {
                    irq_enabled = 1;
                } else {
                    irq_latch = (irq_latch & 0xFF00) | data;
                }
                break;
            case 0xB000 :
                set_PPU_bank0(data);
                break;
            case 0xB001 :
                set_PPU_bank1(data);
                break;
            case 0xB002 :
                set_PPU_bank2(data);
                break;
            case 0xB003 :
                set_PPU_bank3(data);
                break;
            case 0xB004 :
                set_PPU_bank4(data);
                break;
            case 0xB005 :
                set_PPU_bank5(data);
                break;
            case 0xB006 :
                set_PPU_bank6(data);
                break;
            case 0xB007 :
                set_PPU_bank7(data);
                break;
            case 0xA000 :
                set_CPU_bank5(data);
                break;
            case 0xC000 :
                set_CPU_bank6(data);
                break;
        }
    }

    public final void HSync(int scanline) {
        if (irq_enabled != 0) {
            if (patch != 0) {
                if (irq_counter == 0) {
                    nes.do_irq();
                    irq_enabled = 0;
                } else {
                    irq_counter--;
                }
            } else {
                if (irq_counter <= 113) {
                    nes.do_irq();
                    irq_enabled = 0;
                    irq_counter = 0xFFFF;
                } else {
                    irq_counter -= 113;
                }
            }
        }
    }
}
