public final class Mapper068 extends Mapper {

    private int[] regs;

    public Mapper068(final NESCore nes) {
        super(nes);
        regs = new int[4];
    }

    public final void Reset() {
        set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
        regs[0] = 0;
        regs[1] = 0;
        regs[2] = 0;
        regs[3] = 0;
    }

    public final void MemoryWrite(int addr, int data) {
        switch (addr & 0xF000) {
            case 0x8000 :
                set_PPU_bank0(data * 2 + 0);
                set_PPU_bank1(data * 2 + 1);
                break;
            case 0x9000 :
                set_PPU_bank2(data * 2 + 0);
                set_PPU_bank3(data * 2 + 1);
                break;
            case 0xA000 :
                set_PPU_bank4(data * 2 + 0);
                set_PPU_bank5(data * 2 + 1);
                break;
            case 0xB000 :
                set_PPU_bank6(data * 2 + 0);
                set_PPU_bank7(data * 2 + 1);
                break;
            case 0xC000 :
                regs[2] = data;
                SyncMirror();
                break;
            case 0xD000 :
                regs[3] = data;
                SyncMirror();
                break;
            case 0xE000 :
                regs[0] = (data & 0x10) >> 4;
                regs[1] = data & 0x03;
                SyncMirror();
                break;
            case 0xF000 :
                set_CPU_bank4(data * 2);
                set_CPU_bank5(data * 2 + 1);
                break;
        }
    }

    public final void SyncMirror() {
        if (regs[0] != 0) {
            if (regs[1] == 0) {
                nes.set_ppu_bank(8, regs[2] + 0x80);
                nes.set_ppu_bank(9, regs[3] + 0x80);
                nes.set_ppu_bank(10, regs[2] + 0x80);
                nes.set_ppu_bank(11, regs[3] + 0x80);
            } else if (regs[1] == 1) {
                nes.set_ppu_bank(8, regs[2] + 0x80);
                nes.set_ppu_bank(9, regs[2] + 0x80);
                nes.set_ppu_bank(10, regs[3] + 0x80);
                nes.set_ppu_bank(11, regs[3] + 0x80);
            } else if (regs[1] == 2) {
                nes.set_ppu_bank(8, regs[2] + 0x80);
                nes.set_ppu_bank(9, regs[2] + 0x80);
                nes.set_ppu_bank(10, regs[2] + 0x80);
                nes.set_ppu_bank(11, regs[2] + 0x80);
            } else if (regs[1] == 3) {
                nes.set_ppu_bank(8, regs[3] + 0x80);
                nes.set_ppu_bank(9, regs[3] + 0x80);
                nes.set_ppu_bank(10, regs[3] + 0x80);
                nes.set_ppu_bank(11, regs[3] + 0x80);
            }
        } else {
            if (regs[1] == 0) {
                set_mirroring(NESCore.MIRROR_VERT);
            } else if (regs[1] == 1) {
                set_mirroring(NESCore.MIRROR_HORIZ);
            } else if (regs[1] == 2) {
                set_mirroring(0, 0, 0, 0);
            } else if (regs[1] == 3) {
                set_mirroring(1, 1, 1, 1);
            }
        }
    }
}
