public final class Mapper118 extends Mapper {

    private int[] regs;
    private int prg0, prg1;
    private int chr0, chr1, chr2, chr3, chr4, chr5, chr6, chr7;

    private final int chr_swap() {
        return regs[0] & 0x80;
    }

    private final int prg_swap() {
        return regs[0] & 0x40;
    }

    private int irq_enabled;
    private int irq_counter;
    private int irq_latch;

    public Mapper118(final NESCore nes) {
        super(nes);
        regs = new int[8];
    }

    public final void Reset() {
        for (int i = 0; i < 8; i++) regs[i] = 0x00;
        prg0 = 0;
        prg1 = 1;
        MMC3_set_CPU_banks();
        if ((num_1k_VROM_banks) != 0) {
            chr0 = 0;
            chr1 = 1;
            chr2 = 2;
            chr3 = 3;
            chr4 = 4;
            chr5 = 5;
            chr6 = 6;
            chr7 = 7;
            MMC3_set_PPU_banks();
        } else {
            chr0 = chr1 = chr2 = chr3 = chr4 = chr5 = chr6 = chr7 = 0;
        }
        irq_enabled = 0;
        irq_counter = 0;
        irq_latch = 0;
    }

    public final void MemoryWrite(int addr, int data) {
        switch (addr & 0xE001) {
            case 0x8000 :
                regs[0] = data;
                MMC3_set_PPU_banks();
                MMC3_set_CPU_banks();
                break;
            case 0x8001 :
                int bank_num;
                regs[1] = data;
                bank_num = regs[1];
                if ((regs[0] & 0x07) < 6) {
                    if ((data & 0x80) != 0) {
                        set_mirroring(0, 0, 0, 0);
                    } else {
                        set_mirroring(1, 1, 1, 1);
                    }
                }
                switch (regs[0] & 0x07) {
                    case 0x00 :
                        if ((num_1k_VROM_banks) != 0) {
                            bank_num &= 0xfe;
                            chr0 = bank_num;
                            chr1 = bank_num + 1;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x01 :
                        if ((num_1k_VROM_banks) != 0) {
                            bank_num &= 0xfe;
                            chr2 = bank_num;
                            chr3 = bank_num + 1;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x02 :
                        if ((num_1k_VROM_banks) != 0) {
                            chr4 = bank_num;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x03 :
                        if ((num_1k_VROM_banks) != 0) {
                            chr5 = bank_num;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x04 :
                        if ((num_1k_VROM_banks) != 0) {
                            chr6 = bank_num;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x05 :
                        if ((num_1k_VROM_banks) != 0) {
                            chr7 = bank_num;
                            MMC3_set_PPU_banks();
                        }
                        break;
                    case 0x06 :
                        prg0 = bank_num;
                        MMC3_set_CPU_banks();
                        break;
                    case 0x07 :
                        prg1 = bank_num;
                        MMC3_set_CPU_banks();
                        break;
                }
                break;
            case 0xC000 :
                regs[4] = data;
                irq_counter = regs[4];
                break;
            case 0xC001 :
                regs[5] = data;
                irq_latch = regs[5];
                break;
            case 0xE000 :
                regs[6] = data;
                irq_enabled = 0;
                break;
            case 0xE001 :
                regs[7] = data;
                irq_enabled = 1;
                break;
        }
    }

    public final void HSync(final int scanline) {
        if (irq_enabled != 0) {
            if ((scanline >= 0) && (scanline <= 239)) {
                if ((nes.PPU_R1 & 0x18) != 0) {
                    if ((irq_counter--) == 0) {
                        irq_counter = irq_latch;
                        nes.do_irq();
                    }
                }
            }
        }
    }

    private final void MMC3_set_CPU_banks() {
        if (prg_swap() != 0) {
            set_CPU_banks(num_8k_ROM_banks - 2, prg1, prg0, num_8k_ROM_banks - 1);
        } else {
            set_CPU_banks(prg0, prg1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
        }
    }

    private final void MMC3_set_PPU_banks() {
        if ((num_1k_VROM_banks) != 0) {
            if (chr_swap() != 0) {
                set_PPU_banks(chr4, chr5, chr6, chr7, chr0, chr1, chr2, chr3);
            } else {
                set_PPU_banks(chr0, chr1, chr2, chr3, chr4, chr5, chr6, chr7);
            }
        }
    }
}
