public class MapperOther extends Mapper {

    private boolean irq_enabled;
    private int lines_to_irq;

    public MapperOther(final NESCore nes) {
        super(nes);
        irq_enabled = false;
        lines_to_irq = 0;
    }

    public final void Reset() {
        switch (mapper_no) {
            case 0 :
                if (num_8k_ROM_banks > 2) {
                    set_CPU_banks(0, 1, 2, 3);
                } else if (num_8k_ROM_banks > 1) {
                    set_CPU_banks(0, 1, 0, 1);
                } else {
                    set_CPU_banks(0, 0, 0, 0);
                }
                if (num_1k_VROM_banks != 0) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                }
                break;
            case 2 :
            case 94 :
                set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                break;
            case 3 :
                if (num_8k_ROM_banks > 2) {
                    set_CPU_banks(0, 1, 2, 3);
                } else {
                    set_CPU_banks(0, 1, 0, 1);
                }
                set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                break;
            case 7 :
            case 77 :
            case 122 :
            case 184 :
                set_CPU_banks(0, 1, 2, 3);
                break;
            case 40 :
            	nes.set_cpu_bank(3, 6);
                set_CPU_banks(4, 5, 0, 7);
                if(num_1k_VROM_banks != 0)
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                break;
            case 34 :
            case 72 :
            case 78 :
            case 92 :
                set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                if (num_1k_VROM_banks != 0) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                }
                break;
            case 70 :
                patch = 0;
                if (nes.crc == 0xa59ca2ef || nes.crc == 0x10bb8f9a) {
                    patch = 1;
                }
                set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                break;
            case 80 :
                patch = 0;
                if (nes.crc == 0x9832d15a) {
                    patch = 1;
                }
                set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                if (num_1k_VROM_banks != 0) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                }
                break;
            case 86 :
            case 101 :
                set_CPU_banks(0, 1, 2, 3);
                set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                break;
            case 87 :
                set_CPU_banks(0, 1, 2, 3);
                if (num_1k_VROM_banks != 0) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                }
                break;
            case 89 :
            case 93 :
            case 102 :
                set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                break;
            case 96 :
                set_CPU_banks(0, 1, 2, 3);
                set_mirroring(0, 0, 0, 0);
                break;
            case 180 :
                set_CPU_banks(0, 1, 2, 3);
                if (num_1k_VROM_banks != 0) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                }
                if (nes.crc == 0xc68363f6) {
                    set_mirroring(NESCore.MIRROR_HORIZ);
                }
                break;
            case 185 :
                set_CPU_banks(0, 1, 2, 3);
                patch = 0;
                if (nes.crc == 0xb36457c7) {
                    patch = 1;
                }
                break;
            case 188 :
                if (num_8k_ROM_banks > 16) {
                    set_CPU_banks(0, 1, 14, 15);
                } else {
                    set_CPU_banks(0, 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                }
                break;
        }
    }

    public void MemoryWriteSaveRAM(int addr, int data) {
        switch (mapper_no) {
            case 34 :
                switch (addr) {
                    case 0x7FFD :
                        set_CPU_banks(data * 4, data * 4 + 1, data * 4 + 2, data * 4 + 3);
                        break;
                    case 0x7FFE :
                        set_PPU_bank0(data * 4 + 0);
                        set_PPU_bank1(data * 4 + 1);
                        set_PPU_bank2(data * 4 + 2);
                        set_PPU_bank3(data * 4 + 3);
                        break;
                    case 0x7FFF :
                        set_PPU_bank4(data * 4 + 0);
                        set_PPU_bank5(data * 4 + 1);
                        set_PPU_bank6(data * 4 + 2);
                        set_PPU_bank7(data * 4 + 3);
                        break;
                }
                break;
            case 80 :
                switch (addr) {
                    case 0x7EF0 :
                        set_PPU_bank0(data & 0x7F);
                        set_PPU_bank1((data & 0x7F) + 1);
                        if (patch != 0) {
                            if ((data & 0x80) != 0) {
                                set_VRAM_bank(0x8, 1);
                                set_VRAM_bank(0x9, 1);
                            } else {
                                set_VRAM_bank(0x8, 0);
                                set_VRAM_bank(0x9, 0);
                            }
                        }
                        break;
                    case 0x7EF1 :
                        set_PPU_bank2(data & 0x7F);
                        set_PPU_bank3((data & 0x7F) + 1);
                        if (patch != 0) {
                            if ((data & 0x80) != 0) {
                                set_VRAM_bank(0xA, 1);
                                set_VRAM_bank(0xB, 1);
                            } else {
                                set_VRAM_bank(0xA, 0);
                                set_VRAM_bank(0xB, 0);
                            }
                        }
                        break;
                    case 0x7EF2 :
                        set_PPU_bank4(data);
                        break;
                    case 0x7EF3 :
                        set_PPU_bank5(data);
                        break;
                    case 0x7EF4 :
                        set_PPU_bank6(data);
                        break;
                    case 0x7EF5 :
                        set_PPU_bank7(data);
                        break;
                    case 0x7EF6 :
                        if ((data & 0x01) != 0) {
                            set_mirroring(NESCore.MIRROR_VERT);
                        } else {
                            set_mirroring(NESCore.MIRROR_HORIZ);
                        }
                        break;
                    case 0x7EFA :
                    case 0x7EFB :
                        set_CPU_bank4(data);
                        break;
                    case 0x7EFC :
                    case 0x7EFD :
                        set_CPU_bank5(data);
                        break;
                    case 0x7EFE :
                    case 0x7EFF :
                        set_CPU_bank6(data);
                        break;
                }
                break;
            case 86 :
                if (addr == 0x6000) {
                    int chr_bank = data & 0x03 | (data & 0x40) >> 4;
                    int prg_bank = (data & 0x30) >> 4;
                    set_CPU_bank4(prg_bank * 4 + 0);
                    set_CPU_bank5(prg_bank * 4 + 1);
                    set_CPU_bank6(prg_bank * 4 + 2);
                    set_CPU_bank7(prg_bank * 4 + 3);
                    set_PPU_bank0(chr_bank * 8 + 0);
                    set_PPU_bank1(chr_bank * 8 + 1);
                    set_PPU_bank2(chr_bank * 8 + 2);
                    set_PPU_bank3(chr_bank * 8 + 3);
                    set_PPU_bank4(chr_bank * 8 + 4);
                    set_PPU_bank5(chr_bank * 8 + 5);
                    set_PPU_bank6(chr_bank * 8 + 6);
                    set_PPU_bank7(chr_bank * 8 + 7);
                }
                break;
            case 87 :
                if (addr == 0x6000) {
                    int chr_bank = (data & 0x02) >> 1;
                    chr_bank <<= 3;
                    set_PPU_banks(
                        chr_bank + 0,
                        chr_bank + 1,
                        chr_bank + 2,
                        chr_bank + 3,
                        chr_bank + 4,
                        chr_bank + 5,
                        chr_bank + 6,
                        chr_bank + 7);
                }
                break;
            case 93 :
            case 102 :
                if (addr == 0x6000) {
                    set_CPU_bank4(data * 2 + 0);
                    set_CPU_bank5(data * 2 + 1);
                }
                break;
            case 101 :
                data &= 0x03;
                set_PPU_bank0(data * 8 + 0);
                set_PPU_bank1(data * 8 + 1);
                set_PPU_bank2(data * 8 + 2);
                set_PPU_bank3(data * 8 + 3);
                set_PPU_bank4(data * 8 + 4);
                set_PPU_bank5(data * 8 + 5);
                set_PPU_bank6(data * 8 + 6);
                set_PPU_bank7(data * 8 + 7);
                break;
            case 122 :
            case 184 :
                if (addr == 0x6000) {
                    int chr_bank0 = data & 0x07;
                    int chr_bank1 = (data & 0x70) >> 4;
                    chr_bank0 <<= 2;
                    chr_bank1 <<= 2;
                    set_PPU_banks(
                        chr_bank0 + 0,
                        chr_bank0 + 1,
                        chr_bank0 + 2,
                        chr_bank0 + 3,
                        chr_bank1 + 0,
                        chr_bank1 + 1,
                        chr_bank1 + 2,
                        chr_bank1 + 3);
                }
                break;
        }
    }

    public void MemoryWrite(int addr, int data) {
        switch (mapper_no) {
            case 2 :
                data &= num_8k_ROM_banks - 1;
                set_CPU_banks(
                    data * 2,
                    (data * 2) + 1,
                    num_8k_ROM_banks - 2,
                    num_8k_ROM_banks - 1);
                break;
            case 3 :
                int base;
                data &= (num_1k_VROM_banks >> 1) - 1;
                base = (data) << 3;
                set_PPU_banks(
                    base + 0,
                    base + 1,
                    base + 2,
                    base + 3,
                    base + 4,
                    base + 5,
                    base + 6,
                    base + 7);
                break;
            case 7 :
                int bank;
                bank = (data & 0x07) << 2;
                set_CPU_banks(bank + 0, bank + 1, bank + 2, bank + 3);
                if ((data & 0x10) != 0) {
                    set_mirroring(1, 1, 1, 1);
                } else {
                    set_mirroring(0, 0, 0, 0);
                }
                break;
            case 34 :
                set_CPU_banks(data * 4, data * 4 + 1, data * 4 + 2, data * 4 + 3);
                break;
            case 40 :
                switch(addr & 0xE000) {
                case 0x8000:
                    irq_enabled = false;
                    break;
                case 0xA000:
                    irq_enabled = true;
                    lines_to_irq = 37;
                    break;
                case 0xC000:
                    break;
                case 0xE000:
                    set_CPU_bank6(data & 0x07);
                     break;
                }
                break;
            case 70 :
                int chr_bank = data & 0x0F;
                int prg_bank = (data & 0x70) >> 4;
                set_CPU_bank4(prg_bank * 2 + 0);
                set_CPU_bank5(prg_bank * 2 + 1);
                chr_bank <<= 3;
                set_PPU_banks(
                    chr_bank + 0,
                    chr_bank + 1,
                    chr_bank + 2,
                    chr_bank + 3,
                    chr_bank + 4,
                    chr_bank + 5,
                    chr_bank + 6,
                    chr_bank + 7);
                if (patch != 0) {
                    if ((data & 0x80) != 0) {
                        set_mirroring(NESCore.MIRROR_HORIZ);
                    } else {
                        set_mirroring(NESCore.MIRROR_VERT);
                    }
                } else {
                    if ((data & 0x80) != 0) {
                        set_mirroring(1, 1, 1, 1);
                    } else {
                        set_mirroring(0, 0, 0, 0);
                    }
                }
                break;
            case 72 :
                int bank_72 = data & 0x0f;
                bank_72 = data & 0x0f;
                if ((data & 0x80) != 0) {
                    set_CPU_banks(bank_72 * 2, bank_72 * 2 + 1, num_8k_ROM_banks - 2, num_8k_ROM_banks - 1);
                }
                if ((data & 0x40) != 0) {
                    set_PPU_banks(
                        bank_72 * 8,
                        bank_72 * 8 + 1,
                        bank_72 * 8 + 2,
                        bank_72 * 8 + 3,
                        bank_72 * 8 + 4,
                        bank_72 * 8 + 5,
                        bank_72 * 8 + 6,
                        bank_72 * 8 + 7);
                }
                break;
            case 77 :
                int prg_bank_77 = data & 0x07;
                int chr_bank_77 = (data & 0xF0) >> 4;
                set_CPU_bank4(prg_bank_77 * 4 + 0);
                set_CPU_bank5(prg_bank_77 * 4 + 1);
                set_CPU_bank6(prg_bank_77 * 4 + 2);
                set_CPU_bank7(prg_bank_77 * 4 + 3);
                set_PPU_bank0(chr_bank_77 * 2 + 0);
                set_PPU_bank1(chr_bank_77 * 2 + 1);
                break;
            case 78 :
                int prg_bank_78 = data & 0x0F;
                int chr_bank_78 = (data & 0xF0) >> 4;
                set_CPU_bank4(prg_bank_78 * 2 + 0);
                set_CPU_bank5(prg_bank_78 * 2 + 1);
                set_PPU_bank0(chr_bank_78 * 8 + 0);
                set_PPU_bank1(chr_bank_78 * 8 + 1);
                set_PPU_bank2(chr_bank_78 * 8 + 2);
                set_PPU_bank3(chr_bank_78 * 8 + 3);
                set_PPU_bank4(chr_bank_78 * 8 + 4);
                set_PPU_bank5(chr_bank_78 * 8 + 5);
                set_PPU_bank6(chr_bank_78 * 8 + 6);
                set_PPU_bank7(chr_bank_78 * 8 + 7);
                if ((addr & 0xFE00) != 0xFE00) {
                    if ((data & 0x08) != 0) {
                        set_mirroring(1, 1, 1, 1);
                    } else {
                        set_mirroring(0, 0, 0, 0);
                    }
                }
                break;
            case 89 :
                if ((addr & 0xFF00) == 0xC000) {
                    int prg_bank_89 = (data & 0x70) >> 4;
                    int chr_bank_89 = ((data & 0x80) >> 4) | (data & 0x07);
                    set_CPU_bank4(prg_bank_89 * 2 + 0);
                    set_CPU_bank5(prg_bank_89 * 2 + 1);
                    set_PPU_bank0(chr_bank_89 * 8 + 0);
                    set_PPU_bank1(chr_bank_89 * 8 + 1);
                    set_PPU_bank2(chr_bank_89 * 8 + 2);
                    set_PPU_bank3(chr_bank_89 * 8 + 3);
                    set_PPU_bank4(chr_bank_89 * 8 + 4);
                    set_PPU_bank5(chr_bank_89 * 8 + 5);
                    set_PPU_bank6(chr_bank_89 * 8 + 6);
                    set_PPU_bank7(chr_bank_89 * 8 + 7);
                    if ((data & 0x08) != 0) {
                        set_mirroring(1, 1, 1, 1);
                    } else {
                        set_mirroring(0, 0, 0, 0);
                    }
                }
                break;
            case 92 :
                data = (int) addr & 0xff;
                int c_bank = (data & 0x0f) << 1;
                int p_bank = data & 0x0f;
                if (addr >= 0x9000) {
                    if ((data & 0xf0) == 0xd0) {
                        set_CPU_banks(0, 1, c_bank, c_bank + 1);
                    } else if ((data & 0xf0) == 0xe0) {
                        set_PPU_banks(
                            p_bank * 8,
                            p_bank * 8 + 1,
                            p_bank * 8 + 2,
                            p_bank * 8 + 3,
                            p_bank * 8 + 4,
                            p_bank * 8 + 5,
                            p_bank * 8 + 6,
                            p_bank * 8 + 7);
                    }
                } else {
                    if ((data & 0xf0) == 0xb0) {
                        set_CPU_banks(0, 1, c_bank, c_bank + 1);
                    } else if ((data & 0xf0) == 0x70) {
                        set_PPU_banks(
                            p_bank * 8,
                            p_bank * 8 + 1,
                            p_bank * 8 + 2,
                            p_bank * 8 + 3,
                            p_bank * 8 + 4,
                            p_bank * 8 + 5,
                            p_bank * 8 + 6,
                            p_bank * 8 + 7);
                    }
                }
                break;
            case 94 :
                if ((addr & 0xFFF0) == 0xFF00) {
                    int bank_94 = (data & 0x1C) >> 2;
                    set_CPU_bank4(bank_94 * 2 + 0);
                    set_CPU_bank5(bank_94 * 2 + 1);
                }
                break;
            case 96 :
                set_CPU_bank4((data & 0x03) * 4 + 0);
                set_CPU_bank5((data & 0x03) * 4 + 1);
                set_CPU_bank6((data & 0x03) * 4 + 2);
                set_CPU_bank7((data & 0x03) * 4 + 3);
                break;
            case 101 :
                data &= 0x03;
                set_PPU_bank0(data * 8 + 0);
                set_PPU_bank1(data * 8 + 1);
                set_PPU_bank2(data * 8 + 2);
                set_PPU_bank3(data * 8 + 3);
                set_PPU_bank4(data * 8 + 4);
                set_PPU_bank5(data * 8 + 5);
                set_PPU_bank6(data * 8 + 6);
                set_PPU_bank7(data * 8 + 7);
                break;
            case 180 :
                set_CPU_bank6((data & 0x07) * 2 + 0);
                set_CPU_bank7((data & 0x07) * 2 + 1);
                break;
            case 185 :
                if ((patch == 0 && (data & 0x03) != 0)
                        || (patch != 0 && data == 0x21)) {
                    set_PPU_banks(0, 1, 2, 3, 4, 5, 6, 7);
                } else {
                    set_PPU_dummy_bank(0);
                    set_PPU_dummy_bank(1);
                    set_PPU_dummy_bank(2);
                    set_PPU_dummy_bank(3);
                    set_PPU_dummy_bank(4);
                    set_PPU_dummy_bank(5);
                    set_PPU_dummy_bank(6);
                    set_PPU_dummy_bank(7);
                }
                break;
            case 188 :
                if (data != 0) {
                    if ((data & 0x10) != 0) {
                        data &= 0x07;
                        set_CPU_bank4(data * 2);
                        set_CPU_bank5(data * 2 + 1);
                    } else {
                        set_CPU_bank4(data * 2 + 16);
                        set_CPU_bank5(data * 2 + 17);
                    }
                } else {
                    if (num_8k_ROM_banks == 0x10) {
                        set_CPU_bank4(14);
                        set_CPU_bank5(15);
                    } else {
                        set_CPU_bank4(16);
                        set_CPU_bank5(17);
                    }
                }
                break;
        }
    }

    public final void HSync(final int scanline) {
        if(irq_enabled) {
            if((--lines_to_irq) <= 0) {
                nes.do_irq();
            }
        }
    }
}
