import com.nttdocomo.ui.AudioPresenter;

import exsound.*;
import dojapcmsound.PcmAudioHandler;

public class NESApu2 implements NESApu{
	
	public static final int APU_FILTER_NONE = 0;
	public static final int APU_FILTER_LOWPASS = 1;
	public static final int APU_FILTER_WEIGHTED = 2;	
	
	private final int APU_WRA0 = 0x4000;
	private final int APU_WRA1 = 0x4001;
	private final int APU_WRA2 = 0x4002;
	private final int APU_WRA3 = 0x4003;
	private final int APU_WRB0 = 0x4004;
	private final int APU_WRB1 = 0x4005;
	private final int APU_WRB2 = 0x4006;
	private final int APU_WRB3 = 0x4007;
	private final int APU_WRC0 = 0x4008;
	private final int APU_WRC2 = 0x400A;
	private final int APU_WRC3 = 0x400B;
	private final int APU_WRD0 = 0x400C;
	private final int APU_WRD2 = 0x400E;
	private final int APU_WRD3 = 0x400F;
	private final int APU_WRE0 = 0x4010;
	private final int APU_WRE1 = 0x4011;
	private final int APU_WRE2 = 0x4012;
	private final int APU_WRE3 = 0x4013;
	private final int APU_SMASK = 0x4015;

	private final double APU_BASEFREQ   = 1789772.5;
	
	/* length of generated noise */
	private final int  APU_NOISE_32K  = 0x7FFF;
	private final int  APU_NOISE_93   = 93;
	
	/* APU queue structure */
	private final int APUQUEUE_SIZE  = 4096;
	private final int APUQUEUE_MASK  = (APUQUEUE_SIZE - 1);
	
	/* ExSounds */
	private final int VRC6 = 1;
	private final int VRC7 = 2;
	private final int FDS  = 4;
	private final int MMC5 = 8;
	private final int N106 = 16;
	private final int FME7 = 32;
	private final int J106 = 64; //(reserved)

	/* apu ring buffer member */
	class ApuData
	{
		int timestamp;
		int address;
		int value;
	}	
	
	/* channel output volume */
	private int APU_RECTANGLE_OUTPUT(RectangleCh chan) { return chan.output_vol; }
	private int APU_TRIANGLE_OUTPUT(TriangleCh chan) { return chan.output_vol + (chan.output_vol >> 2); }
	private int APU_NOISE_OUTPUT(NoiseCh chan) { return (chan.output_vol + chan.output_vol + chan.output_vol) >> 2; }
	private int APU_DMC_OUTPUT(DmcCh chan) { return (chan.output_vol + chan.output_vol + chan.output_vol) >> 2; }
	
	/* to/from 16.16 fixed point */
	private int APU_TO_FIXED(int x) { return x << 16; }
	private int APU_FROM_FIXED(int x) { return x >> 16; }	
	
	/* APUSOUND */
	private RectangleCh apu_rectangle[];
	private TriangleCh apu_triangle;
	private NoiseCh apu_noise;
	private DmcCh apu_dmc;

	private int apu_enable_reg;
	private int apu_enable_reg_cur;

	private ApuData apu_queue[];
	private int apu_q_head, apu_q_tail;

	private int apu_elapsed_cycles;
	private int apu_num_samples;

	private int apu_cycle_rate;
	private int apu_sample_rate;
	private int apu_sample_bits;
	private int apu_refresh_rate;

	private boolean apu_mix_enable[];
	private int apu_filter_type;
	
	/* apu opt flags */
	private boolean REALTIME_NOISE;
	
	/* for output filtering */
	private int prev_sample;
	
	/* for pseudo-random series generator of noise ch */
	private int sreg1 = 0x4000;
	private int sreg2 = 0x4000;
	
	/* for ExSound */
	private ApuData apu_ex_queue[];
	private int apu_ex_q_head, apu_ex_q_tail;
	private int apu_ex_chip;
	private ExSound ex;

	
	
	/* look up table madness */
	private int decay_lut[] = new int[16];
	private int vbl_lut[] = new int[32];
	private int trilength_lut[] = new int[128];

	/* noise lookups for both modes */
	private byte noise_long_lut[] = new byte[APU_NOISE_32K];
	private byte noise_short_lut[] = new byte[APU_NOISE_93];
	
	/* vblank length table used for rectangles, triangle, noise */
	private final int vbl_length[] =
	{
	    5, 127,
	   10,   1,
	   19,   2,
	   40,   3,
	   80,   4,
	   30,   5,
	    7,   6,
	   13,   7,
	    6,   8,
	   12,   9,
	   24,  10,
	   48,  11,
	   96,  12,
	   36,  13,
	    8,  14,
	   16,  15
	};

	/* frequency limit of rectangle channels */
	private final int freq_limit[] =
	{
	   0x3FF, 0x555, 0x666, 0x71C, 0x787, 0x7C1, 0x7E0, 0x7F0
	};

	/* noise frequency lookup table */
	private final int noise_freq[] =
	{
	     4,    8,   16,   32,   64,   96,  128,  160,
	   202,  254,  380,  508,  762, 1016, 2034, 4068
	};

	/* DMC transfer freqs */
	private final int dmc_clocks[] =
	{
	   428, 380, 340, 320, 286, 254, 226, 214,
	   190, 160, 142, 128, 106,  85,  72,  54
	};

	/* ratios of pos/neg pulse for rectangle waves */
	private final int duty_lut[] = { 2, 4, 8, 12 };

	private boolean apu_enable;
//	 ----------------------------------------------------------------------------
//	 APU Sound struct
	
	/* channel structures */
	/* As much data as possible is precalculated,
	** to keep the sample processing as lean as possible
	*/
	class RectangleCh
	{
	   int regs[] = new int[4];

	   boolean enabled;
	   
	   int phaseacc;
	   int freq;
	   int output_vol;
	   boolean fixed_envelope;
	   boolean holdnote;
	   int volume;

	   int sweep_phase;
	   int sweep_delay;
	   boolean sweep_on;
	   int sweep_shifts;
	   int sweep_length;
	   boolean sweep_inc;

	   /* this may not be necessary in the future */
	   int freq_limit;

	   /* rectangle 0 uses a complement addition for sweep
	   ** increases, while rectangle 1 uses subtraction
	   */
	   boolean sweep_complement;

	   int env_phase;
	   int env_delay;
	   int env_vol;

	   int vbl_length;
	   int adder;
	   int duty_flip;

	   /* for sync read $4105 */
	   boolean enabled_cur;
	   boolean holdnote_cur;
	   int vbl_length_cur;
	}

	class TriangleCh
	{
	   int regs[] = new int[3];

	   boolean enabled;

	   int freq;
	   int phaseacc;
	   int output_vol;

	   int adder;

	   boolean holdnote;
	   boolean counter_started;
	   /* quasi-hack */
	   int write_latency;

	   int vbl_length;
	   int linear_length;

	   /* for sync read $4105 */
	   boolean enabled_cur;
	   boolean holdnote_cur;
	   boolean counter_started_cur;
	   int vbl_length_cur;

	   /* less compatibility, clearer sound if enabled */
	   boolean ideal_triangle;
	}

	class NoiseCh
	{
	   int regs[] = new int[3];

	   boolean enabled;

	   int freq;
	   int phaseacc;
	   int output_vol;

	   int env_phase;
	   int env_delay;
	   int env_vol;
	   boolean fixed_envelope;
	   boolean holdnote;

	   int volume;

	   int vbl_length;

	   int xor_tap;
	   
	   boolean short_sample;
	   int cur_pos;

	   /* for sync read $4105 */
	   boolean enabled_cur;
	   boolean holdnote_cur;
	   int vbl_length_cur;
	}

	class DmcCh
	{
	   int regs[] = new int[4];

	   /* bodge for timestamp queue */
	   boolean enabled;
	   
	   int freq;
	   int phaseacc;
	   int output_vol;

	   int address;
	   int cached_addr;
	   int dma_length;
	   int cached_dmalength;
	   int cur_byte;

	   boolean looping;
	   boolean irq_gen;
	   boolean irq_occurred;

	   /* for sync read $4105 and DPCM IRQ */
	   int freq_cur;
	   int phaseacc_cur;
	   int dma_length_cur;
	   int cached_dmalength_cur;
	   boolean enabled_cur;
	   boolean looping_cur;
	   boolean irq_gen_cur;
	   boolean irq_occurred_cur;
	}
//	----------------------------------------------------------------------------
//	APU Sound struct

	private void apu_enqueue(ApuData d) {
	   apu_queue[apu_q_head] = d;

	   apu_q_head = (apu_q_head + 1) & APUQUEUE_MASK;

	   if (apu_q_head == apu_q_tail) /* APU_QEMPT */
		System.out.println("apu: queue overflow\n");      
	}

	private ApuData apu_dequeue() {
	   int loc;

	   if (apu_q_head == apu_q_tail) /* APU_QEMPT */
		System.out.println("apu: queue empty\n");

	   loc = apu_q_tail;
	   apu_q_tail = (apu_q_tail + 1) & APUQUEUE_MASK;

	   return apu_queue[loc];
	}

	private void apu_ex_enqueue(ApuData d) {
	   apu_ex_queue[apu_ex_q_head] = d;

	   apu_ex_q_head = (apu_ex_q_head + 1) & APUQUEUE_MASK;

	   if (apu_ex_q_head == apu_ex_q_tail) /* EX_QEMPT */
		System.out.println("ex_apu: queue overflow\n");      
	}

	private ApuData apu_ex_dequeue() {
	   int loc;

	   if (apu_ex_q_head == apu_ex_q_tail) /* EX_QEMPT */
		System.out.println("ex_apu: queue empty\n");

	   loc = apu_ex_q_tail;
	   apu_ex_q_tail = (apu_ex_q_tail + 1) & APUQUEUE_MASK;

	   return apu_ex_queue[loc];
	}



	/* emulation of the 15-bit shift register the
	** NES uses to generate pseudo-random series
	** for the white noise channel
	*/
	private int shift_register15(int xor_tap) {
	   int bit0, tap, bit14;

	   bit0 = sreg1 & 1;
	   tap = (sreg1 & xor_tap) != 0 ? 1 : 0;
	   bit14 = (bit0 ^ tap);
	   sreg1 >>= 1;
	   sreg1 |= (bit14 << 14);
	   return (bit0 ^ 1);
	}
	
	private void shift_register15(byte[] buf, int count) {
	   int bit0, bit1, bit6, bit14;
	   int p = 0;

	   if (count == APU_NOISE_93)
	   {
	      while (count-- != 0)
	      {
	         bit0 = sreg2 & 1;
	         bit6 = (sreg2 & 0x40) >> 6;
	         bit14 = (bit0 ^ bit6);
	         sreg2 >>= 1;
	         sreg2 |= (bit14 << 14);
	         buf[p++] = (byte)(bit0 ^ 1);
	      }
	   }
	   else /* 32K noise */
	   {
	      while (count-- != 0)
	      {
	         bit0 = sreg2 & 1;
	         bit1 = (sreg2 & 2) >> 1;
	         bit14 = (bit0 ^ bit1);
	         sreg2 >>= 1;
	         sreg2 |= (bit14 << 14);
	         buf[p++] = (byte)(bit0 ^ 1);
	      }
	   }
	}

	/* RECTANGLE WAVE
	** ==============
	** reg0: 0-3=volume, 4=envelope, 5=hold, 6-7=duty cycle
	** reg1: 0-2=sweep shifts, 3=sweep inc/dec, 4-6=sweep length, 7=sweep on
	** reg2: 8 bits of freq
	** reg3: 0-2=high freq, 7-4=vbl length counter
	*/
	private int apu_rectangle(RectangleCh chan) {
	   int output;
	   int num_times;
	   int total;

	   /* APU_VOLUME_DECAY */
	   chan.output_vol -= (chan.output_vol >> 7);

	   if (false == chan.enabled || 0 == chan.vbl_length)
		   return APU_RECTANGLE_OUTPUT(chan); 

	   /* vbl length counter */
	   if (false == chan.holdnote)
	      chan.vbl_length--;

	   /* envelope decay at a rate of (env_delay + 1) / 240 secs */
	   chan.env_phase -= 4; /* 240/60 */
	   while (chan.env_phase < 0)
	   {
	      chan.env_phase += chan.env_delay;

	      if (chan.holdnote)
	         chan.env_vol = (chan.env_vol + 1) & 0x0F;
	      else if (chan.env_vol < 0x0F)
	         chan.env_vol++;
	   }

	   /* TODO: using a table of max frequencies is not technically
	   ** clean, but it is fast and (or should be) accurate 
	   */
	   if (chan.freq < 8 || (false == chan.sweep_inc && chan.freq > chan.freq_limit))
		   return APU_RECTANGLE_OUTPUT(chan); 

	   /* frequency sweeping at a rate of (sweep_delay + 1) / 120 secs */
	   if (chan.sweep_on && 0 != chan.sweep_shifts)
	   {
	      chan.sweep_phase -= 2; /* 120/60 */
	      while (chan.sweep_phase < 0)
	      {
	         chan.sweep_phase += chan.sweep_delay;

	         if (chan.sweep_inc) /* ramp up */
	         {
	            if (true == chan.sweep_complement)
	               chan.freq += ~(chan.freq >> chan.sweep_shifts);
	            else
	               chan.freq -= (chan.freq >> chan.sweep_shifts);
	         }
	         else /* ramp down */
	         {
	            chan.freq += (chan.freq >> chan.sweep_shifts);
	         }
	      }
	   }

	   chan.phaseacc -= apu_cycle_rate; /* # of cycles per sample */
	   if (chan.phaseacc >= 0)
		   return APU_RECTANGLE_OUTPUT(chan); 

	   num_times = total = 0;

	   if (chan.fixed_envelope)
	      output = chan.volume << 8; /* fixed volume */
	   else
	      output = (chan.env_vol ^ 0x0F) << 8;

	   while (chan.phaseacc < 0)
	   {
	      chan.phaseacc += APU_TO_FIXED(chan.freq + 1);
	      chan.adder = (chan.adder + 1) & 0x0F;

	      if (chan.adder < chan.duty_flip)
	         total += output;
	      else
	         total -= output;

	      num_times++;
	   }

	   chan.output_vol = total / num_times;

	   return APU_RECTANGLE_OUTPUT(chan); 
	}

	/* TRIANGLE WAVE
	 ** =============
	 ** reg0: 7=holdnote, 6-0=linear length counter
	 ** reg2: low 8 bits of frequency
	 ** reg3: 7-3=length counter, 2-0=high 3 bits of frequency
	 */
	private int apu_triangle(TriangleCh chan) {
		/* APU_VOLUME_DECAY */
		chan.output_vol -= (chan.output_vol >> 7);
		
		if (false == chan.enabled || 0 == chan.vbl_length)
			return APU_TRIANGLE_OUTPUT(chan);

		if (chan.counter_started)
		{
			if (chan.linear_length > 0)
				chan.linear_length--;
			if (chan.vbl_length != 0 && false == chan.holdnote)
				chan.vbl_length--;
		}
		else if (false == chan.holdnote && chan.write_latency != 0)
		{
			if (--chan.write_latency == 0)
				chan.counter_started = true;
		}

		if (0 == chan.linear_length || chan.freq < 65536) /* inaudible */
			return APU_TRIANGLE_OUTPUT(chan);

		chan.phaseacc -= apu_cycle_rate; /* # of cycles per sample */
		while (chan.phaseacc < 0)
		{
			chan.phaseacc += chan.freq;
			chan.adder = (chan.adder + 1) & 0x1F;

			if ( 0 != (chan.adder & 0x10) )
				chan.output_vol -= 512;
			else
				chan.output_vol += 512;
		}

		return APU_TRIANGLE_OUTPUT(chan);
	}


	/* WHITE NOISE CHANNEL
	** ===================
	** reg0: 0-3=volume, 4=envelope, 5=hold
	** reg2: 7=small(93 byte) sample,3-0=freq lookup
	** reg3: 7-4=vbl length counter
	*/
	private int apu_noise(NoiseCh chan) {
		int outvol;
		int noise_bit;
		int num_times;
		int total;

		/* APU_VOLUME_DECAY */
		chan.output_vol -= (chan.output_vol >> 7);

		if (false == chan.enabled || 0 == chan.vbl_length)
			return APU_NOISE_OUTPUT(chan);

			/* vbl length counter */
			if (false == chan.holdnote)
				chan.vbl_length--;

			/* envelope decay at a rate of (env_delay + 1) / 240 secs */
			chan.env_phase -= 4; /* 240/60 */
			while (chan.env_phase < 0)
			{
				chan.env_phase += chan.env_delay;

				if (chan.holdnote)
					chan.env_vol = (chan.env_vol + 1) & 0x0F;
				else if (chan.env_vol < 0x0F)
					chan.env_vol++;
			}

			chan.phaseacc -= apu_cycle_rate; /* # of cycles per sample */
			if (chan.phaseacc >= 0)
				return APU_NOISE_OUTPUT(chan);

				num_times = total = 0;

				if (chan.fixed_envelope)
					outvol = chan.volume << 8; /* fixed volume */
				else
					outvol = (chan.env_vol ^ 0x0F) << 8;

				while (chan.phaseacc < 0)
				{
					chan.phaseacc += chan.freq;

					if(REALTIME_NOISE) {

						if (shift_register15(chan.xor_tap) != 0)
							total += outvol;
						else
							total -= outvol;

						num_times++;

					} else { /* !REALTIME_NOISE */
						chan.cur_pos++;

						if (chan.short_sample)
						{
							if (APU_NOISE_93 == chan.cur_pos)
								chan.cur_pos = 0;
						}
						else
						{
							if (APU_NOISE_32K == chan.cur_pos)
								chan.cur_pos = 0;
						}

						if (chan.short_sample)
							noise_bit = noise_short_lut[chan.cur_pos];
						else
							noise_bit = noise_long_lut[chan.cur_pos];

						if (noise_bit != 0)
							total += outvol;
						else
							total -= outvol;

						num_times++;
					}
				}

				chan.output_vol = total / num_times;

				return APU_NOISE_OUTPUT(chan);
	}


	/* DELTA MODULATION CHANNEL
	** =========================
	** reg0: 7=irq gen, 6=looping, 3-0=pointer to clock table
	** reg1: output dc level, 6 bits unsigned
	** reg2: 8 bits of 64-byte aligned address offset : $C000 + (value * 64)
	** reg3: length, (value * 16) + 1
	*/
	private int apu_dmc(DmcCh chan)
	{
		int delta_bit;

		/* APU_VOLUME_DECAY */
		chan.output_vol -= (chan.output_vol >> 7);

		/* only process when channel is alive */
		if (chan.dma_length != 0)
		{
			chan.phaseacc -= apu_cycle_rate; /* # of cycles per sample */

			while (chan.phaseacc < 0)
			{
				chan.phaseacc += chan.freq;

				delta_bit = (chan.dma_length & 7) ^ 7;

				if (7 == delta_bit)
				{
					chan.cur_byte = nes.nes6502_getbyte(chan.address);

					// TODO:steal a cycle from CPU
					// nes.nes6502_setdma(1);

					if (0xFFFF == chan.address)
						chan.address = 0x8000;
					else
						chan.address++;
				}

				if (--chan.dma_length == 0)
				{
					/* if loop bit set, we're cool to retrigger sample */
					if (chan.looping)
					{
						/* apu_dmcreload */
						chan.address = chan.cached_addr;
						chan.dma_length = chan.cached_dmalength;
						chan.irq_occurred = false;
					}
					else
					{
						/* check to see if we should generate an irq */
						if (chan.irq_gen)
						{
							chan.irq_occurred = true;
//							#ifndef NSF_PLAYER
////							DCR                  nes_irq();
//							#endif /* !NSF_PLAYER */
						}

						/* bodge for timestamp queue */
						chan.enabled = false;
						break;
					}
				}

				/* positive delta */
				if ((chan.cur_byte & (1 << delta_bit)) != 0)
				{
					if (chan.regs[1] < 0x7D)
					{
						chan.regs[1] += 2;
						chan.output_vol += (2 << 8);
					}
				}
				/* negative delta */
				else            
				{
					if (chan.regs[1] > 1)
					{
						chan.regs[1] -= 2;
						chan.output_vol -= (2 << 8);
					}
				}
			}
		}

		return APU_DMC_OUTPUT(chan);
	}

	private void apu_regwrite(int address, int value) {  
	   int chan;

	   switch (address)
	   {
	   /* rectangles */
	   case APU_WRA0:
	   case APU_WRB0:
	      chan = (address & 4) != 0 ? 1 : 0;
	      apu_rectangle[chan].regs[0] = value;

	      apu_rectangle[chan].volume = value & 0x0F;
	      apu_rectangle[chan].env_delay = decay_lut[value & 0x0F];
	      apu_rectangle[chan].holdnote = (value & 0x20) != 0 ? true : false;
	      apu_rectangle[chan].fixed_envelope = (value & 0x10) != 0 ? true : false;
	      apu_rectangle[chan].duty_flip = duty_lut[value >> 6];
	      break;

	   case APU_WRA1:
	   case APU_WRB1:
	      chan = (address & 4) != 0 ? 1 : 0;
	      apu_rectangle[chan].regs[1] = value;
	      apu_rectangle[chan].sweep_on = (value & 0x80) != 0 ? true : false;
	      apu_rectangle[chan].sweep_shifts = value & 7;
	      apu_rectangle[chan].sweep_delay = decay_lut[(value >> 4) & 7];
	      
	      apu_rectangle[chan].sweep_inc = (value & 0x08) != 0 ? true : false;
	      apu_rectangle[chan].freq_limit = freq_limit[value & 7];
	      break;

	   case APU_WRA2:
	   case APU_WRB2:
	      chan = (address & 4) != 0 ? 1 : 0;
	      apu_rectangle[chan].regs[2] = value;
	      apu_rectangle[chan].freq = (apu_rectangle[chan].freq & ~0xFF) | value;
	      break;

	   case APU_WRA3:
	   case APU_WRB3:
	      chan = (address & 4) != 0 ? 1 : 0;
	      apu_rectangle[chan].regs[3] = value;

	      apu_rectangle[chan].vbl_length = vbl_lut[value >> 3];
	      apu_rectangle[chan].env_vol = 0;
	      apu_rectangle[chan].freq = ((value & 7) << 8) | (apu_rectangle[chan].freq & 0xFF);
	      apu_rectangle[chan].adder = 0;
	      break;

	   /* triangle */
	   case APU_WRC0:
	      apu_triangle.regs[0] = value;
	      apu_triangle.holdnote = (value & 0x80) != 0 ? true : false;

	      if (false == apu_triangle.counter_started && apu_triangle.vbl_length != 0)
	         apu_triangle.linear_length = trilength_lut[value & 0x7F];

	      break;

	   case APU_WRC2:

	      apu_triangle.regs[1] = value;
	      apu_triangle.freq = APU_TO_FIXED((((apu_triangle.regs[2] & 7) << 8) + value) + 1);
	      break;

	   case APU_WRC3:

	      apu_triangle.regs[2] = value;
	  
	      /* this is somewhat of a hack.  there appears to be some latency on 
	      ** the Real Thing between when trireg0 is written to and when the 
	      ** linear length counter actually begins its countdown.  we want to 
	      ** prevent the case where the program writes to the freq regs first, 
	      ** then to reg 0, and the counter accidentally starts running because 
	      ** of the sound queue's timestamp processing.
	      **
	      ** set latency to a couple hundred cycles -- should be plenty of time 
	      ** for the 6502 code to do a couple of table dereferences and load up 
	      ** the other triregs
	      */

	      /* 06/13/00 MPC -- seems to work OK */
	      apu_triangle.write_latency = 228 / APU_FROM_FIXED(apu_cycle_rate);

	      apu_triangle.freq = APU_TO_FIXED((((value & 7) << 8) + apu_triangle.regs[1]) + 1);
	      apu_triangle.vbl_length = vbl_lut[value >> 3];
	      apu_triangle.counter_started = false;
	      apu_triangle.linear_length = trilength_lut[apu_triangle.regs[0] & 0x7F];

	      break;

	   /* noise */
	   case APU_WRD0:
	      apu_noise.regs[0] = value;
	      apu_noise.env_delay = decay_lut[value & 0x0F];
	      apu_noise.holdnote = (value & 0x20) != 0 ? true : false;
	      apu_noise.fixed_envelope = (value & 0x10) != 0 ? true : false;
	      apu_noise.volume = value & 0x0F;
	      break;

	   case APU_WRD2:
		   apu_noise.regs[1] = value;
		   apu_noise.freq = APU_TO_FIXED(noise_freq[value & 0x0F]);

		   if(REALTIME_NOISE) {
			   apu_noise.xor_tap = (value & 0x80) !=0 ? 0x40: 0x02;
		   }
		   else
		   {
			   /* detect transition from long.short sample */
			   if ((value & 0x80) != 0 && false == apu_noise.short_sample)
			   {
				   /* recalculate short noise buffer */
				   shift_register15(noise_short_lut, APU_NOISE_93);
				   apu_noise.cur_pos = 0;
			   }
			   apu_noise.short_sample = (value & 0x80) != 0 ? true : false;
		   }
		   break;

	   case APU_WRD3:
	      apu_noise.regs[2] = value;

	      apu_noise.vbl_length = vbl_lut[value >> 3];
	      apu_noise.env_vol = 0; /* reset envelope */
	      break;

	   /* DMC */
	   case APU_WRE0:
	      apu_dmc.regs[0] = value;

	      apu_dmc.freq = APU_TO_FIXED(dmc_clocks[value & 0x0F]);
	      apu_dmc.looping = (value & 0x40) != 0 ? true : false;

	      if ( (value & 0x80) != 0 )
	         apu_dmc.irq_gen = true;
	      else
	      {
	         apu_dmc.irq_gen = false;
	         apu_dmc.irq_occurred = false;
	      }
	      break;

	   case APU_WRE1: /* 7-bit DAC */
	      /* add the _delta_ between written value and
	      ** current output level of the volume reg
	      */
	      value &= 0x7F; /* bit 7 ignored */
	      apu_dmc.output_vol += ((value - apu_dmc.regs[1]) << 8);
	      apu_dmc.regs[1] = value;
	      break;

	   case APU_WRE2:
	      apu_dmc.regs[2] = value;
	      apu_dmc.cached_addr = 0xC000 + (value << 6);
	      break;

	   case APU_WRE3:
	      apu_dmc.regs[3] = value;
	      apu_dmc.cached_dmalength = ((value << 4) + 1) << 3;
	      break;

	   case APU_SMASK:
	      /* bodge for timestamp queue */
	      apu_dmc.enabled = (value & 0x10) != 0 ? true : false;

	      apu_enable_reg = value;

	      for (chan = 0; chan < 2; chan++)
	      {
	         if ( (value & (1 << chan)) != 0 )
	            apu_rectangle[chan].enabled = true;
	         else
	         {
	            apu_rectangle[chan].enabled = false;
	            apu_rectangle[chan].vbl_length = 0;
	         }
	      }

	      if ( (value & 0x04) != 0 )
	    	  apu_triangle.enabled = true;
	      else
	      {
	    	  apu_triangle.enabled = false;
	    	  apu_triangle.vbl_length = 0;
	    	  apu_triangle.linear_length = 0;
	    	  apu_triangle.counter_started = false;
	    	  apu_triangle.write_latency = 0;
	      }

	      if ( (value & 0x08) != 0 )
	    	  apu_noise.enabled = true;
	      else
	      {
	    	  apu_noise.enabled = false;
	    	  apu_noise.vbl_length = 0;
	      }

	      if ( (value & 0x10) != 0 )
	      {
	    	  if (0 == apu_dmc.dma_length)
	    		  /* apu_dmcreload */
	    		  apu_dmc.address = apu_dmc.cached_addr;
	    	  apu_dmc.dma_length = apu_dmc.cached_dmalength;
	    	  apu_dmc.irq_occurred = false;
	      }
	      else
	      {
	    	  apu_dmc.dma_length = 0;
	    	  apu_dmc.irq_occurred = false;
	      }
	      break;

	      /* unused, but they get hit in some mem-clear loops */
	   case 0x4009:
	   case 0x400D:
		   break;

	   default:
		   break;
	   }
	}

	/* Read from $4000-$4017 */
	public int apu_read(int address) {
	   if(!apu_enable) return 0;
	   
	   int value;

	   switch (address)
	   {
	   case APU_SMASK:
	      value = 0;
	      /* Return 1 in 0-5 bit pos if a channel is playing */
	      if (apu_rectangle[0].enabled_cur && apu_rectangle[0].vbl_length_cur != 0 )
	         value |= 0x01;
	      if (apu_rectangle[1].enabled_cur && apu_rectangle[1].vbl_length_cur != 0 )
	         value |= 0x02;
	      if (apu_triangle.enabled_cur && apu_triangle.vbl_length_cur != 0 )
	         value |= 0x04;
	      if (apu_noise.enabled_cur && apu_noise.vbl_length_cur != 0 )
	         value |= 0x08;
	      /* bodge for timestamp queue */
	      if (apu_dmc.enabled_cur)
	         value |= 0x10;
	      if (apu_dmc.irq_occurred_cur)
	         value |= 0x80;
	      break;

	   default:
	      value = (address >> 8); /* heavy capacitance on data bus */
	      break;
	   }

	   return value;
	}

	int apu_ex_read(int address) {
	   if(apu_ex_chip == FDS) {
	      return ex.Read (address);
	   }
	   else if(apu_ex_chip == N106) {
		  //return N106SoundReadData (address);
	      ApuData d = new ApuData();
	      d.timestamp = nes.nes6502_getcycles();
	      d.address = address + 0x10000;
	      apu_ex_enqueue(d);
	      return 0x00;
	   }
	   else {
	      return 0x00;
	   }
	}

	public void apu_write(int address, int value) {
		if(!apu_enable) return;
		ApuData d = new ApuData();

		switch (address)
		{
		case 0x4015:
			/* bodge for timestamp queue */
			apu_dmc.enabled = (value & 0x10) != 0 ? true : false;

		case 0x4000: case 0x4001: case 0x4002: case 0x4003:
		case 0x4004: case 0x4005: case 0x4006: case 0x4007:
		case 0x4008: case 0x4009: case 0x400A: case 0x400B:
		case 0x400C: case 0x400D: case 0x400E: case 0x400F:
		case 0x4010: case 0x4011: case 0x4012: case 0x4013:
			d.timestamp = nes.nes6502_getcycles();
			d.address = address;
			d.value = value;
			apu_enqueue(d);
			break;

		default:
			break;
		}
		
		apu_write_cur(address, value);
	}

	private void apu_write_cur(int address, int value) {
		/* for sync read $4015 */
		int chan;
		switch (address)
		{
		case APU_WRA0:
		case APU_WRB0:
			chan = (address & 4) != 0 ? 1 : 0;
			apu_rectangle[chan].holdnote_cur = (value & 0x20) != 0 ? true : false;
			break;
		case APU_WRA3:
		case APU_WRB3:
			chan = (address & 4) != 0 ? 1 : 0;
			apu_rectangle[chan].vbl_length_cur = vbl_length[value >> 3];
			break;
		case APU_WRC0:
			apu_triangle.holdnote_cur = (value & 0x80) != 0 ? true : false;
			break;
		case APU_WRC3:
			apu_triangle.vbl_length_cur = vbl_length[value >> 3];
			apu_triangle.counter_started_cur = true;
			break;
		case APU_WRD0:
			apu_noise.holdnote_cur = (value & 0x20) != 0 ? true : false;
			break;
		case APU_WRD3:
			apu_noise.vbl_length_cur = vbl_length[value >> 3];
			break;
		case APU_WRE0:
			apu_dmc.freq_cur = dmc_clocks[value & 0x0F];
			apu_dmc.phaseacc_cur = 0;
			apu_dmc.looping_cur = (value & 0x40) != 0 ? true : false;
			if ( (value & 0x80) != 0 )
				apu_dmc.irq_gen_cur = true;
			else
			{
				apu_dmc.irq_gen_cur = false;
				apu_dmc.irq_occurred_cur = false;
			}
			break;
		case APU_WRE3:
			apu_dmc.cached_dmalength_cur = (value << 4) + 1;
			break;
		case APU_SMASK:
			for (chan = 0; chan < 2; chan++)
			{
				if ( (value & (1 << chan)) != 0 )
					apu_rectangle[chan].enabled_cur = true;
				else
				{
					apu_rectangle[chan].enabled_cur = false;
					apu_rectangle[chan].vbl_length_cur = 0;
				}
			}
			if ( (value & 0x04) != 0 )
				apu_triangle.enabled_cur = true;
			else
			{
				apu_triangle.enabled_cur = false;
				apu_triangle.vbl_length_cur = 0;
				apu_triangle.counter_started_cur = false;
			}
			if ( (value & 0x08) != 0 )
				apu_noise.enabled_cur = true;
			else
			{
				apu_noise.enabled_cur = false;
				apu_noise.vbl_length_cur = 0;
			}
			if ( (value & 0x10) != 0 )
			{
				if(0 == apu_dmc.dma_length_cur)
				{
					apu_dmc.dma_length_cur = apu_dmc.cached_dmalength_cur;
				}
				apu_dmc.enabled_cur = true;
			}
			else
			{
				apu_dmc.dma_length_cur = 0;
				apu_dmc.enabled_cur = false;
				apu_dmc.irq_occurred_cur = false;
			}
			break;
		}
	}

	void apu_ex_write(int address, int value) {
	   ApuData d = new ApuData();
	   
	   d.timestamp = nes.nes6502_getcycles();
	   d.address = address;
	   d.value = value;
	   apu_ex_enqueue(d);

	   if(apu_ex_chip == FDS)
	   {
	      ex.WriteCurrent(address, value);
	   }
	}

	private void apu_process(PcmAudioHandler pcm, int num_samples) {
	   ApuData d;
	   int elapsed_cycles;
//	   static int prev_sample = 0;
	   int next_sample, accum;

	   if (null == pcm)
	   {
	      /* grab it, keep it local for speed */ //DCR
	      elapsed_cycles = apu_elapsed_cycles;
	      /* just go through the motions... */
	      while (0 != num_samples--)
	      {
	         while (apu_q_head != apu_q_tail && apu_queue[apu_q_tail].timestamp <= elapsed_cycles)
	         {
	            d = apu_dequeue();
	            apu_regwrite(d.address, d.value);
	         }
	         
	         //TODO:
	         while(apu_ex_q_head != apu_ex_q_tail && apu_ex_queue[apu_ex_q_tail].timestamp <= elapsed_cycles)
	         {
	        	 d = apu_ex_dequeue();

	        	 if(ex == null) { 
	        		 continue;
	        	 }
	        	 else if(apu_ex_chip == N106) {
	        		 if(0 != (d.address & 0x10000)) {
	        			 int dummy = ex.Read(d.address & 0xFFFF);
	        		 }
	        		 else {
	        			 ex.Write(d.address, d.value);
	        		 }
	        	 }
	        	 else {
	        		 ex.Write(d.address, d.value);
	        	 }
	         }

	         elapsed_cycles += APU_FROM_FIXED(apu_cycle_rate);
	      }
	   }
	   else
	   {
	      /* grab it, keep it local for speed */
	      elapsed_cycles = apu_elapsed_cycles;

	      while (0 != num_samples--)
	      {
	         while (apu_q_head != apu_q_tail && (apu_queue[apu_q_tail].timestamp <= elapsed_cycles))
	         {
	            d = apu_dequeue();
	            apu_regwrite(d.address, d.value);
	         }

	         while(apu_ex_q_head != apu_ex_q_tail && apu_ex_queue[apu_ex_q_tail].timestamp <= elapsed_cycles)
	         {
	        	 d = apu_ex_dequeue();

	        	 if(ex == null) { 
	        		 continue;
	        	 }
	        	 else if(apu_ex_chip == N106) {
	        		 if(0 != (d.address & 0x10000)) {
	        			 int dummy = ex.Read(d.address & 0xFFFF);
	        		 }
	        		 else {
	        			 ex.Write(d.address, d.value);
	        		 }
	        	 }
	        	 else {
	        		 ex.Write(d.address, d.value);
	        	 }
	         }

	         elapsed_cycles += APU_FROM_FIXED(apu_cycle_rate);

	         accum = 0;
	         if (apu_mix_enable[0]) accum += apu_rectangle(apu_rectangle[0]);
	         if (apu_mix_enable[1]) accum += apu_rectangle(apu_rectangle[1]);
	         if (apu_mix_enable[2]) accum += apu_triangle(apu_triangle);
	         if (apu_mix_enable[3]) accum += apu_noise(apu_noise);
	         if (apu_mix_enable[4]) accum += apu_dmc(apu_dmc);

	         if (apu_mix_enable[5])
	         {
	            if (ex != null) {
	            	accum += ex.Render();
	            }
	         }


	         /* do any filtering */
	         if (APU_FILTER_NONE != apu_filter_type)
	         {
	            next_sample = accum;

	            if (APU_FILTER_LOWPASS == apu_filter_type)
	            {
	               accum += prev_sample;
	               accum >>= 1;
	            }
	            else {
	               accum = (accum + accum + accum + prev_sample) >> 2;
	            }

	            prev_sample = next_sample;
	         }

	         /* little extra kick for the kids */
	         accum <<= 1;

	         /* prevent clipping */
	         if (accum > 0x7FFF)
	            accum = 0x7FFF;
	         else if (accum < -0x8000)
	            accum = -0x8000;

	         //TODO:
	         /* signed 16-bit output */
	         pcm.writeSample(accum);
	      }
	   }

	   /* resync cycle counter */
	   apu_elapsed_cycles = nes.nes6502_getcycles();
	}
	
	private void apu_build_luts(int num_samples) {
	   /* lut used for enveloping and frequency sweeps */
	   for (int i = 0; i < 16; i++)
	      decay_lut[i] = num_samples * (i + 1);

	   /* used for note length, based on vblanks and size of audio buffer */
	   for (int i = 0; i < 32; i++)
	      vbl_lut[i] = vbl_length[i] * num_samples;

	   /* triangle wave channel's linear length table */
	   for (int i = 0; i < 128; i++)
	      trilength_lut[i] = (int) (0.25 * (i * num_samples));

	   /* generate noise samples */
	   shift_register15(noise_long_lut, APU_NOISE_32K);
	   shift_register15(noise_short_lut, APU_NOISE_93);
	}

	private void sync_apu_register() {
		if (!apu_rectangle[0].holdnote_cur && apu_rectangle[0].vbl_length_cur != 0)
		{
			apu_rectangle[0].vbl_length_cur--;
		}
		if (!apu_rectangle[1].holdnote_cur && apu_rectangle[1].vbl_length_cur != 0)
		{
			apu_rectangle[1].vbl_length_cur--;
		}
		if (apu_triangle.counter_started_cur)
		{
			if (apu_triangle.vbl_length_cur != 0 && !apu_triangle.holdnote_cur)
			{
				apu_triangle.vbl_length_cur--;
			}
		}
		if (!apu_noise.holdnote_cur && apu_noise.vbl_length_cur != 0)
		{
			apu_noise.vbl_length_cur--;
		}
	}
	
	//TODO
	private boolean sync_dmc_register(long cpu_cycles) {
	   boolean irq_occurred = false;

	   apu_dmc.phaseacc_cur -= (int)cpu_cycles;
	   while(apu_dmc.phaseacc_cur < 0)
	   {
	      apu_dmc.phaseacc_cur += apu_dmc.freq_cur * 8;
	      if (apu_dmc.dma_length_cur != 0)
	      {
	        if (--apu_dmc.dma_length_cur == 0)
	        {
	           if (apu_dmc.looping_cur)
	           {
	              apu_dmc.dma_length_cur = apu_dmc.cached_dmalength_cur;
	              apu_dmc.irq_occurred_cur = false;
	           }
	           else
	           {
	              apu_dmc.dma_length_cur = 0;
	              if (apu_dmc.irq_gen_cur)
	              {
	                 apu_dmc.irq_occurred_cur = true;
	                 irq_occurred = true;
	              }
	              apu_dmc.enabled_cur = false;
	           }
	        }
	      }
	   }
	   return irq_occurred;
	}
	
	public void apu_reset() {
	   apu_elapsed_cycles = 0;
	   
	   for(int i=0; i < apu_queue.length; i++) apu_queue[i] = null;
	   apu_q_head = apu_q_tail = 0;

	   for(int i=0; i < apu_queue.length; i++) apu_queue[i] = null;
	   apu_ex_q_head = apu_ex_q_tail = 0;
	   
	   //apu_ex_chip = 0;

	   /* use to avoid bugs =) */
	   for (int address = 0x4000; address <= 0x4013; address++)
	   {
	      apu_regwrite(address, 0);
	      apu_write_cur(address, 0);
	   }

	   apu_regwrite(0x4015, 0x00);
	   apu_write_cur(0x4015, 0x00);
	   
	   sreg1 = 0x4000;
	   sreg2 = 0x4000;

	   // for ExSound
	   if(ex != null) {
		   ex.Reset();
		   ex.Volume(1);
	   }
	}
	
	void apu_select_exsound(int mapper_num) {
		//TODO:
		switch(mapper_num) {
		case 24 :
		case 26: 
			apu_ex_chip = VRC6;
			break;
		case 85: 
			apu_ex_chip = VRC7;
			break;
		case 20: 
			apu_ex_chip = FDS;
			ex = new FdsSound();
			break;
		case 5: 
			apu_ex_chip = MMC5;
			break;
		case 19: 
			apu_ex_chip = N106;
			break;
		case 69: 
			apu_ex_chip = FME7;
			break;
		default: 
			apu_ex_chip = 0;
		}
	}
	
    private int apu_frame_count;
    private int sound_buf_frame_count;

	public void apu_emulate_frame() {
		if(!apu_enable) return;
		//TODO:how many samples required for ?
		apu_process(this.pcm, apu_num_samples);
		
		sync_apu_register();
        if((++apu_frame_count % this.sound_buf_frame_count) == 0) pcm.play();
	}
		
	public void apu_setparams(int sample_rate, int refresh_rate, int sample_bits)
	{
		apu_sample_rate = sample_rate;
		apu_refresh_rate = refresh_rate;
		apu_sample_bits = sample_bits;

		apu_num_samples = sample_rate / refresh_rate;

		/* turn into fixed point! */
		apu_cycle_rate = (int)(APU_BASEFREQ * 65536.0 / (double) sample_rate);

		/* build various lookup tables for apu */
		apu_build_luts(apu_num_samples);
	}

	public void apu_setchan(int chan, boolean enabled) {
		if(chan < 0 || chan >= apu_mix_enable.length) return;
		apu_mix_enable[chan] = enabled;
	}
	
	public void apu_setfilter(int type) {
		/* set the filter type */
		apu_filter_type = type;
	}
	
	private void apu_init() {
		/* APUSOUND */
		apu_rectangle = new RectangleCh[2];
		apu_rectangle[0] = new RectangleCh();
		apu_rectangle[1] = new RectangleCh();
		apu_triangle = new TriangleCh();
		apu_noise = new NoiseCh();
		apu_dmc = new DmcCh();

		/* set the stupid flag to tell difference between two rectangles */
		apu_rectangle[0].sweep_complement = true;
		apu_rectangle[1].sweep_complement = false;

		apu_mix_enable = new boolean[6];
		apu_queue = new ApuData[APUQUEUE_SIZE];
		apu_ex_queue = new ApuData[APUQUEUE_SIZE];

		for (int channel = 0; channel < 6; channel++) apu_setchan(channel, true);
		apu_setparams(8000, 60, 16);
		
    	sound_buf_frame_count = 15;
        apu_frame_count = 0;
        
		apu_reset();
	}
	
	private PcmAudioHandler pcm;
	public void apu_setmode(int mode) {
		if (mode >= 3) return;
		pcm = PcmAudioHandler.getPCMHandler(mode);
	}
	public void apu_setvol(int vol) {
		pcm.setVolume(vol);
	}
	public void apu_setrate(int rate) {
		pcm.setSampleRate(rate);
	}
	public void apu_enable(boolean enable) {apu_enable = enable;}
	
	public void apu_setting() {
		pcm.setVolume(gui.soundVolume);
		apu_enable = (gui.soundEnable!=0);
       if(apu_mix_enable.length > 0) apu_mix_enable[0] = !gui.pcm_channelOff_[0];
       if(apu_mix_enable.length > 1) apu_mix_enable[1] = !gui.pcm_channelOff_[1];
       if(apu_mix_enable.length > 2) apu_mix_enable[2] = !gui.pcm_channelOff_[2];
       if(apu_mix_enable.length > 3) apu_mix_enable[3] = !gui.pcm_channelOff_[3];
       if(apu_mix_enable.length > 4) apu_mix_enable[4] = !gui.pcm_channelOff_[4];
       if(apu_mix_enable.length > 5) apu_mix_enable[5] = !gui.pcm_channelOff_[5];
        pcm.setSampleRate(gui.SOUND_RATE_ARRAY[gui.soundSampleRate]);
        apu_filter_type = gui.soundApuFilter;
        apu_setparams(gui.SOUND_RATE_ARRAY[gui.soundSampleRate], 60, 16);
    	sound_buf_frame_count = gui.soundBufCount;
        apu_frame_count = 0;
	}	
	
	private final NESCore nes;
	private final NESCanvas gui;
	
	NESApu2(NESCore nes_, NESCanvas gui_, PSGi psgi_) {
		nes = nes_;
		gui = gui_;
		apu_init();
	}
}
