
public class NESApu_NP2 implements NESApu {
	private static final int CH_PWG1 = 0;
	private static final int CH_PWG2 = 1;
	private static final int CH_TWG = 2;
	private static final int CH_NWG = 3;

	private static final int BITFLAG_SET = 1;
	private static final int BITFLAG_CLEAR = 0;
	
	private static final int PWG_BASEFREQ = 111875;
	private static final int TWG_BASEFREQ = 55938;
	
	private static final int PWG_BANKNUM[] = { 0,1,2,1 };
	private static final int TWG_BANKNUM = 3;
	private static final int NWG_BANKNUM[] = { 4, 4 };
	private static final int PWG_VOL[] = { 0,7,10,14,19,22,25,28,32,34,36,38,40,42,44,46 };
	private static final int TWG_VOL[] = { 0,40 };
	private static final int NWG_VOL[] = { 0,5,7,9,11,13,15,17,19,21,23,25,27,29,31,33 };
	private static final int NWG_NOTENUMTABLE[] = { 
		3,8,12,17,20,24,29,34,
		34,34,34,34,34,34,34,34
	}; 

	/*JE^p̃JE^le[uBJE^120hz쓮Aȉ̓t[ŕ\L*/
	private static final int[] lengthCounterValueTable = {
//			0x0A,0xFE,0x14,0x02,
//			0x28,0x04,0x50,0x06,
//			0xA0,0x08,0x3C,0x0A,
//			0x0E,0x0C,0x1A,0x0E,
//			0x0C,0x10,0x18,0x12,
//			0x30,0x14,0x60,0x16,
//			0xC0,0x18,0x48,0x1A,
//			0x10,0x1C,0x20,0x1E
			0x05,0x7F,0x0A,0x01,
			0x14,0x02,0x28,0x03,
			0x50,0x04,0x1E,0x05,
			0x07,0x06,0x0E,0x07,
			0x06,0x08,0x0C,0x09,
			0x18,0x0A,0x30,0x0B,
			0x60,0x0C,0x24,0x0D,
			0x08,0x0E,0x10,0x0F,
	};

	/*WX^p[^*/
	/*PWG1 $4000 - $4003 */
	private int pwg1_dutyCycle;
	private int pwg1_disableLengthCounter;
	private int pwg1_disableEnvelope;
	private int pwg1_envelopeCycle;
	private int pwg1_enableSweep;
	private int pwg1_sweepCycle;
	private int pwg1_sweepModeNegative;
	private int pwg1_sweepShiftval;
	private int pwg1_deviderClock;
	private int pwg1_lengthCounterIndex;
	/*PWG2 $4004 - $4007 */
	private int pwg2_dutyCycle;
	private int pwg2_disableLengthCounter;
	private int pwg2_disableEnvelope;
	private int pwg2_envelopeCycle;
	private int pwg2_enableSweep;
	private int pwg2_sweepCycle;
	private int pwg2_sweepModeNegative;
	private int pwg2_sweepShiftval;
	private int pwg2_deviderClock;
	private int pwg2_lengthCounterIndex;
	/*TWG $4008 - $400B */
	private int twg_disableLengthCounter;
	private int twg_valueOfLinearCounter;	
	private int twg_deviderClock;
	private int twg_lengthCounterIndex;
	/*NWG $400C - $400F */
	private int nwg_disableLengthCounter;
	private int nwg_disableEnvelope;
	private int nwg_envelopeCycle;
	private int nwg_generationMode;
	private int nwg_deviderClockIndex;
	private int nwg_lengthCounterIndex;
	/*DMC $4010 - $4013 */
	/*STATUS	$4015	*/
	private int pwg1_enable;
	private int pwg2_enable;
	private int twg_enable;
	private int nwg_enable;
	/*SEQ,IRQ flags$4017*/

	/*G~[gp̓p[^*/
	private int pwg1_lengthCounter;
	private int pwg1_envelopeTimer;
	private int pwg1_envelopeCounter;
	private int pwg1_sweepTimer;
	private int pwg2_lengthCounter;
	private int pwg2_envelopeTimer;
	private int pwg2_envelopeCounter;
	private int pwg2_sweepTimer;
	private int twg_lengthCounter;	
	private int twg_linearCounter;
	private int twg_disableLinearCounter;
	private int nwg_lengthCounter;
	private int nwg_envelopeTimer;
	private int nwg_envelopeCounter;
	private int pwg1_dacOutput;
	private int pwg2_dacOutput;
	private int twg_dacOutput;
	private int nwg_dacOutput;
	
	private PSGi psgi;
	
	NESApu_NP2(NESCore nes_, NESCanvas gui_, PSGi psgi_) {
		psgi = psgi_;
		psgi.setChTone(CH_TWG, TWG_BANKNUM);
		psgi.setChTone(CH_NWG, NWG_BANKNUM[nwg_deviderClockIndex]);
		psgi.setChNoiseMode(CH_NWG, true);
	}
	
	void doFrame() {
	//60t[ŃTEhG~[V
		
		//XEB[p
		if(pwg1_enableSweep == BITFLAG_SET && pwg1_sweepShiftval > 0 && pwg1_lengthCounter > 0 ) {
			pwg1_sweepTimer += 2;
			while(pwg1_sweepTimer / (pwg1_sweepCycle+1) > 0) {
				if(pwg1_sweepModeNegative == BITFLAG_SET) {
					pwg1_deviderClock -= pwg1_deviderClock >> pwg1_sweepShiftval; 
				}
				else {
					pwg1_deviderClock += pwg1_deviderClock >> pwg1_sweepShiftval; 
				}

				pwg1_sweepTimer -= (pwg1_sweepCycle+1);

				if(pwg1_deviderClock <= 8) {
					pwg1_deviderClock = 0x08;
					pwg1_lengthCounter = 0;
				} 
				if(pwg1_deviderClock >= 0x7FF) {
					pwg1_deviderClock = 0x7FF;
					pwg1_lengthCounter = 0;
				}
			}
		}

		if(pwg2_enableSweep == BITFLAG_SET && pwg2_sweepShiftval > 0 && pwg2_lengthCounter > 0 ) {
			pwg2_sweepTimer += 2;
			while(pwg2_sweepTimer / (pwg2_sweepCycle+1) > 0) {
				if(pwg2_sweepModeNegative == BITFLAG_SET) {
					pwg2_deviderClock -= (pwg2_deviderClock >> pwg2_sweepShiftval) - 1; 
				}
				else {
					pwg2_deviderClock += pwg2_deviderClock >> pwg2_sweepShiftval; 
				}
				
				pwg2_sweepTimer -= (pwg2_sweepCycle+1);
				
				if(pwg2_deviderClock <= 8) {
					pwg2_deviderClock = 0x08;
					pwg2_lengthCounter = 0;
				}
				if(pwg2_deviderClock >= 0x7FF) {
					pwg2_deviderClock = 0x7FF;
					pwg2_lengthCounter = 0;
				}
			}
		}

		//16bitDACo͌vZ
		/*`g1*/
		if(pwg1_disableLengthCounter == BITFLAG_CLEAR && pwg1_lengthCounter > 0) pwg1_lengthCounter--;

		if(pwg1_lengthCounter == 0) {
			pwg1_dacOutput = 0;
		} else {
			if(pwg1_disableEnvelope == BITFLAG_SET) {
				pwg1_dacOutput = pwg1_envelopeCycle;
			}
			else {
				pwg1_envelopeTimer += 4;
				while(pwg1_envelopeTimer / (pwg1_envelopeCycle+1) > 0) {
					pwg1_envelopeCounter--;

					if(pwg1_envelopeCounter < 0) {
						if(pwg1_disableLengthCounter == BITFLAG_SET) {
							pwg1_envelopeCounter = 15;
						}
						else {
							pwg1_envelopeCounter = 0;
						}
					}

					pwg1_envelopeTimer -= pwg1_envelopeCycle + 1;
				}
				pwg1_dacOutput = pwg1_envelopeCounter;
			}
		}

		/*`g2*/
		if(pwg2_disableLengthCounter == BITFLAG_CLEAR && pwg2_lengthCounter > 0) pwg2_lengthCounter--;

		if(pwg2_lengthCounter == 0) {
			pwg2_dacOutput = 0;
		} else {
			if(pwg2_disableEnvelope == BITFLAG_SET) {
				pwg2_dacOutput = pwg2_envelopeCycle;
			}
			else {
				pwg2_envelopeTimer += 4;
				while(pwg2_envelopeTimer / (pwg2_envelopeCycle+1) > 0) {
					pwg2_envelopeCounter--;

					if(pwg2_envelopeCounter < 0) {
						if(pwg2_disableLengthCounter == BITFLAG_SET) {
							pwg2_envelopeCounter = 15;
						}
						else {
							pwg2_envelopeCounter = 0;
						}
					}

					pwg2_envelopeTimer -= pwg2_envelopeCycle + 1;
				}
				pwg2_dacOutput = pwg2_envelopeCounter;
			}
		}

		/*Opg*/
		if(twg_disableLinearCounter == BITFLAG_SET) {
			twg_linearCounter = (twg_valueOfLinearCounter+3)/4;
		}
		else if(twg_linearCounter > 0) {
			twg_linearCounter--;
		}

		if(twg_disableLengthCounter == BITFLAG_CLEAR) {
			if(twg_lengthCounter > 0) twg_lengthCounter--;
			twg_disableLinearCounter = BITFLAG_CLEAR;
		}

		if(twg_lengthCounter > 0 && twg_linearCounter > 0) {
			twg_dacOutput = 0x01;
		} else {
			twg_dacOutput = 0x00;
		}

		/*mCY*/
		if(nwg_disableLengthCounter == BITFLAG_CLEAR && nwg_lengthCounter > 0) nwg_lengthCounter--;

		if(nwg_lengthCounter == 0) {
			nwg_dacOutput = 0;
		} else {
			if(nwg_disableEnvelope == BITFLAG_SET) {
				nwg_dacOutput = nwg_envelopeCycle;
			}
			else {
				nwg_envelopeTimer += 4;
				while(nwg_envelopeTimer / (nwg_envelopeCycle+1) > 0) {
					nwg_envelopeCounter--;

					if(nwg_envelopeCounter < 0) {
						if(nwg_disableLengthCounter == BITFLAG_SET) {
							nwg_envelopeCounter = 15;
						}
						else {
							nwg_envelopeCounter = 0;
						}
					}

					nwg_envelopeTimer -= nwg_envelopeCycle + 1;
				}
				nwg_dacOutput = nwg_envelopeCounter;
			}
		}
		
		
		//MLDo
		
		//o͎g
		psgi.setChFreq(CH_PWG1, PWG_BASEFREQ / (pwg1_deviderClock + 1));
		psgi.setChFreq(CH_PWG2, PWG_BASEFREQ / (pwg2_deviderClock + 1));
		psgi.setChFreq(CH_TWG, TWG_BASEFREQ / (twg_deviderClock + 1));
		psgi.setChFreq(CH_NWG, NWG_NOTENUMTABLE[nwg_deviderClockIndex]);
		//V[PX[h
		psgi.setChTone(CH_PWG1, PWG_BANKNUM[pwg1_dutyCycle]);
		psgi.setChTone(CH_PWG2, PWG_BANKNUM[pwg2_dutyCycle]);
//		psgi.setChTone(CH_TWG, TWG_BANKNUM); //Opg͉F1B
//		psgi.setChTone(CH_NWG, NWG_BANKNUM[nwg_generationMode]); //mCYȗ		
		//
		psgi.setChVolume(CH_PWG1, PWG_VOL[pwg1_dacOutput]);
		psgi.setChVolume(CH_PWG2, PWG_VOL[pwg2_dacOutput]);
		psgi.setChVolume(CH_TWG, TWG_VOL[twg_dacOutput]);
		psgi.setChVolume(CH_NWG, NWG_VOL[nwg_dacOutput]);
		
		psgi.doFrame();
	}

	void writeReg(int addr, int data) {
		switch(addr) {
		
		//Rectangle1
		case 0x4000:
			pwg1_dutyCycle = data >> 6 & 0x03;
			pwg1_disableLengthCounter = data >> 5 & 0x01;
			pwg1_disableEnvelope = data >> 4 & 0x01;
			pwg1_envelopeCycle = data & 0x0F;
			break;
		case 0x4001:
			pwg1_enableSweep = data >> 7 & 0x01;
			pwg1_sweepCycle = data >> 4 & 0x07;
			pwg1_sweepModeNegative = data >> 3 & 0x01;
			pwg1_sweepShiftval = data & 0x07;
			break;
		case 0x4002:
			pwg1_deviderClock = pwg1_deviderClock & ~0xFF | data & 0xFF;
			break;
		case 0x4003:
			pwg1_lengthCounterIndex = data >> 3 & 0x1F;
			pwg1_deviderClock = (data & 0x07) << 8 | pwg1_deviderClock & 0xFF;
			
			pwg1_sweepTimer = 0;
			pwg1_envelopeTimer = 0;
			pwg1_envelopeCounter = 0x0F;

			if(pwg1_enable == BITFLAG_SET) {
				pwg1_lengthCounter = lengthCounterValueTable[pwg1_lengthCounterIndex];
			}

			break;
			
		//Rectangle2
		case 0x4004:
			pwg2_dutyCycle = data >> 6 & 0x03;
			pwg2_disableLengthCounter = data >> 5 & 0x01;
			pwg2_disableEnvelope = data >> 4 & 0x01;
			pwg2_envelopeCycle = data & 0x0F;
			break;
		case 0x4005:
			pwg2_enableSweep = data >> 7 & 0x01;
			pwg2_sweepCycle = data >> 4 & 0x07;
			pwg2_sweepModeNegative = data >> 3 & 0x01;
			pwg2_sweepShiftval = data & 0x07;
			break;
		case 0x4006:
			pwg2_deviderClock = pwg2_deviderClock & ~0xFF | data & 0xFF;
			break;
		case 0x4007:
			pwg2_lengthCounterIndex = data >> 3 & 0x1F;
			pwg2_deviderClock = (data & 0x07) << 8 | pwg2_deviderClock & 0xFF;
			
			pwg2_sweepTimer = 0;
			pwg2_envelopeTimer = 0;
			pwg2_envelopeCounter = 0x0F;

			if(pwg2_enable == BITFLAG_SET) {
				pwg2_lengthCounter = lengthCounterValueTable[pwg2_lengthCounterIndex];
			}

			break;
			
		//Triangle
		case 0x4008:
			twg_disableLengthCounter = data >> 7 & 0x01;
			twg_valueOfLinearCounter = data & 0x7F;
			break;
		case 0x400A:
			twg_deviderClock = twg_deviderClock & ~0xFF | data & 0xFF;
			break;
		case 0x400B:
			twg_lengthCounterIndex = data >> 3 & 0x1F;
			twg_deviderClock = (data & 0x07) << 8 | twg_deviderClock & 0xFF;
			
			twg_disableLinearCounter = BITFLAG_SET;
			
			if(twg_enable == BITFLAG_SET) {
				twg_lengthCounter = lengthCounterValueTable[twg_lengthCounterIndex];
			}
			break;
			
		//Noise
		case 0x400C:
			nwg_disableLengthCounter = data >> 5 & 0x01;
			nwg_disableEnvelope = data >> 4 & 0x01;
			nwg_envelopeCycle = data & 0x0F;
			break;
		case 0x400E:
			nwg_generationMode = data >> 7 & 0x01;
			nwg_deviderClockIndex = data & 0x0F;
			break;
		case 0x400F:
			nwg_lengthCounterIndex = data >> 3 & 0x1F;

			nwg_envelopeTimer = 0;
			nwg_envelopeCounter = 0x0F;

			if(nwg_enable == BITFLAG_SET) {
				nwg_lengthCounter = lengthCounterValueTable[nwg_lengthCounterIndex];
			}

			break;
			
		//status_register	
		case 0x4015:
			pwg1_enable = data & 0x01;
			pwg2_enable = data >> 1 & 0x01;
			twg_enable = data >> 2 & 0x01;
			nwg_enable = data >> 3 & 0x01;
			if(pwg1_enable == BITFLAG_CLEAR) pwg1_lengthCounter = 0;
			if(pwg2_enable == BITFLAG_CLEAR) pwg2_lengthCounter = 0;
			if(twg_enable == BITFLAG_CLEAR) twg_lengthCounter = 0;
			if(nwg_enable == BITFLAG_CLEAR) nwg_lengthCounter = 0;
			break;
			
//		case 0x4017:
//			disableFrameInterrupt = data >> 6 & 0x01;
//			frameSequencerMode = data >> 7 & 0x01;
//			return;
		}
	}
	
	int readReg() {
		int ret = 0;

		if(nwg_lengthCounter != 0) ret += 0x08;
		if(twg_lengthCounter != 0) ret += 0x04;
		if(pwg2_lengthCounter != 0) ret += 0x02;
		if(pwg1_lengthCounter != 0) ret += 0x01;
		return ret;
	}
	
	void init() {
		for(int i = 0x4000;i <= 0x4015; i++) writeReg(i, 0x00);
	}

	/* ----------------------------------------------------------- */

	public void apu_reset() {
		init();
	}
	public int apu_read(int address) {
		return readReg();
	}
	public void apu_write(int address, int value) {
		writeReg(address, value);
	}
	public void apu_emulate_frame() {
		doFrame();
	}
	public void apu_setting() {
		// XXX
	}
	public void apu_enable(boolean enable) {
		psgi.setSoundEnable(enable);
	}
	public void apu_setvol(int vol) {
		psgi.setMasterVolume(vol);
	}
	public void apu_setmode(int mode) {
		psgi.setMode(mode);
	}
}
