import java.io.*;

import com.nttdocomo.ui.Graphics;

abstract class NESCore extends Thread {

    public static final int REQUEST_START = 0;
    public static final int REQUEST_RESET = 1;
    public static final int REQUEST_STOP = 2;
    public static final int REQUEST_LOAD = 3;
    public static final int REQUEST_SAVE = 4;
    public static final int REQUEST_CHDISK = 5;
    public static final int REQUEST_SLEEP = 7;

    public NESCanvas gui;
    public NESApu audio;
    public PSGi psgi;

    public Mapper mapper;
    public int mapper_no;
    public boolean latch_mapper;
    public boolean exgfx_mapper;
    public int crc;
    public int[] crc_table;
    public int frame_irq_enabled;
    public int int_pending;
    public byte[] SRAM;
    public int SRAM_SIZE;
    public byte[][] ROM;
    public byte[][] DISK;
    public byte[][] VROM;
    public int[][] VROM_CACHE;
    public int[][] VRAM_CACHE;
    public byte[][] CACHE_HIT;
    public byte[] DUMMY_VROM;
    public int[] DUMMY_VROM_CACHE;
    public byte[] BIOS;
    public int num_16k_ROM_banks;
    public int num_8k_VROM_banks;
    public int num_DISK_banks;
    public int mirroring;
    public int request;
    public int param;
    public boolean disk_system;
    public boolean exist;
    public boolean sp_mode;

    public int[] fb;

    //NP2
    public boolean exclk_mapper;

	// =========================
	// 
	//==========================
    public void init() {
        RAM = new byte[0x0800];
        SRAM = new byte[0x2000];
        SRAM_SIZE = 0x2000;
        ROM = null;
        DISK = null;
        BIOS = null;
        cpu_map = new byte[8][];
        VRAM = new byte[12][0x0400];
        VRAM_CACHE = new int[8][512];
        CACHE_HIT = new byte[9][65];
        DUMMY_VROM = null;
        DUMMY_VROM_CACHE = null;
        ppu_map = new byte[12][];
        spr_ram = new int[256];
        pal_ram = new int[32];
        patpix = new int[8][];
        patdirty = new byte[8][];
        sprite_hit = new byte[16];
        fb = new int[256 * 256];
        disk_system = false;
        exist = false;
    }

    public void set_spmode(boolean romcache) {
        sp_mode = romcache;
    }

    abstract int load_rom(InputStream is, boolean from_sp) throws IOException;

	// =========================
	// ubN[h
	//==========================
    public final byte[][] load_blocks(InputStream is, int num_banks, int size)
    throws IOException {
        byte[][] b = new byte[num_banks][size + 2];
        for (int i = 0; i < num_banks; i++) {
            int n, p;
            for (p = 0; p < size; p += n) {
                if ((n = is.read(b[i], p, size - p)) <= 0) {
                    throw new IOException();
                }
            }
            gui.setMessage((i + 1) * size + "oCgǂݍ݂܂B");
        }
        return b;
    }

	// ===============================
	// fBXNVXeubN[h
	//================================
    public final byte[][] load_disk_blocks(InputStream is, int num_disks)
    throws IOException {
        byte[][] b = new byte[num_disks][65500];
        for (int i = 0; i < num_disks; i++) {
            int n = 0;
            for (int p = 0; p < 65500; p += n) {
                if ((n = is.read(b[i], p, 65500 - p)) <= 0) {
                    throw new IOException();
                }
            }
            gui.setMessage((i + 1) * 65500 + "oCgǂݍ݂܂B");
        }
        return b;
    }

	// ===========================
	// fBXNVXeBIOS[h
	//============================
    public final byte[] load_bios(boolean from_sp) {
        byte[] b = new byte[0x2000];
        SDbinding sd = null;
        InputStream is = null;
        try {
            boolean ok = false;
            if(!sp_mode || !from_sp) {
                sd = new SDbinding();
                is = sd.openInputStream("DISKSYS.ROM");
            } else {
                is = gui.openSPBIOSInputStream();
            }
            if(is != null) {
                do {
                    int n = 0;
                    for (int p = 0; p < 0x2000; p += n) {
                        if((n = is.read(b, p, 0x2000 - p)) <= 0) {
                            b = null;
                            break;
                        }
                        gui.setMessage("LOAD DISYSYS.ROM");
                    }
                    if(b[0] == 'N' && b[1] == 'E' && b[2] == 'S') {
                        for (int i = 0x2000; i < 0x6010; i++) is.read();
                    } else {
                        ok = true;
                    }
                } while(!ok);
            }
        } catch (Exception e) {
            b = null;
        } catch (Error err) {
            b = null;
        } finally {
            try{
                is.close();
            }catch(Exception e){}
            is = null;
            if(sd!=null) sd.close();
            sd = null;
        }        
        //Cancel license screen
        //BIOS[0x0FB0] = 0x00;
        return b;
    }

    abstract int reset();


	// =========================
	// ^XNC
	//==========================
    public void run() {
        switch (request) {
            case 1 :
                if(sp_mode) {
                    save_sram_sp();
                } else {
                    save_sram();
                }
                reset();
                break;
            case 2 :
                if(sp_mode) {
                    if(param == 0) {
                        save_sram();
                        save_sram_sp();
                    } else {
                        save_sram_sp();
                    }
                } else {
                    save_sram();
                }
                exist = false;
                break;
            case 3 :
            {
                SDbinding sd = null;
                DataInputStream dis = null;
                try{
                    if(!sp_mode && param > 0) {
                        sd = new SDbinding();
                        dis = sd.openDataInputStream(gui.getSaveFile(param));
                    } else {
                        dis = gui.openSPSaveInputStream(param);
                    }
                    if (dis != null) load_state(dis);
                }catch(Exception e){
                }finally{
                    try{
                        dis.close();
                    }catch(Exception e){}
                    dis = null;
                    if(sd!=null) sd.close();
                    sd = null;
                }
                break;
            }
            case 4 :
            {
                SDbinding sd = null;
                DataOutputStream dos = null;
                try {
                    if(!sp_mode && param > 0) {
                        sd = new SDbinding();
                        dos = sd.openDataOutputStream(gui.getSaveFile(param));
                    } else {
                        dos = gui.openSPSaveOutputStream(param);
                    }
                    if (dos != null){
                        save_state(dos);
                        dos.flush();
                    }
                } catch (IOException e) {
                } finally {
                    try {
                        dos.close();
                    } catch(Exception e) {}
                    dos = null;
                    if(sd!=null) sd.close();
                    sd = null;
                    if (sp_mode || param == 0)
                        gui.closeSPSaveOutputStream(param);
                }
                break;
            }
            case 5 :
                if(disk_system) {
                    mapper.SetDiskSide((byte)param);
                }
                break;
            default :
                break;
        }
    }

	// =========================
	// pbgf[^
	//==========================
//    public final int[] nes_palette, cur_palette;
    public int[] nes_palette;
    public int[] cur_palette;
    public final int[] zero_array;
    public int[] zeroPalette_array;
    {
        nes_palette =
            new int[] {
                0xff787878,
                0xff2000b0,
                0xff2800b8,
                0xff6010a0,
                0xff982078,
                0xffb01030,
                0xffa03000,
                0xff784000,
                0xff485800,
                0xff386800,
                0xff386c00,
                0xff306040,
                0xff305080,
                0xff000000,
                0xff000000,
                0xff000000,
                0xffb0b0b0,
                0xff4060f8,
                0xff4040ff,
                0xff9040f0,
                0xffd840c0,
                0xffd84060,
                0xffe05000,
                0xffc07000,
                0xff888800,
                0xff50a000,
                0xff48a810,
                0xff48a068,
                0xff4090c0,
                0xff000000,
                0xff000000,
                0xff000000,
                0xffe8e8e8,
                0xff60a0ff,
                0xff5080ff,
                0xffa070ff,
                0xfff060ff,
                -40784,
                -34768,
                -24576,
                0xffe8d020,
                0xff98e800,
                0xff70f040,
                0xff70e090,
                0xff60d0e0,
                0xff787878,
                0xff000000,
                0xff000000,
                -1,
                0xff90d0ff,
                0xffa0b8ff,
                0xffc0b0ff,
                0xffe0b0ff,
                -18200,
                -14152,
                -10080,
                -3952,
                0xffc8f080,
                0xffa0f0a0,
                0xffa0ffc8,
                0xffa0fff0,
                0xffa0a0a0,
                0xff000000,
                0xff000000,
                };
        cur_palette = new int[64];
        zero_array = new int[280];
        for (int i = 0; i < 280; i++)
            zero_array[i] = 0;
        zeroPalette_array = new int[280];
        for (int i = 0; i < 64; i++) {
            /*
            int c = nes_palette[i];
            cur_palette[i] =
            nes_palette[i] =
                Graphics.getColorOfRGB((c >> 16) & 0xff, (c >> 8) & 0xff, (c >> 0) & 0xff);
            */
                if(i == 0)
                    cur_palette[i] = nes_palette[i] & 0x00FFFFFF;
                else
                    cur_palette[i] = nes_palette[i];
        }
        for (int i = 0; i < 280; i++) zeroPalette_array[i] = cur_palette[0];
    }


/*==================================================================================================================*/
/*@Z[u/[h֘A		*/
/*------------------------------------------------------------------------------------------------------------------*/

    public final void save_state(DataOutputStream dos) throws IOException {
        mapper.PreStateSave();
        dos.write(RAM);
        if (!disk_system) {
            for (int i = 4; i < 8; i++)
                dos.writeInt(get_index(cpu_map[i], ROM));
        } else {
            dos.writeInt((int)(mapper.GetDiskSide() & 0xFF));
            for (int i = 5; i < 8; i++)
                dos.writeInt(-1);
        }
        dos.writeInt(cyc_pending);
        dos.writeInt(A);
        dos.writeInt(X);
        dos.writeInt(Y);
        dos.writeInt(S);
        dos.writeInt(PC);

        //NP2
        int P = (P_BI & 0x3c) | (P_C & 1) | ((P_V >> 1) & 0x40) | (P_ZN & 0x80);
        if (P_ZN <= 0)
            P |= 2;

        dos.writeInt(P);
        for (int i = 0; i < 12; i++)
            dos.write(VRAM[i]);
        for (int i = 0; i < 12; i++) {
            int j = get_index(ppu_map[i], VROM);
            if (j != -1) {
                dos.writeBoolean(true);
                dos.writeInt(j);
            } else {
                dos.writeBoolean(false);
                dos.writeInt(get_index(ppu_map[i], VRAM));
            }
        }
        for (int i = 0; i < spr_ram.length; i++)
            dos.writeInt(spr_ram[i]);
        for (int i = 0; i < pal_ram.length; i++)
            dos.writeInt(pal_ram[i]);
        dos.writeInt(PPU_R0);
        dos.writeInt(PPU_R1);
        dos.writeInt(PPU_R2);
        dos.writeInt(bg_pattern_table_addr);
        dos.writeInt(spr_pattern_table_addr);
        dos.writeInt(ppu_addr_inc);
        dos.writeBoolean(toggle_2005_2006);
        dos.writeInt(loopy_t);
        dos.writeInt(loopy_v);
        dos.writeInt(loopy_x);
        dos.writeInt(spr_ram_rw_ptr);
        dos.writeInt(read_2007_buffer);
        dos.writeInt(frame_irq_enabled);
        dos.write(SRAM);
        if (disk_system) {
            for(int i = 3; i <= 6; i++) dos.write(cpu_map[i], 0, 0x2000);
        }
        mapper.StateSave(dos);
    }

    public void load_state(DataInputStream dis) throws IOException {
        dis.read(RAM);
        if (!disk_system) {
            for (int i = 4; i < 8; i++) {
                int bank = dis.readInt();
                if(bank != -1) cpu_map[i] = ROM[bank];
            }
        } else {
            mapper.SetDiskSide((byte)dis.readInt());
            for (int i = 5; i < 8; i++) dis.readInt();
        }
        cyc_pending = dis.readInt();
        A = dis.readInt();
        X = dis.readInt();
        Y = dis.readInt();
        S = dis.readInt();
        PC = dis.readInt();
        P = dis.readInt();
        for (int i = 0; i < 12; i++)
            dis.read(VRAM[i]);
        for (int i = 0; i < 12; i++) {
            boolean is_vrom = dis.readBoolean();
            int num = dis.readInt();
            if (is_vrom) {
                ppu_map[i] = VROM[num];
                if (i < 8) {
                    patpix[i] = VROM_CACHE[num];
                    patdirty[i] = CACHE_HIT[8];
                }
            } else {
                ppu_map[i] = VRAM[num];
                if ( i < 8) {
                    patpix[i] = VRAM_CACHE[num];
                    patdirty[i] = CACHE_HIT[num];
                }
            }
        }
        for (int i = 0; i < spr_ram.length; i++)
            spr_ram[i] = dis.readInt();
        for (int i = 0; i < pal_ram.length; i++) {
            final int col = dis.readInt();
            pal_ram[i] = col;
            if(i == 0) {
                cur_palette[0] = nes_palette[col] & 0x00FFFFFF;
                //for(int j = 0; j < 280; j++) zeroPalette_array[j] = cur_palette[0];
            } else {
                cur_palette[i] = nes_palette[col];
            }
        }
        //NP2
        for(int i = 0; i < 280; i++) zeroPalette_array[i] = cur_palette[0];
        PPU_R0 = dis.readInt();
        PPU_R1 = dis.readInt();
        PPU_R2 = dis.readInt();
        bg_pattern_table_addr = dis.readInt();
        spr_pattern_table_addr = dis.readInt();
        ppu_addr_inc = dis.readInt();
        toggle_2005_2006 = dis.readBoolean();
        loopy_t = dis.readInt();
        loopy_v = dis.readInt();
        loopy_x = dis.readInt();
        spr_ram_rw_ptr = dis.readInt();
        read_2007_buffer = dis.readInt();
        for(int i = 0; i < 8; i++)
            for (int j = 0; j < 65; j++)
                    CACHE_HIT[i][j] = 0; 
        frame_irq_enabled = dis.readInt();
        dis.read(SRAM);
        if (disk_system) {
            for(int i = 3; i <= 6; i++) dis.read(cpu_map[i], 0, 0x2000);
            for(int i = 3; i <  6; i++) {
                cpu_map[i][0x2000] = cpu_map[i+1][0];
                cpu_map[i][0x2001] = cpu_map[i+1][1];
            }
        }
        mapper.StateLoad(dis);
    }

    public final int save_sram() {
        int ret = -1;
        SDbinding sd = null;
        DataOutputStream dos = null;
        String file = gui.getSramSaveFile();
        if(file == null) return ret;
        try {
            sd = new SDbinding();
            dos = sd.openDataOutputStream(file);
            if (dos != null){
                if(!disk_system) {
                    dos.write(SRAM);
                    dos.flush();
                    ret = 0;
                } else {
                    dos.write('F');
                    dos.write('D');
                    dos.write('S');
                    dos.write(0x1A);
                    dos.write(num_DISK_banks);
                    for (int i = 5; i < 0x10; i++) dos.write(0x00);
                    if (num_DISK_banks > 0) {
                        for (int i = 0; i < num_DISK_banks; i++) {
                            dos.write(DISK[i], 0, 65500);
                        }
                    }
                    dos.flush();
                    ret = 0;
                }
            }
        } catch (IOException e) {
        } finally {
            try{
                dos.close();
            } catch(Exception e) {}
            dos = null;
            if(sd!=null) sd.close();
            sd = null;
        }
        return ret;
    }

    public final int load_sram() {
        int ret = -1;
        SDbinding sd = null;
        DataInputStream dis = null;
        String file = gui.getSramSaveFile();
        if(file == null) return ret;
        try {
            if(!sp_mode) {
                sd = new SDbinding();
                dis = sd.openDataInputStream(file);
            } else {
                dis = gui.openSPSRAMInputStream();
            }
            if (dis != null){
                if(!disk_system) {
                    dis.read(SRAM);
                    ret = 0;
                } else {
                    if(!sp_mode) {
                        for (int i = 0; i < 0x10; i++) dis.read();
                        if (num_DISK_banks > 0) {
                            for (int i = 0; i < num_DISK_banks; i++) {
                                dis.read(DISK[i], 0, 65500);
                            }
                        }
                    }
                    ret = 0;
                }
            }
        } catch (IOException e) {
        } finally {
            try{
                dis.close();
            }catch(Exception e){}
            dis = null;
            if(sd!=null) sd.close();
            sd = null;
        }
        return ret;
    }
    
    public final int save_sram_sp() {
        int ret = -1;
        if(!disk_system) {
            ret = gui.writeSPSRAM(SRAM);
        } else {
            ret = gui.writeSPFD(num_DISK_banks, DISK);
        }
        return ret;
    }

    public final void request(int _request, int _param) {
        request = _request;
        param = _param;
    }

    private final int get_index(byte[] p, byte[][] a) {
        if (a == null) {
            return -1;
        }
        for (int i = 0; i < a.length; i++) {
            if (p == a[i]) {
                return i;
            }
        }
        return -1;
    }

/*==================================================================================================================*/
/*@֘A			*/
/*------------------------------------------------------------------------------------------------------------------*/

    public final int mem_read(int addr) {
        switch (addr >> 13) {
            case 0 :
                return RAM[addr & 0x07FF] & 0xFF;
            case 1 :
                int data;
                switch (addr & 7) {
                    case 1 :
                        return PPU_R1;
                    case 2 :
                        toggle_2005_2006 = false;
                        data = PPU_R2;
                        PPU_R2 &= 0x7F;
                        return data;
                    case 3 :
                        return spr_ram_rw_ptr;
                    case 4 :
                        data = spr_ram[spr_ram_rw_ptr];
                        spr_ram_rw_ptr = (spr_ram_rw_ptr + 1) & 0xFF;
                        return data;
                    case 7 :
                        addr = loopy_v & 0x3FFF;
                        loopy_v = (loopy_v + ppu_addr_inc) & 0xFFFF;
                        if (addr >= 0x3000)
                            return pal_ram[addr & 0x1F];
                        else {
                            int tmp = read_2007_buffer;
                            read_2007_buffer = ppu_map[addr >> 10][addr & 0x03FF] & 0xFF;
                            return tmp;
                        }
                    default :
                        return 0;
                }
            case 2 :
                if (addr >= 0x4018)
                    return mapper.MemoryReadLow(addr);
                switch (addr & 0x1F) {
                    case 0x15 :
                        if ((frame_irq_enabled & 0xC0) == 0)
                            return 0x40 | audio.apu_read(addr);
                            //return 0x40;
                        else
                            return audio.apu_read(addr);
                            //return 0;
                    case 0x16 :
                        return gui.readJoyPadBit();
                    case 0x17 :
                        return gui.readJoyPadBit2();
                    default :
                        return 0;
                }
            case 3 :
                mapper.MemoryReadSaveRAM(addr);
            default :
                return cpu_map[addr >> 13][addr & 0x1FFF] & 0xFF;
        }
    }

    private final void write_mem(int addr, int data) {
        switch (addr >> 13) {
            case 0 :
                RAM[addr & 0x07FF] = (byte) data;
                return;
            case 1 :
                switch (addr & 7) {
                    case 0 :
                        PPU_R0 = data;
                        final int b_addr = ((data & 0x10) << 8);
                        if (bg_pattern_table_addr != b_addr) {
                            bg_pattern_table_addr = b_addr;
                            //System.arraycopy(zero_array, 0, patdirty, b_addr >> 4, 256);
                        }
                        final int s_addr = ((data & 0x08) << 9);
                        if (spr_pattern_table_addr != s_addr) {
                            spr_pattern_table_addr = s_addr;
                            //System.arraycopy(zero_array, 0, patdirty, s_addr >> 4, 256);
                        }
                        ppu_addr_inc = (data & 4) != 0 ? 32 : 1;
                        loopy_t = (loopy_t & 0xF3FF) | ((data & 3) << 10);
                        return;
                    case 1 :
                        PPU_R1 = data;
                        return;
                    case 3 :
                        spr_ram_rw_ptr = data;
                        return;
                    case 4 :
                        spr_ram[spr_ram_rw_ptr++] = data;
                        return;
                    case 5 :
                        if (toggle_2005_2006 =
                                    !toggle_2005_2006) {
                            loopy_t = (loopy_t & 0xFFE0) | ((data & 0xF8) >> 3);
                            loopy_x = data & 7;
                        } else {
                            loopy_t =
                                (loopy_t & 0x8C1F)
                                | ((data & 0xF8) << 2)
                                | ((data & 7) << 12);
                        }
                        return;
                    case 6 :
                        if (toggle_2005_2006 = !toggle_2005_2006) {
                            loopy_t = (loopy_t & 0x00FF) | ((data & 0x3F) << 8);
                        } else {
                            loopy_v = loopy_t = (loopy_t & 0xFF00) | data;
                        }
                        return;
                    case 7 :
                        addr = loopy_v & 0x3FFF;
                        loopy_v = (loopy_v + ppu_addr_inc) & 0xFFFF;
                        if (addr >= 0x3F00) {
                            data &= 0x3F;
                            if ((addr & 0x0F) == 0) {
                                pal_ram[0] = pal_ram[0x10] = data;
                                cur_palette[0]    = nes_palette[data] & 0x00FFFFFF;
                                cur_palette[0x10] = nes_palette[data];
                                for(int i = 0; i < 280; i++) zeroPalette_array[i] = cur_palette[0];
                            } else {
                                final int pal = addr & 0x1F;
                                pal_ram[pal] = data;
                                cur_palette[pal] = nes_palette[data];
                            }
                        } else if (addr >= 0x2000) {
                            addr &= 0x2FFF;
                            ppu_map[addr >> 10][addr & 0x03FF] = (byte) data;
                        } else if (!vram_write_protect && patdirty[addr >> 10][64] == 0) {
                            ppu_map[addr >> 10][addr & 0x03FF] = (byte) data;
                            patdirty[addr >> 10][(addr >> 4) & 0x3F] = 0;
                        }
                        return;
                    default :
                        return;
                }
            case 2 :
                if (addr >= 0x4018) {
                    mapper.MemoryWriteLow(addr, data);
                    return;
                }
                switch (addr & 0x1F) {
                    case 0x00 :
                    case 0x01 :
                    case 0x02 :
                    case 0x03 :
                    case 0x04 :
                    case 0x05 :
                    case 0x06 :
                    case 0x07 :
                    case 0x08 :
                    case 0x09 :
                    case 0x0A :
                    case 0x0B :
                    case 0x0C :
                    case 0x0D :
                    case 0x0E :
                    case 0x0F :
                    case 0x10 :
                    case 0x11 :
                    case 0x12 :
                    case 0x13 :
                    case 0x15 :
                        //if (soundOut) audio.apuHandler(addr, data);
                        audio.apu_write(addr, data);
                        return;
                    case 0x14 :
                        byte[] mem = cpu_map[data >> 5];
                        int off = (data & 0x1F) << 8;
                        for (int i = 0; i < 256; i++)
                            spr_ram[i] = mem[off + i] & 0xFF;
                        cyc_pending -= 514;
                        return;
                    case 0x16 :
                        if ((data & 1) == 0)
                            gui.resetJoyPad();
                        return;
                    case 0x17 :
                        if ((data & 1) == 0)
                            gui.resetJoyPad2();
                        frame_irq_enabled = data;
                        return;
                    default :
                        return;
                }
            case 3 :
                SRAM[addr & 0x1FFF] = (byte) data;
                mapper.MemoryWriteSaveRAM(addr, data);
                return;
            default :
                mapper.MemoryWrite(addr, data);
                return;
        }
    }


/*==================================================================================================================*/
/*@CPU֘A					*/
/*------------------------------------------------------------------------------------------------------------------*/

    public byte[] RAM;
    public byte[][] cpu_map;
    public int cyc_pending;
    public int cyc_normal;
    public int cyc_before_nmi;
    public int cyc_after_nmi;
    public int cyc_before_hsync;
    public int cyc_after_hsync;
    public int cyc_correct;
    public int A;
    public int X;
    public int Y;
    public int S;
    public int PC;
    public int P;

    //NP2
    public int P_BI;
    public int P_C;
    public int P_ZN;
    public int P_V;

    public final void set_cpu_bank(int page, int bank) {
        if (num_16k_ROM_banks == 0) return;
        cpu_map[page] = ROM[bank % (num_16k_ROM_banks << 1)];
        if (page > 4) {
            try {
                cpu_map[page - 1][0x2000] = cpu_map[page][0x0000];
                cpu_map[page - 1][0x2001] = cpu_map[page][0x0001];
            } catch (Exception e) {}
        }
        if (page < 7) {
            try {
                cpu_map[page][0x2000] = cpu_map[page + 1][0x0000];
                cpu_map[page][0x2001] = cpu_map[page + 1][0x0001];
            } catch (Exception e) {}
        }
    }

    public void do_irq() {
        if ((P & 4) == 0) {
            P &= ~0x10;
            RAM[S + 0x100] = (byte) (PC >> 8);
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) PC;
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) P;
            S = (S - 1) & 0xFF;
            PC = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
            P |= 4;
            cyc_pending -= 7;
        }
    }

    public void do_pending_irq() {
        if ((P & 4) == 0) {
            P &= ~0x10;
            RAM[S + 0x100] = (byte) (PC >> 8);
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) PC;
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) P;
            S = (S - 1) & 0xFF;
            PC = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
            P |= 4;
            cyc_pending -= 7;
        } else {
            int_pending = 1;
        }
    }

    public final void cpu_emulate_CPU_cycles(int cyc) {
        int A = this.A;
        int PClo, PChi;
        int X = this.X;
        int Y = this.Y;
        int S = this.S;
        int P = this.P;
        int C, V, T;
        final byte[] RAM = this.RAM;
        final byte[][] cpu_map = this.cpu_map;
        byte[] curpage;

        cyc += cyc_pending;
        cyc_pending = 0;

        if(int_pending == 1 && cyc > 0 && (P & 4) == 0) {
            P &= ~0x10;
            RAM[S + 0x100] = (byte) (PC >> 8);
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) (PC);
            S = (S - 1) & 0xFF;
            RAM[S + 0x100] = (byte) P;
            S = (S - 1) & 0xFF;
            PC = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
            P |= 4;
            cyc -= 7;
            int_pending = 0;
        }

        C = P & 1;
        V = P << 1;
        T = ((P & 0x82) ^ 2) | ((P & 2) << 30);
        L0 : for (;;) {
            PChi = PC >> 13;
            PClo = PC & 0x1FFF;
            int cur_PClo = 0;
            try {
                while (cyc > 0) {
                    int tmp, ea;
                    cur_PClo = PClo;
                    curpage = cpu_map[PChi];
                    switch (curpage[PClo++] & 0xFF/*OPC*/) {
                        case 0x1 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            A =
                                T =
                                    A
                                    | mem_read(
                                        ((RAM[tmp] & 0xFF)
                                         | ((RAM[tmp + 1] & 0xFF) << 8)));
                            cyc -= 6;
                            break;
                        case 0x5 :
                            A = T = A | (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0x9 :
                            A = T = A | (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        case 0xd :
                            A =
                                T =
                                    A
                                    | mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF)
                                            << 8)));
                            cyc -= 4;
                            break;
                        case 0x11 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A | mem_read(ea);
                            cyc -= 5;
                            break;
                        case 0x15 :
                            A =
                                T =
                                    A
                                    | (RAM[((curpage[PClo++] & 0xFF) + X)
                                           & 0xFF]
                                       & 0xFF);
                            cyc -= 4;
                            break;
                        case 0x19 :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A | mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x1d :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            A = T = A | mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x21 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            A =
                                T =
                                    A
                                    & mem_read(
                                        ((RAM[tmp] & 0xFF)
                                         | ((RAM[tmp + 1] & 0xFF) << 8)));
                            cyc -= 6;
                            break;
                        case 0x25 :
                            A = T = A & (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0x29 :
                            A = T = A & (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        case 0x2d :
                            A =
                                T =
                                    A
                                    & mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF)
                                            << 8)));
                            cyc -= 4;
                            break;
                        case 0x31 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A & mem_read(ea);
                            cyc -= 5;
                            break;
                        case 0x35 :
                            A =
                                T =
                                    A
                                    & (RAM[((curpage[PClo++] & 0xFF) + X)
                                           & 0xFF]
                                       & 0xFF);
                            cyc -= 4;
                            break;
                        case 0x39 :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A & mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x3d :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            A = T = A & mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x41 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            A =
                                T =
                                    A
                                    ^ mem_read(
                                        ((RAM[tmp] & 0xFF)
                                         | ((RAM[tmp + 1] & 0xFF) << 8)));
                            cyc -= 6;
                            break;
                        case 0x45 :
                            A = T = A ^ (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0x49 :
                            A = T = A ^ (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        case 0x4d :
                            A =
                                T =
                                    A
                                    ^ mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF)
                                            << 8)));
                            cyc -= 4;
                            break;
                        case 0x51 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A ^ mem_read(ea);
                            cyc -= 5;
                            break;
                        case 0x55 :
                            A =
                                T =
                                    A
                                    ^ (RAM[((curpage[PClo++] & 0xFF) + X)
                                           & 0xFF]
                                       & 0xFF);
                            cyc -= 4;
                            break;
                        case 0x59 :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = A ^ mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x5d :
                            ea = (((curpage[PClo++] & 0xFF)
                                    | ((curpage[PClo++] & 0xFF)
                                   << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            A = T = A ^ mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0x61 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp =
                                mem_read(
                                    ((RAM[tmp] & 0xFF)
                                     | ((RAM[tmp + 1] & 0xFF) << 8)));
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 6;
                            break;
                        case 0x65 :
                            tmp = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 3;
                            break;
                        case 0x69 :
                            tmp = (curpage[PClo++] & 0xFF);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 2;
                            break;
                        case 0x6d :
                            tmp =
                                mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x71 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 5;
                            break;
                        case 0x75 :
                            tmp =
                                (RAM[((curpage[PClo++] & 0xFF) + X)
                                     & 0xFF]
                                 & 0xFF);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x79 :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x7d :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A + tmp + C;
                            C = (T >> 8) & 1;
                            V = (T ^ A) & (T ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x81 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            write_mem(
                                ((RAM[tmp] & 0xFF)
                                 | ((RAM[tmp + 1] & 0xFF) << 8)),
                                A);
                            cyc -= 6;
                            break;
                        case 0x85 :
                            RAM[(curpage[PClo++] & 0xFF)] = (byte) A;
                            cyc -= 3;
                            break;
                        case 0x89 :
                            cyc -= 2;
                            break;
                        case 0x8d :
                            write_mem(
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8)),
                                A);
                            cyc -= 4;
                            break;
                        case 0x91 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = (((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            write_mem(ea, A);
                            cyc -= 6;
                            break;
                        case 0x95 :
                            RAM[((curpage[PClo++] & 0xFF) + X) & 0xFF] =
                                (byte) A;
                            cyc -= 4;
                            break;
                        case 0x99 :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            write_mem(
                                ea,
                                A);
                            cyc -= 5;
                            break;
                        case 0x9d :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            write_mem(ea, A);
                            cyc -= 5;
                            break;
                        case 0xa1 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            A =
                                T =
                                    mem_read(
                                        ((RAM[tmp] & 0xFF)
                                         | ((RAM[tmp + 1] & 0xFF) << 8)));
                            cyc -= 6;
                            break;
                        case 0xa5 :
                            A = T = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0xa9 :
                            A = T = (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        case 0xad :
                            A =
                                T =
                                    mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF) << 8)));
                            cyc -= 4;
                            break;
                        case 0xb1 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = mem_read(ea);
                            cyc -= 5;
                            break;
                        case 0xb5 :
                            A =
                                T =
                                    (RAM[((curpage[PClo++] & 0xFF) + X)
                                         & 0xFF]
                                     & 0xFF);
                            cyc -= 4;
                            break;
                        case 0xb9 :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            A = T = mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0xbd :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            A = T = mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0xc1 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            T =
                                A
                                - mem_read(
                                    ((RAM[tmp] & 0xFF)
                                     | ((RAM[tmp + 1] & 0xFF) << 8)));
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 6;
                            break;
                        case 0xc5 :
                            T = A - (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 3;
                            break;
                        case 0xc9 :
                            T = A - (curpage[PClo++] & 0xFF);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 2;
                            break;
                        case 0xcd :
                            T =
                                A
                                - mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xd1 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            T = A - mem_read(ea);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 5;
                            break;
                        case 0xd5 :
                            T =
                                A
                                - (RAM[((curpage[PClo++] & 0xFF) + X)
                                       & 0xFF]
                                   & 0xFF);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xd9 :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            T = A - mem_read(ea);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xdd :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            T = A - mem_read(ea);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xe1 :
                            tmp = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp =
                                mem_read(
                                    ((RAM[tmp] & 0xFF)
                                     | ((RAM[tmp + 1] & 0xFF) << 8)));
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 6;
                            break;
                        case 0xe5 :
                            tmp = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 3;
                            break;
                        case 0xe9 :
                            tmp = (curpage[PClo++] & 0xFF);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 2;
                            break;
                        case 0xed :
                            tmp =
                                mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xf1 :
                            tmp = (curpage[PClo++] & 0xFF);
                            ea = ((((RAM[tmp] & 0xFF)
                                   | ((RAM[tmp + 1] & 0xFF) << 8)))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 5;
                            break;
                        case 0xf5 :
                            tmp =
                                (RAM[((curpage[PClo++] & 0xFF) + X)
                                     & 0xFF]
                                 & 0xFF);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xf9 :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xfd :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = A - tmp - 1 + C;
                            C = (~T >> 8) & 1;
                            V = (T ^ A) & (A ^ tmp);
                            A = T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x6 :
                            ea = (curpage[PClo++] & 0xFF);
                            tmp = (RAM[ea] & 0xFF);
                            T = (tmp << 1) & 0xFF;
                            C = (tmp >> 7) & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0xe :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            tmp = mem_read(ea);
                            T = (tmp << 1) & 0xFF;
                            C = (tmp >> 7) & 1;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0x16 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp = (RAM[ea] & 0xFF);
                            T = (tmp << 1) & 0xFF;
                            C = (tmp >> 7) & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0x1e :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = (tmp << 1) & 0xFF;
                            C = (tmp >> 7) & 1;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0x26 :
                            ea = (curpage[PClo++] & 0xFF);
                            tmp = (RAM[ea] & 0xFF);
                            T = ((tmp << 1) & 0xFF) | C;
                            C = (tmp >> 7) & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0x2e :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            tmp = mem_read(ea);
                            T = ((tmp << 1) & 0xFF) | C;
                            C = (tmp >> 7) & 1;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0x36 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp = (RAM[ea] & 0xFF);
                            T = ((tmp << 1) & 0xFF) | C;
                            C = (tmp >> 7) & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0x3e :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = ((tmp << 1) & 0xFF) | C;
                            C = (tmp >> 7) & 1;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0x46 :
                            ea = (curpage[PClo++] & 0xFF);
                            tmp = (RAM[ea] & 0xFF);
                            T = tmp >> 1;
                            C = tmp & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0x4e :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            tmp = mem_read(ea);
                            T = tmp >> 1;
                            C = tmp & 1;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0x56 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp = (RAM[ea] & 0xFF);
                            T = tmp >> 1;
                            C = tmp & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0x5e :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = tmp >> 1;
                            C = tmp & 1;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0x66 :
                            ea = (curpage[PClo++] & 0xFF);
                            tmp = (RAM[ea] & 0xFF);
                            T = (tmp >> 1) | (C << 7);
                            C = tmp & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0x6e :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            tmp = mem_read(ea);
                            T = (tmp >> 1) | (C << 7);
                            C = tmp & 1;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0x76 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            tmp = (RAM[ea] & 0xFF);
                            T = (tmp >> 1) | (C << 7);
                            C = tmp & 1;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0x7e :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            tmp = mem_read(ea);
                            T = (tmp >> 1) | (C << 7);
                            C = tmp & 1;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0x86 :
                            RAM[(curpage[PClo++] & 0xFF)] = (byte) X;
                            cyc -= 3;
                            break;
                        case 0x8e :
                            write_mem(
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8)),
                                X);
                            cyc -= 4;
                            break;
                        case 0x96 :
                            RAM[((curpage[PClo++] & 0xFF) + Y) & 0xFF] =
                                (byte) X;
                            cyc -= 4;
                            break;
                        case 0xa6 :
                            X = T = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0xae :
                            X =
                                T =
                                    mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF) << 8)));
                            cyc -= 4;
                            break;
                        case 0xb6 :
                            X =
                                T =
                                    (RAM[((curpage[PClo++] & 0xFF) + Y)
                                         & 0xFF]
                                     & 0xFF);
                            cyc -= 4;
                            break;
                        case 0xbe :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + Y)
                                 & 0xFFFF;
                            if((ea & 0xFF) < Y) cyc -= 1;
                            X = T = mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0xc6 :
                            ea = (curpage[PClo++] & 0xFF);
                            T = ((RAM[ea] & 0xFF) - 1) & 0xFF;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0xce :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            T = (mem_read(ea) - 1) & 0xFF;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0xd6 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            T = ((RAM[ea] & 0xFF) - 1) & 0xFF;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0xde :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            T = (mem_read(ea) - 1) & 0xFF;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0xe6 :
                            ea = (curpage[PClo++] & 0xFF);
                            T = ((RAM[ea] & 0xFF) + 1) & 0xFF;
                            RAM[ea] = (byte) T;
                            cyc -= 5;
                            break;
                        case 0xee :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            T = (mem_read(ea) + 1) & 0xFF;
                            write_mem(ea, T);
                            cyc -= 6;
                            break;
                        case 0xf6 :
                            ea = ((curpage[PClo++] & 0xFF) + X) & 0xFF;
                            T = ((RAM[ea] & 0xFF) + 1) & 0xFF;
                            RAM[ea] = (byte) T;
                            cyc -= 6;
                            break;
                        case 0xfe :
                            ea =
                                (((curpage[PClo++] & 0xFF)
                                  | ((curpage[PClo++] & 0xFF) << 8))
                                 + X)
                                & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            T = (mem_read(ea) + 1) & 0xFF;
                            write_mem(ea, T);
                            cyc -= 7;
                            break;
                        case 0x24 :
                            T = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            V = T << 1;
                            if ((A & T) == 0)
                                T |= ~0x1FF;
                            cyc -= 3;
                            break;
                        case 0x2c :
                            T =
                                mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            V = T << 1;
                            if ((A & T) == 0)
                                T |= ~0x1FF;
                            cyc -= 4;
                            break;
                        case 0x84 :
                            RAM[(curpage[PClo++] & 0xFF)] = (byte) Y;
                            cyc -= 3;
                            break;
                        case 0x8c :
                            write_mem(
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8)),
                                Y);
                            cyc -= 4;
                            break;
                        case 0x94 :
                            RAM[((curpage[PClo++] & 0xFF) + X) & 0xFF] =
                                (byte) Y;
                            cyc -= 4;
                            break;
                        case 0xa4 :
                            Y = T = (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            cyc -= 3;
                            break;
                        case 0xac :
                            Y =
                                T =
                                    mem_read(
                                        ((curpage[PClo++] & 0xFF)
                                         | ((curpage[PClo++] & 0xFF) << 8)));
                            cyc -= 4;
                            break;
                        case 0xb4 :
                            Y =
                                T =
                                    (RAM[((curpage[PClo++] & 0xFF) + X)
                                         & 0xFF]
                                     & 0xFF);
                            cyc -= 4;
                            break;
                        case 0xbc :
                            ea = (((curpage[PClo++] & 0xFF)
                                   | ((curpage[PClo++] & 0xFF) << 8))
                                  + X)
                                 & 0xFFFF;
                            if((ea & 0xFF) < X) cyc -= 1;
                            Y = T = mem_read(ea);
                            cyc -= 4;
                            break;
                        case 0xc4 :
                            T = Y - (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 3;
                            break;
                        case 0xcc :
                            T =
                                Y
                                - mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0xe4 :
                            T = X - (RAM[(curpage[PClo++] & 0xFF)] & 0xFF);
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 3;
                            break;
                        case 0xec :
                            T =
                                X
                                - mem_read(
                                    ((curpage[PClo++] & 0xFF)
                                     | ((curpage[PClo++] & 0xFF) << 8)));
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 4;
                            break;
                        case 0x0 :
                            tmp = (PChi << 13) + PClo + 1;
                            RAM[S + 0x100] = (byte) (tmp >> 8);
                            S = (S - 1) & 0xFF;
                            RAM[S + 0x100] = (byte) (tmp);
                            S = (S - 1) & 0xFF;
                            P = (P & 0x3c) | (C) | ((V >> 1) & 0x40) | (T & 0x80);
                            if (T <= 0)
                                P |= 2;
                            P |= 0x10;
                            RAM[S + 0x100] = (byte) P;
                            S = (S - 1) & 0xFF;
                            ea = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
                            curpage = cpu_map[PChi = ea >> 13];
                            PClo = ea & 0x1FFF;
                            P |= 4;
                            cyc -= 7;
                            break;
                        case 0x8 :
                            P = (P & 0x3c) | (C) | ((V >> 1) & 0x40) | (T & 0x80);
                            if (T <= 0)
                                P |= 2;
                            RAM[S + 0x100] = (byte) (P | 0x10);
                            S = (S - 1) & 0xFF;
                            cyc -= 3;
                            break;
                        case 0x18 :
                            C = 0;
                            cyc -= 2;
                            break;
                        case 0x20 :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            tmp = (PChi << 13) + PClo - 1;
                            RAM[S + 0x100] = (byte) (tmp >> 8);
                            S = (S - 1) & 0xFF;
                            RAM[S + 0x100] = (byte) (tmp);
                            S = (S - 1) & 0xFF;
                            curpage = cpu_map[PChi = ea >> 13];
                            PClo = ea & 0x1FFF;
                            cyc -= 6;
                            break;
                        case 0x28 :
                            S = (S + 1) & 0xFF;
                            P = RAM[S + 0x100] & 0xFF;
                            C = P & 1;
                            V = P << 1;
                            T = ((P & 0x82) ^ 2) | ((P & 2) << 30);
                            cyc -= 4;
                            break;
                        case 0x38 :
                            C = 1;
                            cyc -= 2;
                            break;
                        case 0x40 :
                            S = (S + 1) & 0xFF;
                            P = RAM[S + 0x100] & 0xFF;
                            C = P & 1;
                            V = P << 1;
                            T = ((P & 0x82) ^ 2) | ((P & 2) << 30);
                            S = (S + 1) & 0xFF;
                            PClo = RAM[S + 0x100] & 0xFF;
                            S = (S + 1) & 0xFF;
                            tmp = RAM[S + 0x100] & 0xFF;
                            PClo += ((tmp & 0x1F) << 8);
                            curpage = cpu_map[PChi = tmp >> 5];
                            cyc -= 6;
                            if(int_pending == 1 && cyc > 0 && (P & 4) == 0) {
                                P &= ~0x10;
                                tmp = (PChi << 13) + PClo;
                                RAM[S + 0x100] = (byte) (tmp >> 8);
                                S = (S - 1) & 0xFF;
                                RAM[S + 0x100] = (byte) (tmp);
                                S = (S - 1) & 0xFF;
                                RAM[S + 0x100] = (byte) P;
                                S = (S - 1) & 0xFF;
                                ea = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
                                curpage = cpu_map[PChi = ea >> 13];
                                PClo = ea & 0x1FFF;
                                P |= 4;
                                cyc -= 7;
                                int_pending = 0;
                            }
                            break;
                        case 0x48 :
                            RAM[S + 0x100] = (byte) (A);
                            S = (S - 1) & 0xFF;
                            cyc -= 3;
                            break;
                        case 0x58 :
                            P &= ~4;
                            cyc -= 2;
                            if(int_pending == 1 && cyc > 0) {
                                P &= ~0x10;
                                tmp = (PChi << 13) + PClo;
                                RAM[S + 0x100] = (byte) (tmp >> 8);
                                S = (S - 1) & 0xFF;
                                RAM[S + 0x100] = (byte) (tmp);
                                S = (S - 1) & 0xFF;
                                P = (P & 0x3c) | (C) | ((V >> 1) & 0x40) | (T & 0x80);
                                if (T <= 0)
                                    P |= 2;
                                RAM[S + 0x100] = (byte) P;
                                S = (S - 1) & 0xFF;
                                ea = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
                                curpage = cpu_map[PChi = ea >> 13];
                                PClo = ea & 0x1FFF;
                                P |= 4;
                                cyc -= 7;
                                int_pending = 0;
                            }
                            break;
                        case 0x60 :
                            S = (S + 1) & 0xFF;
                            PClo = RAM[S + 0x100] & 0xFF;
                            S = (S + 1) & 0xFF;
                            tmp = RAM[S + 0x100] & 0xFF;
                            PClo += ((tmp & 0x1F) << 8) + 1;
                            curpage = cpu_map[PChi = tmp >> 5];
                            cyc -= 6;
                            break;
                        case 0x68 :
                            S = (S + 1) & 0xFF;
                            T = RAM[S + 0x100] & 0xFF;
                            A = T;
                            cyc -= 4;
                            break;
                        case 0x78 :
                            P |= 4;
                            cyc -= 2;
                            break;
                        case 0x88 :
                            Y = T = (Y - 1) & 0xFF;
                            cyc -= 2;
                            break;
                        case 0x98 :
                            A = T = Y;
                            cyc -= 2;
                            break;
                        case 0xa0 :
                            Y = T = (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        case 0xa8 :
                            Y = T = A;
                            cyc -= 2;
                            break;
                        case 0xb8 :
                            V = 0;
                            cyc -= 2;
                            break;
                        case 0xc0 :
                            tmp = (curpage[PClo++] & 0xFF);
                            T = Y - tmp;
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 2;
                            break;
                        case 0xc8 :
                            Y = T = (Y + 1) & 0xFF;
                            cyc -= 2;
                            break;
                        case 0xd8 :
                            P &= ~8;
                            cyc -= 2;
                            break;
                        case 0xe0 :
                            tmp = (curpage[PClo++] & 0xFF);
                            T = X - tmp;
                            C = (~T >> 8) & 1;
                            T &= 0xFF;
                            cyc -= 2;
                            break;
                        case 0xe8 :
                            X = T = (X + 1) & 0xFF;
                            cyc -= 2;
                            break;
                        case 0xf8 :
                            P |= 8;
                            cyc -= 2;
                            break;
                        case 0xa :
                            T = (A << 1) & 0xFF;
                            C = (A >> 7) & 1;
                            A = T;
                            cyc -= 2;
                            break;
                        case 0x2a :
                            T = ((A << 1) & 0xFF) | C;
                            C = (A >> 7) & 1;
                            A = T;
                            cyc -= 2;
                            break;
                        case 0x4a :
                            T = A >> 1;
                            C = A & 1;
                            A = T;
                            cyc -= 2;
                            break;
                        case 0x6a :
                            T = (A >> 1) | (C << 7);
                            C = A & 1;
                            A = T;
                            cyc -= 2;
                            break;
                        case 0x8a :
                            A = T = X;
                            cyc -= 2;
                            break;
                        case 0xaa :
                            X = T = A;
                            cyc -= 2;
                            break;
                        case 0xca :
                            X = T = (X - 1) & 0xFF;
                            cyc -= 2;
                            break;
                        case 0xea :
                            cyc -= 2;
                            break;
                        case 0x9a :
                            S = X;
                            cyc -= 2;
                            break;
                        case 0xba :
                            X = T = S;
                            cyc -= 2;
                            break;
                        case 0x10 :
                            tmp = curpage[PClo++];
                            if ((T & 0x80) == 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0x30 :
                            tmp = curpage[PClo++];
                            if ((T & 0x80) != 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0x50 :
                            tmp = curpage[PClo++];
                            if ((V & 0x80) == 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0x70 :
                            tmp = curpage[PClo++];
                            if ((V & 0x80) != 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0x90 :
                            tmp = curpage[PClo++];
                            if (C == 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0xb0 :
                            tmp = curpage[PClo++];
                            if (C != 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0xd0 :
                            tmp = curpage[PClo++];
                            if (T > 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0xf0 :
                            tmp = curpage[PClo++];
                            if (T <= 0) {
                                if((((PClo & 0xFF) + tmp) & 0x100) != 0)
                                    cyc -= 1;
                                PClo += tmp;
                                cyc -= 3;
                                break;
                            } else {
                                cyc -= 2;
                                break;
                            }
                        case 0x4c :
                            ea =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            curpage = cpu_map[PChi = ea >> 13];
                            PClo = ea & 0x1FFF;
                            cyc -= 3;
                            break;
                        case 0x6c :
                            tmp =
                                ((curpage[PClo++] & 0xFF)
                                 | ((curpage[PClo++] & 0xFF) << 8));
                            if ((tmp & 0xFF) != 0xFF)
                                ea = mem_read(tmp) | (mem_read(tmp + 1) << 8);
                            else
                                ea = mem_read(tmp) | (mem_read(tmp & 0xFF00) << 8);
                            curpage = cpu_map[PChi = ea >> 13];
                            PClo = ea & 0x1FFF;
                            cyc -= 5;
                            break;
                        case 0xa2 :
                            X = T = (curpage[PClo++] & 0xFF);
                            cyc -= 2;
                            break;
                        default :
                            break;
                    }
                    if((PClo & ~0x1FFF) != 0) {
                        ea = (PChi << 13) + PClo;
                        curpage = cpu_map[PChi = ea >> 13];
                        PClo = ea & 0x1FFF;
                        //System.out.println(" -> PC (" + Integer.toHexString(ea) + ")");
                    }
                }
                PC = (PChi << 13) + PClo;
                break L0;
            } catch (ArrayIndexOutOfBoundsException e) {
                PC = (PChi << 13) + cur_PClo;
                continue L0;
            }
        }
        this.A = A;
        this.X = X;
        this.Y = Y;
        this.S = S;
        P = (P & 0x3c) | (C) | ((V >> 1) & 0x40) | (T & 0x80);
        if (T <= 0)
            P |= 2;
        this.P = P;
        cyc_pending += cyc;
    }


/*==================================================================================================================*/
/*@PPU֘A					*/
/*------------------------------------------------------------------------------------------------------------------*/

    public static final int MIRROR_HORIZ = 0;
    public static final int MIRROR_VERT = 1;
    public static final int MIRROR_FOUR_SCREEN = 2;
    public byte[][] VRAM;
    public byte[][] ppu_map;
    public int[] spr_ram;
    public int[] pal_ram;
    public int PPU_R0;
    public int PPU_R1;
    public int PPU_R2;
    public int bg_pattern_table_addr;
    public int spr_pattern_table_addr;
    public int ppu_addr_inc;
    public boolean toggle_2005_2006;
    public int loopy_t;
    public int loopy_v;
    public int loopy_x;
    public int spr_ram_rw_ptr;
    public int read_2007_buffer;
    public boolean vram_write_protect;
    public int[][] patpix;
    public byte[][] patdirty;
    public byte[] sprite_hit;


    public final void set_ppu_bank(int page, int bank) {
        if (num_8k_VROM_banks != 0) {
            bank %= (num_8k_VROM_banks << 3);
            ppu_map[page] = VROM[bank];
            if (page < 8) {
                patpix[page] = VROM_CACHE[bank];
                patdirty[page] = CACHE_HIT[8];
                //System.arraycopy(true_array, 0, patdirty[page], 0, 64);
            }
        } else {
            if (page < 8) {
                bank &= 7;
                ppu_map[page] = VRAM[bank];
                patpix[page] = VRAM_CACHE[bank];
                patdirty[page] = CACHE_HIT[bank];
                //System.arraycopy(zero_array, 0, patdirty[page], 0, 64);
            } else {
                ppu_map[page] = VRAM[(bank & 3) + 8];
            }
        }
    }

    public final void set_dummy_ppu_bank(int page) {
        ppu_map[page] = DUMMY_VROM;
        if (page < 8) {
            patpix[page] = DUMMY_VROM_CACHE;
            patdirty[page] = CACHE_HIT[8];
            //System.arraycopy(true_array, 0, patdirty[page], 0, 64);
        }
    }

    public final void set_vram_bank(int page, int bank) {
        if (page < 8) {
            bank &= 7;
            ppu_map[page] = VRAM[bank];
            patpix[page] = VRAM_CACHE[bank];
            patdirty[page] = CACHE_HIT[bank];
            //System.arraycopy(zero_array, 0, patdirty[page], 0, 64);
        } else {
            ppu_map[page] = VRAM[(bank & 3) + 8];
        }
    }

    public final void set_mirroring(int mirroring) {
        switch (mirroring) {
            case MIRROR_HORIZ :
                set_mirroring(0, 0, 1, 1);
                break;
            case MIRROR_VERT :
                set_mirroring(0, 1, 0, 1);
                break;
            case MIRROR_FOUR_SCREEN :
                set_mirroring(0, 1, 2, 3);
                break;
        }
    }

    public final void set_mirroring(int nt0, int nt1, int nt2, int nt3) {
        ppu_map[0x8] = VRAM[(nt0 & 3) + 8];
        ppu_map[0x9] = VRAM[(nt1 & 3) + 8];
        ppu_map[0xA] = VRAM[(nt2 & 3) + 8];
        ppu_map[0xB] = VRAM[(nt3 & 3) + 8];
    }


/*------------------------------------------------------------------------------------------------------------------*/

    public final byte get_disk_side() {
        if (mapper == null) return 0;
        return mapper.GetDiskSide();
    }

    //NP2
    public int get_default_cycle_line() {
        //return CYCLE_PER_LINE;
        return 0;
    }

    public final void setApuVolume(int enable, int vol) {
        if (enable > 0) {
            audio.apu_enable(true);
            audio.apu_setvol(vol);
        } else {
            audio.apu_enable(false);
        }
    }

    public final void setApuUserSetting() {
        audio.apu_setting();
    }
    
    public final void setApuMode(int mode) {
        audio.apu_setmode(mode);
    }

    public final boolean isDiskSystem() {
        return disk_system;
    }

    public int nes6502_getcycles() {return 0;}
    
    public int nes6502_getbyte(int addr) {
        return mem_read(addr);
    }


    // NP2
    public boolean is_blank_line() {
        return false;
    }

    public void set_mapper_irq() {
    }

    public void clear_mapper_irq() {
    }
    
    public void set_mapper2_irq() {
    }

    public void clear_mapper2_irq() {
    }

    /*************************************
        APU
    *************************************/
    public final void init_apu() {
        audio.apu_reset();
    }

}
