import java.io.*;

import com.nttdocomo.ui.Graphics;

public final class NESCore_NP extends NESCore {

    private int CYCLE_BEFORE_NMI = 1;
    private float CYCLE_PER_LINE = 114.0f;
    private boolean BANKSWITCH_PER_TILE;

    public NESCore_NP(NESCanvas _gui, PSGi _psgi) {
        gui = _gui;
        psgi = _psgi;
        crc_table = new int[256];
        for (int n = 0; n < 256; n++) {
            int c = n;
            for (int k = 8; --k >= 0;) {
                if ((c & 1) != 0)
                    c = 0xedb88320 ^ (c >>> 1);
                else
                    c = c >>> 1;
            }
            crc_table[n] = c;
        }
        switch (gui.getApuNum()) {
            case 1 :
                audio = new NESApu1(this, gui, psgi);
                break;
            case 2 :
                audio = new NESApu2(this, gui, psgi);
                break;
            case 3 :
                audio = new NESApu_NP2(this, gui, psgi);
                break;
            default :
                break;
        }
    }

    public final int load_rom(InputStream is, boolean from_sp) throws IOException {
        int b0, b1, b2, b3, b6, b7;
        b6 = b7 = 0;

        b0 = is.read();
        b1 = is.read();
        b2 = is.read();
        b3 = is.read();
        if (b0 == 'N' && b1 == 'E' && b2 == 'S' && b3 == 0x1A) {
            num_16k_ROM_banks = is.read();
            num_8k_VROM_banks = is.read();
            num_DISK_banks = 0;
            b6 = is.read();
            b7 = is.read();
            mirroring =
                (b6 & 1) != 0
                ? MIRROR_VERT
                : (b6 & 8) != 0
                ? MIRROR_FOUR_SCREEN
                : MIRROR_HORIZ;
            mapper_no = ((b6 >> 4) & 0x0F) | (b7 & 0xF0);
            CYCLE_PER_LINE = 114.0f;
            CYCLE_BEFORE_NMI = 1;
            for (int i = 0; i < 8; i++) {
                if (is.read() != 0) {
                    mapper_no &= 0x0F;
                }
            }
            if ((b6 & 4) != 0) {
                return -2;
            }
            ROM = load_blocks(is, num_16k_ROM_banks << 1, 0x2000);
            int c = ~crc;
            for (int i = 0; i < num_16k_ROM_banks << 1; i++) {
                for (int j = 0; j < 0x2000; j++) {
                    c = crc_table[(c ^ ROM[i][j]) & 0xff] ^ (c >>> 8);
                }
            }
            crc = ~c;
            correctRom(crc);
            DISK = null;
            BIOS = null;
            if (mapper_no == 5) {
                SRAM = new byte[0x10000];
                SRAM_SIZE = 0x10000;
            } else if( crc == 0xb8747abf) { // Pro Yakyuu Special
                SRAM = new byte[0x4000];
                SRAM_SIZE = 0x4000;
            }
            disk_system = false;
            if (mapper_no == 0x14) {
                return -3;
            }
        } else if(b0 == 'F' && b1 == 'D' && b2 == 'S' && b3 == 0x1A) {
            num_16k_ROM_banks = 0;
            num_8k_VROM_banks = 0;
            num_DISK_banks = is.read();
            mirroring = MIRROR_HORIZ;
            mapper_no = 0x14;
            CYCLE_PER_LINE = 113.0f;
            CYCLE_BEFORE_NMI = 1;
            for(int i = 5; i < 0x10; i++) is.read();
            ROM  = null;
            DISK = load_disk_blocks(is, num_DISK_banks);
            BIOS = load_bios(from_sp);
            if(BIOS == null) return -4;
            disk_system = true;
        } else {
            return -1;
        }
        if ((mapper = Mapper.createMapper(mapper_no, this)) == null) {
            return -3;
        }
        if (num_8k_VROM_banks > 0) {
            VROM = load_blocks(is, num_8k_VROM_banks << 3, 0x0400);
            vram_write_protect = (crc != 0xadb47286);
            VROM_CACHE = new int[num_8k_VROM_banks << 3][512];
            for (int i = 0; i < (num_8k_VROM_banks << 3); i++) {
                for (int j = 0; j < 64; j++) {
                    for (int k = 0; k < 8; k++) {
                        int lo = VROM[i][(j << 4) + k    ] & 0xFF;
                        int hi = VROM[i][(j << 4) + k + 8] & 0xFF;
                        int pat;
                        pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
                            | ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
                            | ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
                            | ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
                            | ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
                            | ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
                            | ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
                            | ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
                        VROM_CACHE[i][(j << 3) + k]=
                            pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
                    }
                }
            }
        }
        // INIT or LOAD SRAM
        int c = load_sram();
        if(c != 0)
            for (int i = 0; i < SRAM_SIZE; i++) SRAM[i] = (byte) 0xFF;
        // SAVE ROM@SP
        if(sp_mode && !from_sp) {
            if(!disk_system) {
                gui.writeSPROM(num_16k_ROM_banks, num_8k_VROM_banks, b6, b7, ROM, VROM);
            } else {
                gui.writeSPFD(num_DISK_banks, DISK);
                gui.writeSPBIOS(BIOS);
            }
        }
        // SET DUMMY CHARCTER ROM
        if (mapper_no == 185) {
            DUMMY_VROM = new byte[0x400];
            DUMMY_VROM_CACHE = new int[512];
            for(int i = 0; i < 0x400; i++)
                DUMMY_VROM[i] = (byte)0xFF;
            for (int i = 0; i < 512; i++)
                DUMMY_VROM_CACHE[i] = (int)0xFFFFFFFF;
        }
        return 0;
    }

    private final void correctRom(int crc) {
        switch(crc){
        case 0x5c707ac4: //MOTHER
            BANKSWITCH_PER_TILE = true;
            CYCLE_PER_LINE = 113.75f;
            //gui.adjust_cycle = 0.0f;
            break;
        case 0x761ccfb5:
            CYCLE_PER_LINE = 119.0f;
            CYCLE_BEFORE_NMI = 1;
            //gui.adjust_cycle = 0.0f;
            break;
        case 0xb00abe1c:
        case 0xa0b8d310:
            CYCLE_PER_LINE = 113.0f;
            CYCLE_BEFORE_NMI = 1;
            //gui.adjust_cycle = 0.0f;
            break;
        case 0xa52a8119:
            CYCLE_PER_LINE = 115.0f;
            CYCLE_BEFORE_NMI = 1;
            //gui.adjust_cycle = 0.0f;
            break;
        case 0x57970078: //F-1 Race
            ROM[1][0x1FE1] = (byte)0xFF;
            ROM[1][0x1FE6] = (byte)0x00;
            //gui.adjust_cycle = 0.0f;
            break;
        }
    }
    

    public final int reset() {
        // reset memory
        for (int i = 0; i < 8; i++)
            cpu_map[i] = null;
        for (int i = 0; i < 0x0800; i++)
            RAM[i] = 0x00;
        cpu_map[0] = RAM;
        cpu_map[3] = SRAM;
        // reset ppu
        for (int i = 0; i < 12; i++)
            for(int j = 0; j < 0x400; j++)
                VRAM[i][j] = (byte)0x00;
        for (int i = 0; i < 65; i++) {
            for(int j = 0; j < 8; j++)
                CACHE_HIT[j][i] = 0; 
            CACHE_HIT[8][i] = 1;
        }
        for (int i = 0; i < 8; i++) {
            ppu_map[i] = VRAM[i];
            patpix[i] = VRAM_CACHE[i];
            patdirty[i] = CACHE_HIT[i];
        }
        set_mirroring(mirroring);
        PPU_R0 = 0;
        PPU_R1 = 0;
        PPU_R2 = 0;
        bg_pattern_table_addr = 0;
        spr_pattern_table_addr = 0;
        ppu_addr_inc = 0;
        toggle_2005_2006 = false;
        loopy_v = 0;
        loopy_t = 0;
        loopy_x = 0;
        spr_ram_rw_ptr = 0;
        read_2007_buffer = 0;
        frame_irq_enabled = 0xFF;
        for (int i = 0; i < 256; i++)
            spr_ram[i] = 0;
        for (int i = 0; i < 32; i++)
            pal_ram[i] = 0;
        // reset mmc
        try{
            mapper.Reset();
        } catch (Exception e) {
            return -1;
        }
        // reset cpu
        A = X = Y = 0;
        S = 0xFF;
        P = 0x26;
        PC = mem_read(0xFFFC) | (mem_read(0xFFFD) << 8);
        int_pending = 0;
        cyc_pending = -6;
        cyc_normal = (int)Math.floor(CYCLE_PER_LINE + gui.adjust_cycle);
        cyc_before_nmi = CYCLE_BEFORE_NMI;
        cyc_after_nmi = (int)Math.floor(CYCLE_PER_LINE - CYCLE_BEFORE_NMI + gui.adjust_cycle);
        cyc_correct = (int)Math.floor(((CYCLE_PER_LINE + gui.adjust_cycle) - cyc_normal) * 4);
        cyc_correct = (cyc_correct == 1) ? (2) : (cyc_correct == 2) ? (1) : (cyc_correct);
        if(BANKSWITCH_PER_TILE) {
            cyc_before_hsync = 13;
            cyc_after_hsync = cyc_normal - cyc_before_hsync - 88;
        } else {
            cyc_before_hsync = 0;
            cyc_after_hsync = 0;
        }
        //System.out.println("CYC = " + cyc_normal + ": CORRECT = " + cyc_correct);
        //System.out.println("BEFORE HSYNC = " + cyc_before_hsync + ": AFTER HSYNC = " + cyc_after_hsync);
        // reset apu
        switch (gui.getApuNum()) {
            case 1 :
                audio = new NESApu1(this, gui, psgi);
                break;
            case 2 :
                audio = new NESApu2(this, gui, psgi);
                break;
            case 3 :
                audio = new NESApu_NP2(this, gui, psgi);
                break;
            default :
                break;
        }
        audio.apu_setmode(gui.soundMode);
        audio.apu_setting();
        audio.apu_reset();
        return 0;
    }


    public final void run() {
        final int[] line_buf = new int[280];
        final int[] bg_flags = new int[280];
        final int[] fg_flags = new int[280]; //FOR Mapper #9
        final byte[] sprite_hit = new byte[16];
        exist = true;
        for (;;) {
            super.run();
            switch (request) {
                case 2 :
                    request = 0;
                    exist = false;
                    return;
                case 7 :
                    continue;
                default :
                    break;
            }
            request = 0;
            try {
                if ((PPU_R1 & 0x18) != 0)
                    loopy_v = loopy_t;
                boolean render_frame = gui.startFrame();
                for (int i = 0; i < 240; i++) {
                    if (render_frame) {
                        if (!BANKSWITCH_PER_TILE) {
                            cpu_emulate_CPU_cycles(cyc_normal);
                            mapper.HSync(i);
                        }
                        if ((PPU_R1 & 8) == 0)
                            System.arraycopy(zeroPalette_array, 0, line_buf, 0, 280);
                        if ((PPU_R1 & 0x18) != 0) {
                            System.arraycopy(zero_array, 0, bg_flags, 0, 280);
                            loopy_v = (loopy_v & 0xFBE0) | (loopy_t & 0x041F);
                            if ((PPU_R1 & 8) != 0) {
                                final int bg_pattern_table_addr = this.bg_pattern_table_addr;
                                final int loopy_v = this.loopy_v;
                                final byte[][] patdirty = this.patdirty;
                                final int[][] patpix = this.patpix;
                                final boolean latch_mapper = this.latch_mapper;
                                final boolean exgfx_mapper = this.exgfx_mapper;
                                int tileX = (loopy_v & 0x001F);
                                final int tileY = (loopy_v & 0x03E0) >> 5;
                                int name_table_no = ((loopy_v & 0x0C00) >> 10) + 8;
                                byte[] name_table = ppu_map[name_table_no];
                                int nameAddr = loopy_v & 0x03FF;
                                int attribAddr = 0x03C0 + ((tileY >> 2) << 3) + (tileX >> 2);
                                final int pos_y = (tileY & 2) << 1;
                                int attribBits =
                                    (((name_table[attribAddr] & 0xFF) << 2) >> (pos_y | (tileX & 2))) & 0xC;
                                int p = 8 - loopy_x;
                                final int subtile_y = (loopy_v & 0x7000) >> 12;
                                for (int j = 0; j < 33; j++) {
                                    if (BANKSWITCH_PER_TILE && j != 32) {
                                        if ((i & 1) == 0)
                                            cpu_emulate_CPU_cycles(2);
                                        else
                                            cpu_emulate_CPU_cycles(3);
                                    }
                                    if (exgfx_mapper) {
                                        int mmc5_pal = mapper.PPU_Latch_RenderScreen(1, nameAddr & 0x03FF);
                                        if( mmc5_pal != 0) {
                                            attribBits = mmc5_pal;
                                        }
                                    }
                                    final int patternAddr =
                                        bg_pattern_table_addr
                                        + ((name_table[nameAddr] & 0xFF) << 4)
                                        + subtile_y;
                                    final int patno = (patternAddr >> 4) & 0x1FF;
                                    int patno_hi = patno >> 6;
                                    int patno_lo = ((patno & 0x3F) << 3);
                                    if (patdirty[patno_hi][patno & 0x3F] == 0) {
                                        int patno_lo2 = (patno_lo << 1);
                                        for (int v = 0; v < 8; v++) {
                                            int lo = (ppu_map[patno_hi][patno_lo2 + v    ] & 0xFF);
                                            int hi = (ppu_map[patno_hi][patno_lo2 + v + 8] & 0xFF);
                                            int pat;
                                            pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
                                                | ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
                                                | ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
                                                | ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
                                                | ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
                                                | ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
                                                | ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
                                                | ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
                                            patpix[patno_hi][patno_lo | v] =
                                                pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
                                        }
                                        patdirty[patno_hi][patno & 0x3F] = 1;
                                    }
                                    final int pat = patpix[patno_hi][patno_lo | subtile_y];
                                    if (latch_mapper) {
                                        if ((((patternAddr) & 0x0FF0) == 0x0FD0) || (((patternAddr) & 0x0FF0) == 0x0FE0)) {
                                            mapper.PPU_Latch_FDFE(patternAddr);
                                        }
                                    }
                                    attribBits |= 3;
                                    int tmp;
                                    tmp = bg_flags[p    ] = attribBits & pat;
                                    line_buf[p    ] = cur_palette[tmp];
                                    tmp = bg_flags[p + 1] = attribBits & (pat >>  4);
                                    line_buf[p + 1] = cur_palette[tmp];
                                    tmp = bg_flags[p + 2] = attribBits & (pat >>  8);
                                    line_buf[p + 2] = cur_palette[tmp];
                                    tmp = bg_flags[p + 3] = attribBits & (pat >> 12);
                                    line_buf[p + 3] = cur_palette[tmp];
                                    tmp = bg_flags[p + 4] = attribBits & (pat >> 16);
                                    line_buf[p + 4] = cur_palette[tmp];
                                    tmp = bg_flags[p + 5] = attribBits & (pat >> 20);
                                    line_buf[p + 5] = cur_palette[tmp];
                                    tmp = bg_flags[p + 6] = attribBits & (pat >> 24);
                                    line_buf[p + 6] = cur_palette[tmp];
                                    tmp = bg_flags[p + 7] = attribBits & (pat >> 28);
                                    line_buf[p + 7] = cur_palette[tmp];
                                    p += 8;
                                    tileX++;
                                    nameAddr++;
                                    if ((tileX & 0x0001) == 0) {
                                        if ((tileX & 0x0003) == 0) {
                                            if ((tileX & 0x001F) == 0) {
                                                nameAddr = nameAddr - 0x0020;
                                                attribAddr = attribAddr - 0x0008;
                                                name_table = ppu_map[name_table_no ^ 1];
                                                tileX -= 0x0020;
                                            }
                                            attribAddr++;
                                        }
                                        attribBits =
                                            (((name_table[attribAddr] & 0xFF) << 2) >> (pos_y | (tileX & 2)))
                                            & 0xC;
                                    }
                                }
                            }
                            if (BANKSWITCH_PER_TILE) cpu_emulate_CPU_cycles(8);
                            if ((PPU_R1 & 0x10) != 0) {
                                int PPU_R2 = this.PPU_R2;
                                final int[] spr_ram = this.spr_ram;
                                final boolean latch_mapper = this.latch_mapper;
                                //final boolean exgfx_mapper = this.exgfx_mapper;
                                final byte[][] patdirty = this.patdirty;
                                final int[][] patpix = this.patpix;
                                int spritesInScanLine = 0;
                                final int sprHeight = ((PPU_R0 & 0x20) != 0) ? 16 : 8;
                                if (mapper_no != 9) {
                                    if (exgfx_mapper) {
                                        mapper.PPU_Latch_RenderScreen(0, 0);
                                    }
                                    for (int si = 63; si >= 0; si--) {
                                        final int s = (si << 2);
                                        final int lineOfSprite = i - (spr_ram[s] + 1);
                                        if (lineOfSprite < 0 || lineOfSprite >= sprHeight)
                                            continue;
                                        spritesInScanLine++;
                                        int tileIndex = spr_ram[s + 1];
                                        final int sprAttribs = spr_ram[s + 2];
                                        final int sprVFlip = (sprAttribs & 0x80);
                                        final int sprHFlip = (sprAttribs & 0x40);
                                        final int sprBGPriority = (sprAttribs & 0x20);
                                        int y = lineOfSprite;
                                        int p = spr_ram[s + 3] + 8;
                                        if (sprVFlip != 0)
                                            y = (sprHeight - 1) - y;
                                        if (latch_mapper) {
                                            if ((((tileIndex) & 0x00FF) == 0x00FD) || (((tileIndex) & 0x00FF) == 0x00FE)) {
                                                mapper.PPU_Latch_FDFE(tileIndex << 4);
                                            }
                                        }
                                        if (sprHeight == 16) {
                                            if ((tileIndex & 0x01) != 0) {
                                                tileIndex += 0x100;
                                                if (y < 8)
                                                    tileIndex--;
                                            } else {
                                                if (y >= 8)
                                                    tileIndex++;
                                            }
                                        } else {
                                            tileIndex += spr_pattern_table_addr >> 4;
                                        }
                                        int patno_hi = tileIndex >> 6;
                                        int patno_lo = ((tileIndex & 0x3F) << 3);
                                        if (patdirty[patno_hi][tileIndex & 0x3F] == 0) {
                                            int patno_lo2 = (patno_lo << 1);
                                            for (int v = 0; v < 8; v++) {
                                                int lo = (ppu_map[patno_hi][patno_lo2 + v    ] & 0xFF);
                                                int hi = (ppu_map[patno_hi][patno_lo2 + v + 8] & 0xFF);
                                                int pat;
                                                pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
                                                    | ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
                                                    | ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
                                                    | ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
                                                    | ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
                                                    | ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
                                                    | ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
                                                    | ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
                                                patpix[patno_hi][patno_lo | v] =
                                                    pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
                                            }
                                            patdirty[patno_hi][tileIndex & 0x3F] = 1;
                                        }
                                        final int pat = patpix[patno_hi][patno_lo | (y & 7)];
                                        final int colhi = ((sprAttribs & 3) << 2) | 0x10;
                                        if (si == 0) {
                                            sprite_hit[lineOfSprite] = 0;
                                            if (pat != 0) {
                                                if (bg_flags[p + 0] != 0 || bg_flags[p + 1] != 0 ||
                                                    bg_flags[p + 2] != 0 || bg_flags[p + 3] != 0 ||
                                                    bg_flags[p + 4] != 0 || bg_flags[p + 5] != 0 ||
                                                    bg_flags[p + 6] != 0 || bg_flags[p + 7] != 0) {
                                                        PPU_R2 |= 0x40;
                                                        sprite_hit[lineOfSprite] = 1;
                                                }
                                            }
                                        }
                                        {
                                            int c;
                                            if (sprBGPriority != 0)
                                                if (sprHFlip != 0) {
                                                    if (bg_flags[p    ] == 0 && (c = (pat >> 28) & 3) != 0)
                                                        line_buf[p    ] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 1] == 0  && (c = (pat >> 24) & 3) != 0)
                                                        line_buf[p + 1] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 2] == 0  && (c = (pat >> 20) & 3) != 0)
                                                        line_buf[p + 2] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 3] == 0  && (c = (pat >> 16) & 3) != 0)
                                                        line_buf[p + 3] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 4] == 0  && (c = (pat >> 12) & 3) != 0)
                                                        line_buf[p + 4] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 5] == 0  && (c = (pat >>  8) & 3) != 0)
                                                        line_buf[p + 5] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 6] == 0  && (c = (pat >>  4) & 3) != 0)
                                                        line_buf[p + 6] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 7] == 0  && (c = pat & 3) != 0)
                                                        line_buf[p + 7] = cur_palette[colhi | c];
                                                } else {
                                                    if (bg_flags[p    ] == 0 && (c = pat & 3) != 0)
                                                        line_buf[p    ] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 1] == 0  && (c = (pat >>  4) & 3) != 0)
                                                        line_buf[p + 1] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 2] == 0  && (c = (pat >>  8) & 3) != 0)
                                                        line_buf[p + 2] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 3] == 0  && (c = (pat >> 12) & 3) != 0)
                                                        line_buf[p + 3] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 4] == 0  && (c = (pat >> 16) & 3) != 0)
                                                        line_buf[p + 4] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 5] == 0  && (c = (pat >> 20) & 3) != 0)
                                                        line_buf[p + 5] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 6] == 0  && (c = (pat >> 24) & 3) != 0)
                                                        line_buf[p + 6] = cur_palette[colhi | c];
                                                    if (bg_flags[p + 7] == 0  && (c = (pat >> 28) & 3) != 0)
                                                        line_buf[p + 7] = cur_palette[colhi | c];
                                                }
                                            else if (sprHFlip != 0) {
                                                if ((c = (pat >> 28) & 3) != 0)
                                                    line_buf[p    ] = cur_palette[colhi | c];
                                                if ((c = (pat >> 24) & 3) != 0)
                                                    line_buf[p + 1] = cur_palette[colhi | c];
                                                if ((c = (pat >> 20) & 3) != 0)
                                                    line_buf[p + 2] = cur_palette[colhi | c];
                                                if ((c = (pat >> 16) & 3) != 0)
                                                    line_buf[p + 3] = cur_palette[colhi | c];
                                                if ((c = (pat >> 12) & 3) != 0)
                                                    line_buf[p + 4] = cur_palette[colhi | c];
                                                if ((c = (pat >> 8) & 3) != 0)
                                                    line_buf[p + 5] = cur_palette[colhi | c];
                                                if ((c = (pat >> 4) & 3) != 0)
                                                    line_buf[p + 6] = cur_palette[colhi | c];
                                                if ((c = pat & 3) != 0)
                                                    line_buf[p + 7] = cur_palette[colhi | c];
                                            } else {
                                                if ((c = pat & 3) != 0)
                                                    line_buf[p    ] = cur_palette[colhi | c];
                                                if ((c = (pat >> 4) & 3) != 0)
                                                    line_buf[p + 1] = cur_palette[colhi | c];
                                                if ((c = (pat >> 8) & 3) != 0)
                                                    line_buf[p + 2] = cur_palette[colhi | c];
                                                if ((c = (pat >> 12) & 3) != 0)
                                                    line_buf[p + 3] = cur_palette[colhi | c];
                                                if ((c = (pat >> 16) & 3) != 0)
                                                    line_buf[p + 4] = cur_palette[colhi | c];
                                                if ((c = (pat >> 20) & 3) != 0)
                                                    line_buf[p + 5] = cur_palette[colhi | c];
                                                if ((c = (pat >> 24) & 3) != 0)
                                                    line_buf[p + 6] = cur_palette[colhi | c];
                                                if ((c = (pat >> 28) & 3) != 0)
                                                    line_buf[p + 7] = cur_palette[colhi | c];
                                            }
                                        }
                                    }
                                } else {
                                    // FOR MAPPER #9
                                    System.arraycopy(zero_array, 0, fg_flags, 0, 280);
                                    for (int si = 0; si < 64; si++) {
                                        final int s = (si << 2);
                                        final int sprY = spr_ram[s] + 1;
                                        final int lineOfSprite = i - sprY;
                                        if (lineOfSprite < 0 || lineOfSprite >= sprHeight)
                                            continue;
                                        int tileIndex = spr_ram[s + 1];
                                        final int sprAttribs = spr_ram[s + 2];
                                        final int sprVFlip = (sprAttribs & 0x80);
                                        final int sprHFlip = (sprAttribs & 0x40);
                                        final int sprBGPriority = (sprAttribs & 0x20);
                                        int y = lineOfSprite;
                                        int p = spr_ram[s + 3] + 8;
                                        if (sprVFlip != 0)
                                            y = (sprHeight - 1) - y;
                                        if ((((tileIndex) & 0x00FF) == 0x00FD) || (((tileIndex) & 0x00FF) == 0x00FE)) {
                                            mapper.PPU_Latch_FDFE(tileIndex << 4);
                                        }
                                        if (sprHeight == 16) {
                                            if ((tileIndex & 0x01) != 0) {
                                                tileIndex += 0x100;
                                                if (y < 8)
                                                    tileIndex--;
                                            } else {
                                                if (y >= 8)
                                                    tileIndex++;
                                            }
                                        } else {
                                            tileIndex += spr_pattern_table_addr >> 4;
                                        }
                                        int patno_hi = tileIndex >> 6;
                                        int patno_lo = ((tileIndex & 0x3F) << 3);
                                        if (patdirty[patno_hi][tileIndex & 0x3F] == 0) {
                                            int patno_lo2 = (patno_lo << 1);
                                            for (int v = 0; v < 8; v++) {
                                                int lo = (ppu_map[patno_hi][patno_lo2 + v    ] & 0xFF);
                                                int hi = (ppu_map[patno_hi][patno_lo2 + v + 8] & 0xFF);
                                                int pat;
                                                pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
                                                    | ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
                                                    | ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
                                                    | ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
                                                    | ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
                                                    | ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
                                                    | ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
                                                    | ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
                                                patpix[patno_hi][patno_lo | v] =
                                                    pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
                                            }
                                            patdirty[patno_hi][tileIndex & 0x3F] = 1;
                                        }
                                        final int pat = patpix[patno_hi][patno_lo | (y & 7)];
                                        final int colhi = ((sprAttribs & 3) << 2) | 0x10;
                                        if (si == 0) {
                                            sprite_hit[lineOfSprite] = 0;
                                            if (pat != 0) {
                                                if (bg_flags[p + 0] != 0 || bg_flags[p + 1] != 0 ||
                                                    bg_flags[p + 2] != 0 || bg_flags[p + 3] != 0 ||
                                                    bg_flags[p + 4] != 0 || bg_flags[p + 5] != 0 ||
                                                    bg_flags[p + 6] != 0 || bg_flags[p + 7] != 0) {
                                                        PPU_R2 |= 0x40;
                                                        sprite_hit[lineOfSprite] = 1;
                                                }
                                            }
                                        }
                                        {
                                            int c;
                                            if (sprBGPriority != 0) {
                                                if (sprHFlip != 0) {
                                                    if (fg_flags[p    ] == 0 && bg_flags[p    ] == 0 && (c = (pat >> 28) & 3) != 0) {
                                                        line_buf[p    ] = cur_palette[colhi | c];
                                                        fg_flags[p    ] = 1;
                                                    }
                                                    if (fg_flags[p + 1] == 0 && bg_flags[p + 1] == 0  && (c = (pat >> 24) & 3) != 0) {
                                                        line_buf[p + 1] = cur_palette[colhi | c];
                                                        fg_flags[p + 1] = 1;
                                                    }
                                                    if (fg_flags[p + 2] == 0 && bg_flags[p + 2] == 0  && (c = (pat >> 20) & 3) != 0) {
                                                        line_buf[p + 2] = cur_palette[colhi | c];
                                                        fg_flags[p + 2] = 1;
                                                    }
                                                    if (fg_flags[p + 3] == 0 && bg_flags[p + 3] == 0  && (c = (pat >> 16) & 3) != 0) {
                                                        line_buf[p + 3] = cur_palette[colhi | c];
                                                        fg_flags[p + 3] = 1;
                                                    }
                                                    if (fg_flags[p + 4] == 0 && bg_flags[p + 4] == 0  && (c = (pat >> 12) & 3) != 0) {
                                                        line_buf[p + 4] = cur_palette[colhi | c];
                                                        fg_flags[p + 4] = 1;
                                                    }
                                                    if (fg_flags[p + 5] == 0 && bg_flags[p + 5] == 0  && (c = (pat >>  8) & 3) != 0) {
                                                        line_buf[p + 5] = cur_palette[colhi | c];
                                                        fg_flags[p + 5] = 1;
                                                    }
                                                    if (fg_flags[p + 6] == 0 && bg_flags[p + 6] == 0  && (c = (pat >>  4) & 3) != 0) {
                                                        line_buf[p + 6] = cur_palette[colhi | c];
                                                        fg_flags[p + 6] = 1;
                                                    }
                                                    if (fg_flags[p + 7] == 0 && bg_flags[p + 7] == 0  && (c = pat & 3) != 0) {
                                                        line_buf[p + 7] = cur_palette[colhi | c];
                                                        fg_flags[p + 7] = 1;
                                                    }
                                                } else {
                                                    if (fg_flags[p    ] == 0 && bg_flags[p    ] == 0 && (c = pat & 3) != 0) {
                                                        line_buf[p    ] = cur_palette[colhi | c];
                                                        fg_flags[p    ] = 1;
                                                    }
                                                    if (fg_flags[p + 1] == 0 && bg_flags[p + 1] == 0  && (c = (pat >>  4) & 3) != 0) {
                                                        line_buf[p + 1] = cur_palette[colhi | c];
                                                        fg_flags[p + 1] = 1;
                                                    }
                                                    if (fg_flags[p + 2] == 0 && bg_flags[p + 2] == 0  && (c = (pat >>  8) & 3) != 0) {
                                                        line_buf[p + 2] = cur_palette[colhi | c];
                                                        fg_flags[p + 2] = 1;
                                                    }
                                                    if (fg_flags[p + 3] == 0 && bg_flags[p + 3] == 0  && (c = (pat >> 12) & 3) != 0) {
                                                        line_buf[p + 3] = cur_palette[colhi | c];
                                                        fg_flags[p + 3] = 1;
                                                    }
                                                    if (fg_flags[p + 4] == 0 && bg_flags[p + 4] == 0  && (c = (pat >> 16) & 3) != 0) {
                                                        line_buf[p + 4] = cur_palette[colhi | c];
                                                        fg_flags[p + 4] = 1;
                                                    }
                                                    if (fg_flags[p + 5] == 0 && bg_flags[p + 5] == 0  && (c = (pat >> 20) & 3) != 0) {
                                                        line_buf[p + 5] = cur_palette[colhi | c];
                                                        fg_flags[p + 5] = 1;
                                                    }
                                                    if (fg_flags[p + 6] == 0 && bg_flags[p + 6] == 0  && (c = (pat >> 24) & 3) != 0) {
                                                        line_buf[p + 6] = cur_palette[colhi | c];
                                                        fg_flags[p + 6] = 1;
                                                    }
                                                    if (fg_flags[p + 7] == 0 && bg_flags[p + 7] == 0  && (c = (pat >> 28) & 3) != 0) {
                                                        line_buf[p + 7] = cur_palette[colhi | c];
                                                        fg_flags[p + 7] = 1;
                                                    }
                                                }
                                            } else if (sprHFlip != 0) {
                                                if (fg_flags[p    ] == 0 && (c = (pat >> 28) & 3) != 0) {
                                                    line_buf[p    ] = cur_palette[colhi | c];
                                                    fg_flags[p    ] = 1;
                                                }
                                                if (fg_flags[p + 1] == 0 && (c = (pat >> 24) & 3) != 0) {
                                                    line_buf[p + 1] = cur_palette[colhi | c];
                                                    fg_flags[p + 1] = 1;
                                                }
                                                if (fg_flags[p + 2] == 0 && (c = (pat >> 20) & 3) != 0) {
                                                    line_buf[p + 2] = cur_palette[colhi | c];
                                                    fg_flags[p + 2] = 1;
                                                }
                                                if (fg_flags[p + 3] == 0 && (c = (pat >> 16) & 3) != 0) {
                                                    line_buf[p + 3] = cur_palette[colhi | c];
                                                    fg_flags[p + 3] = 1;
                                                }
                                                if (fg_flags[p + 4] == 0 && (c = (pat >> 12) & 3) != 0) {
                                                    line_buf[p + 4] = cur_palette[colhi | c];
                                                    fg_flags[p + 4] = 1;
                                                }
                                                if (fg_flags[p + 5] == 0 && (c = (pat >> 8) & 3) != 0) {
                                                    line_buf[p + 5] = cur_palette[colhi | c];
                                                    fg_flags[p + 5] = 1;
                                                }
                                                if (fg_flags[p + 6] == 0 && (c = (pat >> 4) & 3) != 0) {
                                                    line_buf[p + 6] = cur_palette[colhi | c];
                                                    fg_flags[p + 6] = 1;
                                                }
                                                if (fg_flags[p + 7] == 0 && (c = pat & 3) != 0) {
                                                    line_buf[p + 7] = cur_palette[colhi | c];
                                                    fg_flags[p + 7] = 1;
                                                }
                                            } else {
                                                if (fg_flags[p    ] == 0 && (c = pat & 3) != 0) {
                                                    line_buf[p    ] = cur_palette[colhi | c];
                                                    fg_flags[p    ] = 1;
                                                }
                                                if (fg_flags[p + 1] == 0 && (c = (pat >> 4) & 3) != 0) {
                                                    line_buf[p + 1] = cur_palette[colhi | c];
                                                    fg_flags[p + 1] = 1;
                                                }
                                                if (fg_flags[p + 2] == 0 && (c = (pat >> 8) & 3) != 0) {
                                                    line_buf[p + 2] = cur_palette[colhi | c];
                                                    fg_flags[p + 2] = 1;
                                                }
                                                if (fg_flags[p + 3] == 0 && (c = (pat >> 12) & 3) != 0) {
                                                    line_buf[p + 3] = cur_palette[colhi | c];
                                                    fg_flags[p + 3] = 1;
                                                }
                                                if (fg_flags[p + 4] == 0 && (c = (pat >> 16) & 3) != 0) {
                                                    line_buf[p + 4] = cur_palette[colhi | c];
                                                    fg_flags[p + 4] = 1;
                                                }
                                                if (fg_flags[p + 5] == 0 && (c = (pat >> 20) & 3) != 0) {
                                                    line_buf[p + 5] = cur_palette[colhi | c];
                                                    fg_flags[p + 5] = 1;
                                                }
                                                if (fg_flags[p + 6] == 0 && (c = (pat >> 24) & 3) != 0) {
                                                    line_buf[p + 6] = cur_palette[colhi | c];
                                                    fg_flags[p + 6] = 1;
                                                }
                                                if (fg_flags[p + 7] == 0 && (c = (pat >> 28) & 3) != 0) {
                                                    line_buf[p + 7] = cur_palette[colhi | c];
                                                    fg_flags[p + 7] = 1;
                                                }
                                            }
                                        }
                                        if(++spritesInScanLine == 8) break;
                                    }
                                }
                                if (spritesInScanLine >= 8)
                                    PPU_R2 |= 0x20;
                                else
                                    PPU_R2 &= 0xDF;
                                this.PPU_R2 = PPU_R2;
                            }
                            if ((loopy_v & 0x7000) == 0x7000) {
                                loopy_v &= 0x8FFF;
                                if ((loopy_v & 0x03E0) == 0x03A0)
                                    loopy_v = (loopy_v ^ 0x0800) & 0xFC1F;
                                else if ((loopy_v & 0x03E0) == 0x03E0)
                                    loopy_v &= 0xFC1F;
                                else
                                    loopy_v += 0x0020;
                            } else
                                loopy_v += 0x1000;
                        }
                        if (BANKSWITCH_PER_TILE) {
                            cpu_emulate_CPU_cycles(cyc_before_hsync);
                            mapper.HSync(i);
                            cpu_emulate_CPU_cycles(cyc_after_hsync);
                            if (i == 0) {
                                cpu_emulate_CPU_cycles(88 + cyc_before_hsync);
                                mapper.HSync(i);
                                cpu_emulate_CPU_cycles(cyc_after_hsync);
                            }
                        }
                        if (render_frame && i >= 8 && i < 232) {
                            System.arraycopy(line_buf, 8, fb, ((i - 8) << 8), 256);
                        }
                    } else {
                        // !render_frame
                        cpu_emulate_CPU_cycles(cyc_normal);
                        mapper.HSync(i);
                        if ((PPU_R1 & 0x18) != 0) {
                            loopy_v = (loopy_v & 0xFBE0) | (loopy_t & 0x041F);
                            if ((PPU_R1 & 0x10) != 0) {
                                final int sprHeight = ((PPU_R0 & 0x20) != 0) ? 16 : 8;
                                final int sprY = spr_ram[0] + 1;
                                final int lineOfSprite = i - sprY;
                                if (lineOfSprite >= 0 && lineOfSprite < sprHeight)
                                    if (sprite_hit[lineOfSprite] != 0)
                                        PPU_R2 |= 0x40;
                            }
                            if ((loopy_v & 0x7000) == 0x7000) {
                                loopy_v &= 0x8FFF;
                                if ((loopy_v & 0x03E0) == 0x03A0)
                                    loopy_v = (loopy_v ^ 0x0800) & 0xFC1F;
                                else if ((loopy_v & 0x03E0) == 0x03E0)
                                    loopy_v &= 0xFC1F;
                                else
                                    loopy_v += 0x0020;
                            } else
                                loopy_v += 0x1000;
                        }
                    }
                    if((i & cyc_correct) != 0) cyc_pending += 1;
                }
                if ((frame_irq_enabled & 0xC0) == 0) {
                    if ((P & 4) == 0) {
                        P &= ~0x10;
                        RAM[S + 0x100] = (byte) (PC >> 8);
                        S = (S - 1) & 0xFF;
                        RAM[S + 0x100] = (byte) PC;
                        S = (S - 1) & 0xFF;
                        RAM[S + 0x100] = (byte) P;
                        S = (S - 1) & 0xFF;
                        PC = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
                        P |= 4;
                        cyc_pending -= 7;
                    } else {
                        int_pending = 1;
                    }
                }
                for (int i = 240; i <= 261; i++) {
                    if (i == 261)
                        PPU_R2 &= 0x3F;
                    if (i == 241) {
                        PPU_R2 |= 0x80;
                        mapper.VSync();
                        cpu_emulate_CPU_cycles(cyc_before_nmi);
                        if ((PPU_R0 & 0x80) != 0) {
                            P &= ~0x10;
                            RAM[S + 0x100] = (byte) (PC >> 8);
                            S = (S - 1) & 0xFF;
                            RAM[S + 0x100] = (byte) PC;
                            S = (S - 1) & 0xFF;
                            RAM[S + 0x100] = (byte) P;
                            S = (S - 1) & 0xFF;
                            PC = mem_read(0xFFFA) | (mem_read(0xFFFB) << 8);
                            P |= 4;
                            cyc_pending -= 7;
                        }
                        cpu_emulate_CPU_cycles(cyc_after_nmi);
                        mapper.HSync(i);
                        if((i & cyc_correct) != 0) cyc_pending += 1;
                        continue;
                    }
                    cpu_emulate_CPU_cycles(cyc_normal);
                    mapper.HSync(i);
                    if((i & cyc_correct) != 0) cyc_pending += 1;
                }

                audio.apu_emulate_frame();

                if (render_frame) gui.endFrame();
            } catch (Throwable th) {
            }
        }
    }
}

