import java.io.*;

import com.nttdocomo.ui.Graphics;

/*=============================================================================
	NESRANX
  =============================================================================*/
public final class NESCore_SB extends NESCore {

	private			int			CYCLE_PER_LINE = 114;
	private			int			CYCLES_BEFORE_NMI = 1;
	private			boolean		BANKSWITCH_PER_TILE;
	private			boolean		has_SRAM;

	// =========================
	// RXgN^
	//==========================
	public NESCore_SB(NESCanvas _gui, PSGi _psgi) {
		gui = _gui;
        psgi = _psgi;
		crc_table = new int[256];
		for( int n = 0 ; n < 256 ; n++ ){
			int c = n;
			for( int k = 8 ; --k >= 0; ){
				if( (c & 1) != 0 ){
					c = 0xedb88320 ^ (c >>> 1);
				}else{
					c = c >>> 1;
				}
			}
			crc_table[n] = c;
		}

		// TEh
		switch (gui.getApuNum()) {
			case 1 :
				audio = new NESApu1(this, gui, psgi);
				break;
			case 2 :
				audio = new NESApu2(this, gui, psgi);
				break;
			case 3 :
				audio = new NESApu_NP2(this, gui, psgi);
				break;
			default :
				break;
		}
	}

	// =========================
	// ROM[h
	//==========================
	public final int load_rom(InputStream is, boolean from_sp) throws IOException {
		int b0, b1, b2, b3, b6, b7, iNum16kBank = 0, iNum8kBank = 0;
		b6 = b7 = 0;

		b0 = is.read();
		b1 = is.read();
		b2 = is.read();
		b3 = is.read();
		if( (b0 == 'N') && (b1 == 'E') && (b2 == 'S') && (b3 == 0x1A) ){	// t@~R̎
			num_16k_ROM_banks = is.read();
			num_8k_VROM_banks = is.read();
			num_DISK_banks = 0;
			b6 = is.read();
			b7 = is.read();
			mirroring =
				(b6 & 1) != 0
				? 1
				: (b6 & 8) != 0
				? 2
				: 0;
			has_SRAM = (b6 & 2) != 0; 								// ĂɂȂ˂BBB
			mapper_no = ((b6 >> 4) & 0x0F) | (b7 & 0xF0);
			for( int i = 0 ; i < 8 ; i++ ){
				if( is.read() != 0 ){
					mapper_no &= 0x0F;
				}
			}
			if( (b6 & 4) != 0 ){
				return -2;
			}
			
			iNum16kBank = num_16k_ROM_banks << 1;
			iNum8kBank  = num_8k_VROM_banks << 3;
			
			ROM = load_blocks(is, iNum16kBank, 0x2000);
			int c = ~crc;
			for( int i = 0 ; i < iNum16kBank ; i++ ){
				for( int j = 0 ; j < 0x2000 ; j++ ){
					c = crc_table[(c ^ ROM[i][j]) & 0xff] ^ (c >>> 8);
				}
			}
			crc = ~c;
			System.out.println("CRC=" + Integer.toHexString(crc));
			DISK = null;
			BIOS = null;
			disk_system = false;
			if( mapper_no == 0x14 ){
				return -3;
			}
		}else if( (b0 == 'F') && (b1 == 'D') && (b2 == 'S') && (b3 == 0x1A) ){	// fBXNVXe̎
			num_16k_ROM_banks = 0;
			num_8k_VROM_banks = 0;
			num_DISK_banks = is.read();
			mirroring = 0;
			mapper_no = 0x14;
			CYCLE_PER_LINE = 113;
			for( int i = 5 ; i < 0x10 ; i++ ){
				is.read();
			}
			
			ROM  = null;
			DISK = load_disk_blocks(is, num_DISK_banks);
			BIOS = load_bios(from_sp);
			if( BIOS == null ){
				return -4;
			}
			disk_system = true;
		}else{
			return -1;
		}
		
		BANKSWITCH_PER_TILE = false;
		
		System.out.println("pb`O ROM=" + num_16k_ROM_banks + " VROM=" + num_8k_VROM_banks + " mapper_no=" + mapper_no +
											" SRAM=" + has_SRAM + " mirror=" + mirroring + " CYCLE_PER_LINE=" + CYCLE_PER_LINE);
		rom_patch( crc, mapper_no);		// ROMpb`
		System.out.println("pb` ROM=" + num_16k_ROM_banks + " VROM=" + num_8k_VROM_banks + " mapper_no=" + mapper_no +
											" SRAM=" + has_SRAM + " mirror=" + mirroring + " CYCLE_PER_LINE=" + CYCLE_PER_LINE);

		if( (mapper = Mapper.createMapper(mapper_no, this)) == null ){
			return -3;
		}
		if( num_8k_VROM_banks > 0 ){
			VROM = load_blocks( is, iNum8kBank, 0x0400);
			vram_write_protect = (crc != 0xadb47286);
			VROM_CACHE = new int[iNum8kBank][512];
			for( int i = 0 ; i < iNum8kBank ; i++ ){
				for( int j = 0 ; j < 64 ; j++ ){
					for( int k = 0 ; k < 8 ; k++ ){
						int lo = VROM[i][(j << 4) + k	 ] & 0xFF;
						int hi = VROM[i][(j << 4) + k + 8] & 0xFF;
						int pat;
						pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
							| ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
							| ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
							| ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
							| ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
							| ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
							| ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
							| ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
						VROM_CACHE[i][(j << 3) + k]= pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
					}
				}
			}
		}
	// INIT or LOAD SRAM
		int c = load_sram();
		if(c != 0)
		for( int i = 0 ; i < SRAM_SIZE ; i++ ){
			SRAM[i] = (byte) 0xFF;
		}
		// SAVE ROM@SP
		if(sp_mode && !from_sp) {
			if(!disk_system) {
				gui.writeSPROM(num_16k_ROM_banks, num_8k_VROM_banks, b6, b7, ROM, VROM);
			} else {
				gui.writeSPFD(num_DISK_banks, DISK);
				gui.writeSPBIOS(BIOS);
			}
		}
		
		// SET DUMMY CHARCTER ROM
		if( mapper_no == 185 ){
			DUMMY_VROM = new byte[0x400];
			DUMMY_VROM_CACHE = new int[512];
			for( int i = 0 ; i < 0x400 ; i++ ){
				DUMMY_VROM[i] = (byte)0xFF;
			}
			for( int i = 0 ; i < 512 ; i++ ){
				DUMMY_VROM_CACHE[i] = (int)0xFFFFFFFF;
			}
		}
		return 0;
	}

	// =========================
	// Zbg
	//==========================
	public final int reset() {
		// reset memory
		for( int i = 0 ; i < 8 ; i++ ){
			cpu_map[i] = null;
		}
		
		if( crc == 0x29401686 ){	// ݂Ȃ̂V
			for( int i = 0 ; i < 0x0800 ; i++ ){
				RAM[i] = (byte)0xFF;
			}
		}else{
			for( int i = 0 ; i < 0x0800 ; i++ ){
				RAM[i] = 0x00;
			}
		}
		
		cpu_map[0] = RAM;
		cpu_map[3] = SRAM;
		
		// reset ppu
		for( int i = 0 ; i < 12 ; i++ ){
			for( int j = 0 ; j < 0x400 ; j++ ){
				VRAM[i][j] = (byte)0x00;
			}
		}
		for( int i = 0 ; i < 65 ; i++ ){
			for( int j = 0; j < 8 ; j++ ){
				CACHE_HIT[j][i] = 0;
			}
			CACHE_HIT[8][i] = 1;
		}
		for( int i = 0 ; i < 8 ; i++ ){
			ppu_map[i] = VRAM[i];
			patpix[i] = VRAM_CACHE[i];
			patdirty[i] = CACHE_HIT[i];
		}
		set_mirroring(mirroring);
		PPU_R0 = 0;
		PPU_R1 = 0;
		PPU_R2 = 0;
		bg_pattern_table_addr = 0;
		spr_pattern_table_addr = 0;
		ppu_addr_inc = 0;
		toggle_2005_2006 = false;
		loopy_v = 0;
		loopy_t = 0;
		loopy_x = 0;
		spr_ram_rw_ptr = 0;
		read_2007_buffer = 0;
		frame_irq_enabled = 0xFF;
		for( int i = 0 ; i < 256 ; i++ ){
			spr_ram[i] = 0;
		}
		for( int i = 0 ; i < 32 ; i++ ){
			pal_ram[i] = 0;
		}
		// reset mmc
		try{
			mapper.Reset();
		} catch (Exception e) {
			return -1;
		}
		// reset cpu
		A = X = Y = 0;
		S = 0xFF;
		P = 0x26;
		PC = mem_read(0xFFFC) | (mem_read(0xFFFD) << 8);
		int_pending = 0;
		cyc_pending = -6;
		switch (gui.getApuNum()) {
			case 1 :
				audio = new NESApu1(this, gui, psgi);
				break;
			case 2 :
				audio = new NESApu2(this, gui, psgi);
				break;
			case 3 :
				audio = new NESApu_NP2(this, gui, psgi);
				break;
			default :
				break;
		}
		audio.apu_setmode(gui.soundMode);
		audio.apu_setting();
		audio.apu_reset();
		return 0;
	}
	
	// =========================
	// ^XNC
	//==========================
	public final void run() {
		final int[] line_buf = new int[280];
		final int[] bg_flags = new int[280];
		final int[] fg_flags = new int[280]; //FOR Mapper #9
		final byte[] sprite_hit = new byte[16];
		exist = true;
		for (;;) {
			super.run();
			switch (request) {
				case 2 :
					request = 0;
					exist = false;
					return;
				case 7 :
					continue;
				default :
				break;
			}
			request = 0;
			emulate_frame();	// G~[gC
		}
	}

	private int			current_frame_line;
	private	boolean		render_frame;

	// =========================
	// G~[gC
	//==========================
	private final void emulate_frame() {
		try {
			int i;

			// t[Jn
			current_frame_line = 0;
			if( (PPU_R1 & 0x18) != 0 ){
				loopy_v = loopy_t;
			}
			render_frame = gui.startFrame();
			// MOTHER̎
			if( BANKSWITCH_PER_TILE ){
				for( i = 0 ; i < 240 ; i++ ){
					if( (PPU_R1 & 0x18) != 0 ){
						loopy_v = (loopy_v & 0xFBE0) | (loopy_t & 0x041F);
						if( render_frame ){
							render_background();	// obNOEh`
							cpu_emulate_CPU_cycles(24);
						}else{
							cpu_emulate_CPU_cycles(87);
						}
						
						if( (PPU_R1 & 0x10) != 0 ){
							render_spr();			// XvCg`
						}
						
						// ̍s
						if( (loopy_v & 0x7000) == 0x7000 ){
							loopy_v &= 0x8FFF;
							if( (loopy_v & 0x03E0) == 0x03A0 ){
								loopy_v = (loopy_v ^ 0x0800) & 0xFC1F;
							}else if( (loopy_v & 0x03E0) == 0x03E0 ){
								loopy_v &= 0xFC1F;
							}else{
								loopy_v += 0x0020;
							}
						} else{
							loopy_v += 0x1000;
						}
					}
					current_frame_line++;
					
					cpu_emulate_CPU_cycles(13);
					mapper.HSync(i);
					cpu_emulate_CPU_cycles(14);
					if( i == 0 ){
						cpu_emulate_CPU_cycles(100);
						mapper.HSync(i);
						cpu_emulate_CPU_cycles(14);
					}
				}
			
			// MOTHERł͂Ȃ
			}else{
				for( i = 0 ; i < 240 ; i++ ){
					cpu_emulate_CPU_cycles(CYCLE_PER_LINE);
					mapper.HSync(i);
					if( (PPU_R1 & 0x18) != 0 ){
						loopy_v = (loopy_v & 0xFBE0) | (loopy_t & 0x041F);
						if( render_frame ){
							render_background();	// obNOEh`
						}
						if( (PPU_R1 & 0x10) != 0 ){
							render_spr();			// XvCg`
						}
						
						// ̍s
						if( (loopy_v & 0x7000) == 0x7000 ){
							loopy_v &= 0x8FFF;
							if( (loopy_v & 0x03E0) == 0x03A0 ){
								loopy_v = (loopy_v ^ 0x0800) & 0xFC1F;
							}else if( (loopy_v & 0x03E0) == 0x03E0 ){
								loopy_v &= 0xFC1F;
							}else{
								loopy_v += 0x0020;
							}
						} else{
							loopy_v += 0x1000;
						}
					}
					current_frame_line++;
				}
			}
			
			
			// ʊOp
			if( (frame_irq_enabled & 0xC0) == 0 ){
				if( (P & 4) == 0 ){
					P &= ~0x10;
					RAM[S + 0x100] = (byte) (PC >> 8);
					S = (S - 1) & 0xFF;
					RAM[S + 0x100] = (byte) PC;
					S = (S - 1) & 0xFF;
					RAM[S + 0x100] = (byte) P;
					S = (S - 1) & 0xFF;
					PC = mem_read(0xFFFE) | (mem_read(0xFFFF) << 8);
					P |= 4;
					cyc_pending -= 7;
				}else{
					int_pending = 1;
				}
			}
			
			for( i = 240 ; i <= 261 ; i++ ){
				if( i == 261 ){
					PPU_R2 &= 0x3F;
				}else if( i == 241 ){
					PPU_R2 |= 0x80;
					mapper.VSync();
					cpu_emulate_CPU_cycles(CYCLES_BEFORE_NMI);
					if( (PPU_R0 & 0x80) != 0 ){
						P &= ~0x10;
						RAM[S + 0x100] = (byte) (PC >> 8);
						S = (S - 1) & 0xFF;
						RAM[S + 0x100] = (byte) PC;
						S = (S - 1) & 0xFF;
						RAM[S + 0x100] = (byte) P;
						S = (S - 1) & 0xFF;
						PC = mem_read(0xFFFA) | (mem_read(0xFFFB) << 8);
						P |= 4;
						cyc_pending -= 7;
					}
					cpu_emulate_CPU_cycles(CYCLE_PER_LINE - CYCLES_BEFORE_NMI);
					mapper.HSync(i);
					continue;
				}
				cpu_emulate_CPU_cycles(CYCLE_PER_LINE);
				mapper.HSync(i);
			}
			// TEhĐ
			audio.apu_emulate_frame();
			
			// t[I
			if( render_frame ){
				gui.endFrame();
			}
		}catch( Throwable th ){
		}
	}

	// =========================
	// obOOEh쐬
	//==========================
	private final void render_background() {
		if( (PPU_R1 & 8) == 0 ){
//			System.arraycopy(zero_array, 0, line_buf, 0, 272);
		}else{
			final	int			bg_pattern_table_addr = this.bg_pattern_table_addr;
			final	int			loopy_v = this.loopy_v;
			final	byte[][]	patdirty = this.patdirty;
			final	int[][]		patpix = this.patpix;
			final	boolean		latch_mapper = this.latch_mapper;
			final	boolean		exgfx_mapper = this.exgfx_mapper;
					int			tileX = (loopy_v & 0x001F);
			final	int			tileY = (loopy_v & 0x03E0) >> 5;
					int			name_table_no = ((loopy_v & 0x0C00) >> 10) + 8;
					byte[]		name_table = ppu_map[name_table_no];
					int			nameAddr = loopy_v & 0x03FF;
					int			attribAddr = 0x03C0 + ((tileY >> 2) << 3) + (tileX >> 2);
			final	int			pos_y = (tileY & 2) << 1;
					int			attribBits = (((name_table[attribAddr] & 0xFF) << 2) >> (pos_y | (tileX & 2))) & 0xC;
					int			p = -loopy_x + (current_frame_line-8)*256 + 8;	// +8 [͕`悵Ȃł
			final	int			subtile_y = (loopy_v & 0x7000) >> 12;
			
			tileX++;
			nameAddr++;
			if( (tileX & 0x0001) == 0 ){
				if( (tileX & 0x0003) == 0 ){
					if( (tileX & 0x001F) == 0 ){
						nameAddr = nameAddr - 0x0020;
						attribAddr = attribAddr - 0x0008;
						name_table = ppu_map[name_table_no ^ 1];
						tileX -= 0x0020;
					}
					attribAddr++;
				}
				attribBits = (((name_table[attribAddr] & 0xFF) << 2) >> (pos_y | (tileX & 2))) & 0xC;
			}
			
			for( int j = 0 ; j < 33 ; j++ ){
				if( BANKSWITCH_PER_TILE ){
					if( j != 32 ){
						cpu_emulate_CPU_cycles(2);
					}
				}
				if( exgfx_mapper ){
					int mmc5_pal = mapper.PPU_Latch_RenderScreen(1, nameAddr & 0x03FF);
					if( mmc5_pal != 0) {
						attribBits = mmc5_pal;
					}
				}
				final	int	patternAddr = bg_pattern_table_addr + ((name_table[nameAddr] & 0xFF) << 4) + subtile_y;
				final	int	patno = (patternAddr >> 4) & 0x1FF;
						int	patno_hi = patno >> 6;
						int	patno_lo = ((patno & 0x3F) << 3);
				if( patdirty[patno_hi][patno & 0x3F] == 0 ){
					int patno_lo2 = (patno_lo << 1);
					for( int v = 0 ; v < 8 ; v++ ){
						int lo = (ppu_map[patno_hi][patno_lo2 + v	 ] & 0xFF);
						int hi = (ppu_map[patno_hi][patno_lo2 + v + 8] & 0xFF);
						int pat;
						pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
							| ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
							| ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
							| ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
							| ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
							| ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
							| ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
							| ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
						patpix[patno_hi][patno_lo | v] = pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
					}
					patdirty[patno_hi][patno & 0x3F] = 1;
				}
				final int pat = patpix[patno_hi][patno_lo | subtile_y];
				if( latch_mapper ){
					final int addr = patternAddr;
					if( ((addr) & 0x0FC0) == 0x0FC0 ){
						if( (((addr) & 0x0FF0) == 0x0FD0) || (((addr) & 0x0FF0) == 0x0FE0) ){
							mapper.PPU_Latch_FDFE(addr);
						}
					}
				}
				attribBits |= 3;
				if (p >= 0){
					fb[p	] = cur_palette[(attribBits &  pat		 )];
					fb[p + 1] = cur_palette[(attribBits & (pat >>  4))];
					fb[p + 2] = cur_palette[(attribBits & (pat >>  8))];
					fb[p + 3] = cur_palette[(attribBits & (pat >> 12))];
					fb[p + 4] = cur_palette[(attribBits & (pat >> 16))];
					fb[p + 5] = cur_palette[(attribBits & (pat >> 20))];
					fb[p + 6] = cur_palette[(attribBits & (pat >> 24))];
					fb[p + 7] = cur_palette[(attribBits & (pat >> 28))];
				}
				p += 8;
				
				tileX++;
				nameAddr++;
				if( (tileX & 0x0001) == 0 ){
					if( (tileX & 0x0003) == 0 ){
						if( (tileX & 0x001F) == 0 ){
							nameAddr = nameAddr - 0x0020;
							attribAddr = attribAddr - 0x0008;
							name_table = ppu_map[name_table_no ^ 1];
							tileX -= 0x0020;
						}
						attribAddr++;
					}
					attribBits = (((name_table[attribAddr] & 0xFF) << 2) >> (pos_y | (tileX & 2))) & 0xC;
				}
			}
		}
	}
	
	// =========================
	// XvCg쐬
	//==========================
	private final void render_spr(){
		
		// XvCg`悠
		if( render_frame ){
			final	int[]		spr_ram = this.spr_ram;
			final	int			loopy_x = this.loopy_x;
			final	boolean 	latch_mapper = this.latch_mapper;
			//final	boolean		exgfx_mapper = this.exgfx_mapper;
			final	byte[][]	patdirty = this.patdirty;
			final	int[][]		patpix = this.patpix;
			final	int			sprHeight = ((PPU_R0 & 0x20) != 0) ? 16 : 8;
					int			PPU_R2 = this.PPU_R2;
					int			spritesInScanLine = 0;

			if( exgfx_mapper ){
				mapper.PPU_Latch_RenderScreen(0, 0);
			}
			for( int si = 63 ; si >= 0 ; si-- ){
				final int	s = (si << 2);
				final int	sprY = spr_ram[s] + 1;
				final int	lineOfSprite = current_frame_line - sprY;
				if( (lineOfSprite < 0) || (lineOfSprite >= sprHeight) ){
					continue;
				}
				spritesInScanLine++;
					  int		tileIndex = spr_ram[s + 1];
				final int		sprAttribs = spr_ram[s + 2];
				final boolean	sprVFlip = (sprAttribs & 0x80) != 0;
				final boolean	sprHFlip = (sprAttribs & 0x40) != 0;
				final boolean	sprBGPriority = (sprAttribs & 0x20) != 0;
					  int		y = lineOfSprite;
					  int		p = spr_ram[s + 3] + ((current_frame_line-8) * 256);
				if( sprVFlip ){
					y = (sprHeight - 1) - y;
				}
				if( latch_mapper ){
					final int addr = tileIndex << 4;
					if( ((addr) & 0x0FC0) == 0x0FC0 ){
						if( (((addr) & 0x0FF0) == 0x0FD0) || (((addr) & 0x0FF0) == 0x0FE0) ){
							mapper.PPU_Latch_FDFE(addr);
						}
					}
				}
				if( sprHeight == 16 ){
					if( (tileIndex & 0x01) != 0 ){
						tileIndex += 0x100;
						if( y < 8 ){
							tileIndex--;
						}
					}else{
						if( y >= 8 ){
							tileIndex++;
						}
					}
				}else{
					tileIndex += spr_pattern_table_addr >> 4;
				}
				int patno_hi = tileIndex >> 6;
				int patno_lo = ((tileIndex & 0x3F) << 3);
				if( patdirty[patno_hi][tileIndex & 0x3F] == 0 ){
					int patno_lo2 = (patno_lo << 1);
					for( int v = 0 ; v < 8 ; v++ ){
						int lo = (ppu_map[patno_hi][patno_lo2 + v	 ] & 0xFF);
						int hi = (ppu_map[patno_hi][patno_lo2 + v + 8] & 0xFF);
						int pat;
						pat = ((lo & 0x01) << 28) | ((hi & 0x01) << 29)
							| ((lo & 0x02) << 23) | ((hi & 0x02) << 24)
							| ((lo & 0x04) << 18) | ((hi & 0x04) << 19)
							| ((lo & 0x08) << 13) | ((hi & 0x08) << 14)
							| ((lo & 0x10) <<  8) | ((hi & 0x10) <<  9)
							| ((lo & 0x20) <<  3) | ((hi & 0x20) <<  4)
							| ((lo & 0x40) >>  2) | ((hi & 0x40) >>  1)
							| ((lo & 0x80) >>  7) | ((hi & 0x80) >>  6);
						patpix[patno_hi][patno_lo | v] = pat | (((pat + 0x77777777) ^ 0x44444444) & 0xCCCCCCCC);
					}
					patdirty[patno_hi][tileIndex & 0x3F] = 1;
				}
				final int pat = patpix[patno_hi][patno_lo | (y & 7)];
				final int colhi = ((sprAttribs & 3) << 2) | 0x10;
				if( si == 0 ){
					sprite_hit[lineOfSprite] = 0;
					if( pat != 0 ){
						for( int k = 0 ; k < 8 ; k++ ){
							if( fb[p + k] != 0 ){
								PPU_R2 |= 0x40;
								sprite_hit[lineOfSprite] = 1;
								break;
							}
						}
					}
				}
				if (p >= 0){
					int c;
					if( sprBGPriority ){
						if( sprHFlip ){
							if ((c = (pat >> 28) & 3) != 0 && fb[p	  ] == cur_palette[0]) fb[p	   ] = cur_palette[colhi | c];
							if ((c = (pat >> 24) & 3) != 0 && fb[p + 1] == cur_palette[0]) fb[p + 1] = cur_palette[colhi | c];
							if ((c = (pat >> 20) & 3) != 0 && fb[p + 2] == cur_palette[0]) fb[p + 2] = cur_palette[colhi | c];
							if ((c = (pat >> 16) & 3) != 0 && fb[p + 3] == cur_palette[0]) fb[p + 3] = cur_palette[colhi | c];
							if ((c = (pat >> 12) & 3) != 0 && fb[p + 4] == cur_palette[0]) fb[p + 4] = cur_palette[colhi | c];
							if ((c = (pat >>  8) & 3) != 0 && fb[p + 5] == cur_palette[0]) fb[p + 5] = cur_palette[colhi | c];
							if ((c = (pat >>  4) & 3) != 0 && fb[p + 6] == cur_palette[0]) fb[p + 6] = cur_palette[colhi | c];
							if ((c =  pat		 & 3) != 0 && fb[p + 7] == cur_palette[0]) fb[p + 7] = cur_palette[colhi | c];
						}else{
							if ((c =  pat		 & 3) != 0 && fb[p	  ] == cur_palette[0]) fb[p	   ] = cur_palette[colhi | c];
							if ((c = (pat >>  4) & 3) != 0 && fb[p + 1] == cur_palette[0]) fb[p + 1] = cur_palette[colhi | c];
							if ((c = (pat >>  8) & 3) != 0 && fb[p + 2] == cur_palette[0]) fb[p + 2] = cur_palette[colhi | c];
							if ((c = (pat >> 12) & 3) != 0 && fb[p + 3] == cur_palette[0]) fb[p + 3] = cur_palette[colhi | c];
							if ((c = (pat >> 16) & 3) != 0 && fb[p + 4] == cur_palette[0]) fb[p + 4] = cur_palette[colhi | c];
							if ((c = (pat >> 20) & 3) != 0 && fb[p + 5] == cur_palette[0]) fb[p + 5] = cur_palette[colhi | c];
							if ((c = (pat >> 24) & 3) != 0 && fb[p + 6] == cur_palette[0]) fb[p + 6] = cur_palette[colhi | c];
							if ((c = (pat >> 28) & 3) != 0 && fb[p + 7] == cur_palette[0]) fb[p + 7] = cur_palette[colhi | c];
						}
					}else{
						if( sprHFlip ){
							if ((c = (pat >> 28) & 3) != 0) fb[p	] = cur_palette[colhi | c];
							if ((c = (pat >> 24) & 3) != 0) fb[p + 1] = cur_palette[colhi | c];
							if ((c = (pat >> 20) & 3) != 0) fb[p + 2] = cur_palette[colhi | c];
							if ((c = (pat >> 16) & 3) != 0) fb[p + 3] = cur_palette[colhi | c];
							if ((c = (pat >> 12) & 3) != 0) fb[p + 4] = cur_palette[colhi | c];
							if ((c = (pat >>  8) & 3) != 0) fb[p + 5] = cur_palette[colhi | c];
							if ((c = (pat >>  4) & 3) != 0) fb[p + 6] = cur_palette[colhi | c];
							if ((c =  pat		 & 3) != 0) fb[p + 7] = cur_palette[colhi | c];
						}else{
							if ((c =  pat		 & 3) != 0) fb[p	] = cur_palette[colhi | c];
							if ((c = (pat >>  4) & 3) != 0) fb[p + 1] = cur_palette[colhi | c];
							if ((c = (pat >>  8) & 3) != 0) fb[p + 2] = cur_palette[colhi | c];
							if ((c = (pat >> 12) & 3) != 0) fb[p + 3] = cur_palette[colhi | c];
							if ((c = (pat >> 16) & 3) != 0) fb[p + 4] = cur_palette[colhi | c];
							if ((c = (pat >> 20) & 3) != 0) fb[p + 5] = cur_palette[colhi | c];
							if ((c = (pat >> 24) & 3) != 0) fb[p + 6] = cur_palette[colhi | c];
							if ((c = (pat >> 28) & 3) != 0) fb[p + 7] = cur_palette[colhi | c];
						}
					}
				}
			}
			if( spritesInScanLine >= 8 ){
				PPU_R2 |= 0x20;
			}else{
				PPU_R2 &= 0xDF;
			}
			this.PPU_R2 = PPU_R2;
			
		// XvCg`Ȃ
		}else{
			final int sprHeight = ((PPU_R0 & 0x20) != 0) ? 16 : 8;
			final int sprY = spr_ram[0] + 1;
			final int lineOfSprite = current_frame_line - sprY;
			if( (lineOfSprite >= 0) && (lineOfSprite < sprHeight) ){
				if( sprite_hit[lineOfSprite] != 0 ){
					PPU_R2 |= 0x40;
				}
			}
		}
	}
	
	

/*==================================================================================================================*/
/*																						@ROMpb`				*/
/*------------------------------------------------------------------------------------------------------------------*/
	
	// =========================
	// ROMpb`
	//==========================
	private final void rom_patch( int icrc, int imapper_no){
		
		if( imapper_no == 5 ){			// }bp[5(hƂ)
			SRAM = new byte[0x10000];
			SRAM_SIZE = 0x10000;
		}
		
		switch(icrc){
		
		// Mapper 001
		case	0x7831b2ff:		// AJ哝̑I
		case	0x190a3e11:		// r[EobvEnCXN[ ZɊy`
		case	0x52449508:		// z[iC^[ yig[O
		case	0x0973f714:		// 
		case	0x7172f3d4:		// 
		case	0xa5781280:		// E2
		case	0x8ce9c87b:		// UE}l[Q[
		case	0xec47296d:		// Xc
		case	0xcee5857b:		// E҂zC!
		case	0xe63d9193:		// J_ȉw3
		case	0xd54f5da9:		// ςEH[Y
			has_SRAM = true;
			break;
		case	0xb8747abf:		// xXgv[v싅XyV
			has_SRAM = true;
			SRAM = new byte[0x4000];
			SRAM_SIZE = 0x4000;
			break;
		
		// Mapper 004
		case	0xf3feb3ab:		// _E^EXyV ɂ̎㌀SW
		case	0xa524ae9b:		// ̐
		case	0x46dc6e57:		// SDK_ K`|m2 JvZL
		case	0x66b2dec7:		// SDK_ K`|m3 pYL
		case	0x92b07fd9:		// SDK_ K`|m4 j[^CvXg[[
		case	0x8ee6463a:		// SDK_ K`|m5 og Iu jo[T Z`[
		case	0xaf754426:		// Eg}y3
		case	0xfe4e5b11:		// ƂƂ [̑d
		case	0x57c12c17:		// RуTXyX seNEl
			has_SRAM = true;
			break;
		case	0x42e03e4a:		// RPGlQ[
			has_SRAM = true;
			mapper_no = 118;
			break;
		case	0x5c707ac4:		// MOTHER
			BANKSWITCH_PER_TILE = true;
			break;
		case	0xa52a8119:		// Eg}y2
			CYCLE_PER_LINE = 115;
			break;
		case	0xa0b8d310:		// KibN
			CYCLE_PER_LINE = 113;
			break;
			
		// Mapper 005
		case	0xf4cd4998:		// qC
		case	0x8fa95456:		// ېV̗
		case	0x98c8e090:		// M̖] 퍑QY`
		case	0xe6c28c5f:		// ` V̐
			SRAM = new byte[0x8000];
			SRAM_SIZE = 0x8000;
			break;
		case	0xf4120e58:		// TƔĎ j
		case	0x286613d8:		// M̖] _^
		case	0x95ba5733:		// Ou2
			SRAM = new byte[0x10000];
			SRAM_SIZE = 0x10000;
			break;
		case	0x5d9d9891:		// WXgu[h
		case	0xe91548d8:		// V4lł}[W 𖞓V
			has_SRAM = true;
			break;
		
		// Mapper 010
		case 0xc9cce8f2:		// Famicom Wars (J)
			has_SRAM = true;
			break;
		
		// Mapper 016
		case	0x983d8175:		// f[^bN ogbV
		case	0x894efdbc:		// f[^bN N񂿂 Iƃ|C|C
		case	0x19e81461:		// f[^bN hS{[Z Vꕐ
		case	0xbe06853f:		// f[^bN J[OX[p[gbvvC[Y
		case	0x0be0a328:		// f[^bN SDK_ GUNDAM WAR
		case	0x5b457641:		// f[^bN Eg}y X|t@Cg!
		case	0xf51a7f46:		// f[^bN HV Íp
		case	0x31cd9903:		// hS{[Z P!TCl
		case	0xe49fc53e:		// hS{[Z2 _t[U!!
		case	0x09499f4d:		// hS{[Z3 ll
		case	0x2e991109:		// hS{[ZO` TClŌv
			has_SRAM = true;
			break;
		case	0x6c6c2feb:		// hS{[3 `
			CYCLE_PER_LINE = 120;
			break;
		
		// Mapper 019
		case	0x3296ff7a:		// ogt[g.nes
		case	0x429fd177:		// t@~X^'90
		case	0xdd454208:		// nChCh3 ł̖K
		case	0xb1b9e187:		// Lb
		case	0xaf15338f:		// }ChV[J[
			has_SRAM = true;
			break;
		case	0x761ccfb5:		// _]2
			CYCLE_PER_LINE = 119;
			break;
			
		// Mapper 026
		case	0x836cc1ab:		// 鲐L}_
			has_SRAM = true;
			break;
		
		// Mapper 033
		case	0x547e6cc1:		// tgXg[
			mapper_no = 48;
			break;
		
		// Mapper 066
		case	0xb00abe1c:		// h
			CYCLE_PER_LINE = 113;
			break;
		
		// Mapper 068
		case	0xfde79681:		// }nW
			has_SRAM = true;
			break;
		
		// Mapper 069
		case	0xfeac6916:		// ̓hbWe2
			has_SRAM = true;
			break;
		
		// Mapper 080
		case	0x95aaed34:		// _bW[@X
		case	0x17627d4b:		// ^Cg[Ov hւ̃CZX
			has_SRAM = true;
			break;
		
		// Mapper 082
		case	0x4819a595:		// ɃnLX^WA N
			has_SRAM = true;
			break;
		
		// Mapper 118
		case	0x3b0fb600:		// C[X3 _[Y t C[X
			has_SRAM = true;
			break;
			
		default:
			break;
		}
	}
}
