import java.io.*;

import javax.microedition.io.Connector;

import com.nttdocomo.io.*;
import com.nttdocomo.fs.*;
import com.nttdocomo.device.*;

public final class ROMCache{

    private NESCanvas gui;

    private static final int VERSION = 1;

    /** ROM LbV **/
    private static final int START_SP_ROMCACHE = 0x10000; // 0x3000
    private static final int END_SP_ROMCACHE = 0x100000; // 0xA0000g(ASPsizeݒl0xA0000`0xC0000xŏ\)
    private static final int SIZE_SP_HEADER = 32;
    private static final int SIZE_SP_BLOCK_HEADER = 8;
    private static final int SIZE_SP_SAVEx12_HEADER = 4 * 12;
    private static final int SIZE_SP_SAVEBLOCK_HEADER = 12;
    // ȈՃZ[u(0)
    private static final int START_SP_SAVE_0 = 1300 + (30 * 12) * 20; // SIZE_SP_SETTING + (SIZE_ROM_NAME * NUM_SDROM) * MAX_SDROM_PAGE
    private static final int SIZE_SP_SAVE_0 = START_SP_ROMCACHE - START_SP_SAVE_0;
    // 
    private int SIZE_SP_BIOS;
    private int SIZE_SP_SRAM;
    private int SIZE_SP_ROM;
    private int SIZE_SP_SAVEx12;
    private int[] SIZE_SP_SAVE = new int[13];
    private int SIZE_SP;

    private int SIZE_BIOS_RAW = 0;
    private int SIZE_BIOS_COMPRESS = 0;
    private int SIZE_SRAM_RAW = 0;
    private int SIZE_SRAM_COMPRESS = 0;
    private int SIZE_ROM_RAW = 0;
    private int SIZE_ROM_COMPRESS = 0;
    private int[] SIZE_SAVE_RAW = new int[13];
    private int[] SIZE_SAVE_COMPRESS = new int[13];
    private int[] SAVE_ROMID = new int[13];

    /*
     * ɃubN\}A\
     *
     */

    private byte[] BIOS = null;
    private byte[] SRAM = null;
    private byte[] ROM = null;
    private byte[][] SAVE = new byte[13][];
    private ByteArrayOutputStream[] saveStream = new ByteArrayOutputStream[13];

    private boolean noRomCache_ = false;
    // k[h -1:kȂA0:Xe[gZ[û݈kA1:Sf[^k
    private int compressMode_ = 0;
    private boolean compressDisable_ = false;

    // ԕ\
    private boolean showMemory = false;

    public boolean onSaving = false;
    public boolean onMoving = false;

    public ROMCache(NESCanvas _gui){
        gui = _gui;

        for (int i = 0; i <= 12; i++) {
            SIZE_SP_SAVE[i] = -1;
            SIZE_SAVE_RAW[i] = -1;
            SIZE_SAVE_COMPRESS[i] = -1;
            SAVE_ROMID[i] = -1;
            SAVE[i] = null;
            saveStream[i] = null;
        }
    }

    public final int init(){
        boolean ok = true;
        DataInputStream in = null;
        int version, size_bios, size_sram, size_rom, size_save, size_sp;
        version = size_bios = size_sram = size_rom = size_save = size_sp = 0;
        byte c1, c2, c3, c4, c5, c6, c7, c8;
        c1 = c2 = c3 = c4 = c5 = c6 = c7 = c8 = 0;

        ok = initSizeSP(false);
        if (!ok) {
            noRomCache_ = true;
            return -1;
        }

        try {
            in = openSPInputStream(START_SP_ROMCACHE, SIZE_SP_HEADER);
            c1 = in.readByte();
            c2 = in.readByte();
            c3 = in.readByte();
            c4 = in.readByte();
            c5 = in.readByte();
            c6 = in.readByte();
            c7 = in.readByte();
            c8 = in.readByte();
            version = in.readInt();
            size_bios = in.readInt();
            size_sram = in.readInt();
            size_rom = in.readInt();
            size_save = in.readInt();
            size_sp = in.readInt();
        } catch (Exception e) {
            gui.alert("err1:" + e);
            ok = false;
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        if (!ok) {
            noRomCache_ = true;
            gui.alert("ROMLbV͎g܂");
            return -1;
        }

        if (c1 == 'R' && c2 == 'O' && c3 == 'M' && c4 == 'C'
            && c5 == 'A' && c6 == 'C' && c7 == 'H' && c8 == 'E') {
            SIZE_SP_BIOS = size_bios;
            SIZE_SP_SRAM = size_sram;
            SIZE_SP_ROM = size_rom;
            SIZE_SP_SAVEx12 = size_save;
            SIZE_SP = size_sp;
            ok = readBlockHeader();
            if (ok) ok = readSaveHeader();
            if (ok) ok = readSaveBlockHeader();
            if (ok) ok = readSave0Header();
        } else {
            gui.infomsg("ROMLbVGA܂");

            ok = initSizeSP(true);
            if (!ok) {
                noRomCache_ = true;
                gui.alert("ROMLbV͎g܂");
                return -1;
            }

            DataOutputStream out = null;
            SIZE_SP_BIOS = SIZE_SP_BLOCK_HEADER;
            SIZE_SP_SRAM = SIZE_SP_BLOCK_HEADER;
            SIZE_SP_ROM = SIZE_SP_BLOCK_HEADER;
            SIZE_SP_SAVEx12 = SIZE_SP_SAVEx12_HEADER + SIZE_SP_SAVEBLOCK_HEADER * 12;
            SIZE_BIOS_RAW = 0;
            SIZE_BIOS_COMPRESS = 0;
            SIZE_SRAM_RAW = 0;
            SIZE_SRAM_COMPRESS = 0;
            SIZE_ROM_RAW = 0;
            SIZE_ROM_COMPRESS = 0;
            try {
                out = openSPOutputStream(START_SP_ROMCACHE, SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM + SIZE_SP_ROM);
                out.writeByte('R');
                out.writeByte('O');
                out.writeByte('M');
                out.writeByte('C');
                out.writeByte('A');
                out.writeByte('C');
                out.writeByte('H');
                out.writeByte('E');
                out.writeInt(VERSION);
                out.writeInt(SIZE_SP_BIOS);
                out.writeInt(SIZE_SP_SRAM);
                out.writeInt(SIZE_SP_ROM);
                out.writeInt(SIZE_SP_SAVEx12);
                out.writeInt(SIZE_SP);
                out.writeInt(SIZE_BIOS_RAW);
                out.writeInt(SIZE_BIOS_COMPRESS);
                out.writeInt(SIZE_SRAM_RAW);
                out.writeInt(SIZE_SRAM_COMPRESS);
                out.writeInt(SIZE_ROM_RAW);
                out.writeInt(SIZE_ROM_COMPRESS);
                out.flush();
            } catch (Exception e) {
                gui.alert("err2:" + e);
                ok = false;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
            // ďI[ɃZ[uf[^̃wb_\
            if (ok) ok = initSaveBlockHeader(0);
            if (ok) ok = initSave0Header();
            if (ok) {
                gui.dialogNotReady = true;
                boolean rYes = true;
                rYes = gui.confirm("SD̾ްROMر\n߰܂H");
                if (rYes) {
                    gui.infomsg("΂炭҂");
                    sd2cache(0);
                }
            }
        }
        if (!ok) {
            noRomCache_ = true;
            gui.alert("ROMLbV͎g܂");
            return -1;
        }

        memoryStatus("");

        return 0;
    }

    private final DataInputStream openSPInputStream(final int pos, final int length) throws IOException{
        DataInputStream in = null;
        in = gui.openSPInputStream(pos, length);
        //in = Connector.openDataInputStream(
        //    "scratchpad:///0;pos=" + pos + ",length=" + length);
        return in;
    }

    private final DataOutputStream openSPOutputStream(final int pos, final int length) throws IOException{
        DataOutputStream out = null;
        out = gui.openSPOutputStream(pos, length);
        //out = Connector.openDataOutputStream(
        //    "scratchpad:///0;pos=" + pos + ",length=" + length);
        return out;
    }

    private final DataInputStream openCompressInputStream(DataInputStream in) {
        ByteArrayOutputStream out = null;

        //WJGW̏
        SevenZip.Compression.LZMA.Decoder decoder = new SevenZip.Compression.LZMA.Decoder();

        memoryStatus("LO");

        try {
            gui.setMessage("kf[^LEEE");

            // WJ
            long outSize = 0;
            int propertiesSize = 5;
            byte[] properties = new byte[propertiesSize];
            if (in.read(properties, 0, propertiesSize) != propertiesSize)
                throw new Exception("input .lzma file is too short");
            if (!decoder.SetDecoderProperties(properties))
                throw new Exception("Incorrect stream properties");
            int v = 0;
            for (int i = 0; i < 8; i++) {
                v = in.read();
                if (v < 0) {
                    throw new Exception("Can't read stream size");
                }
                outSize |= ((long)v) << (8 * i);
            }

            // in(SP)  out ɓWJ
            out = new ByteArrayOutputStream();
            if (!decoder.Code(in, out, outSize)) {
                throw new Exception("Error in data stream");
            }
            gui.setMessage("");
        } catch (Exception e) {
            gui.alert("err(openCompressInputStream1) : " + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        decoder = null;

        memoryStatus("L");

        ByteArrayInputStream bin = null;
        try {
            // out  buf ɃRs[Abuf  is 쐬
            byte[] buf = out.toByteArray();
            bin = new ByteArrayInputStream(buf);
            in = new DataInputStream(bin);
            buf = null;
        } catch (Exception e) {
            gui.alert("err(openCompressInputStream2) :" + e);
        } finally {
            try {
                bin.close();
                bin = null;
            } catch (Exception e) {
                bin = null;
            }
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        System.gc();
        return in;
    }

    public final byte[] compressedData(final byte[] data) {
        ByteArrayInputStream in = null;
        ByteArrayOutputStream out = null;

        memoryStatus("kO");

        //kGW̏
        SevenZip.Compression.LZMA.Encoder encoder = new SevenZip.Compression.LZMA.Encoder();
        // fomãƎTCY16bitxEB
        // TCY8bitƃgpʂ16bit̔AROM̈k͂Ȃɕς炸B
        encoder.SetDictionarySize(1 << 8);
        encoder.SeNumFastBytes(128);
        encoder.SetMatchFinder(1);
        encoder.SetLcLpPb(0, 2, 2);
        encoder.SetEndMarkerMode(true);

        try {
            in = new ByteArrayInputStream(data);
            out = new ByteArrayOutputStream();

            gui.setMessage("ki΂炭҂j");
            encoder.WriteCoderProperties(out);
            long fileSize = -1;
            for (int i = 0; i < 8; i++)
                out.write((int)(fileSize >>> (8 * i)) & 0xFF);

            memoryStatus("k");

            encoder.Code(in, out, -1, -1);
            out.flush();

            gui.setMessage("");
        } catch (Exception e) {
            gui.alert("err in Encode: " + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        byte[] buf = out.toByteArray();
        try {
            out.close();
            out = null;
        } catch (Exception e) {
            out = null;
        }

        memoryStatus("k");

        encoder = null;
        System.gc();
        return buf;
    }


    public final int writeSPBIOS(final byte[] bios) {
        int ret = -1;
        DataInputStream in = null;
        DataOutputStream out = null;
        boolean rewrite = false;

        //if (!compressDisable_ && compressMode_ > 0) {
        if (compressMode_ > 0) {
            // łɈkԂŕۑĂ珑݂XLbv
            if (SIZE_BIOS_COMPRESS > 0 && SIZE_BIOS_RAW > SIZE_BIOS_COMPRESS)
                return 0;

            BIOS = compressedData(bios);
        } else {
            BIOS = bios;
        }
        SIZE_BIOS_RAW = bios.length;
        SIZE_BIOS_COMPRESS = BIOS.length;

        if((SIZE_SP_BLOCK_HEADER + SIZE_BIOS_COMPRESS) > SIZE_SP_BIOS) {
            try {
                //SRAM  ROM ǂ݂B
                SRAM = new byte[SIZE_SP_SRAM];
                ROM = new byte[SIZE_SP_ROM];
                in = openSPInputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS, SIZE_SP_SRAM + SIZE_SP_ROM);
                in.read(SRAM);
                in.read(ROM);
                ret = 0;
            } catch (Exception e) {
                gui.alert("writeSP(BIOS): SPɕۑł܂B\n" + e);
                BIOS = null;
                SRAM = null;
                ROM = null;
            } finally {
                try {
                    in.close();
                    in = null;
                } catch (Exception e) {
                    in = null;
                }
            }
            rewrite = true;
            SIZE_SP_BIOS = SIZE_SP_BLOCK_HEADER + SIZE_BIOS_COMPRESS;
        } else {
            ret = 0;
        }

        // 󂫗eʃ`FbN
        if ((START_SP_ROMCACHE
            + SIZE_SP_HEADER
            + SIZE_SP_BIOS
            + SIZE_SP_SRAM
            + SIZE_SP_ROM
            + SIZE_SP_SAVEx12) > SIZE_SP) {
            // eʕs
            noRomCache_ = true;
            ret = -1;
        }

        if (ret == 0) {
            try {
                int length = rewrite ? (SIZE_SP_BIOS + SIZE_SP_SRAM + SIZE_SP_ROM) : SIZE_SP_BIOS;
                out = openSPOutputStream(START_SP_ROMCACHE + SIZE_SP_HEADER, length);
                out.writeInt(SIZE_BIOS_RAW);
                out.writeInt(SIZE_BIOS_COMPRESS);
                out.write(BIOS);
                if (rewrite) {
                    //SRAM  ROM ď
                    out.write(SRAM);
                    out.write(ROM);
                    out.flush();
                    out.close();
                    //SIZE_SP_BIOS XV
                    out = openSPOutputStream(START_SP_ROMCACHE + 12, 4);
                    out.writeInt(SIZE_SP_BIOS);
                }
                out.flush();
            } catch (Exception e) {
                gui.alert("writeSP(BIOS): ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        BIOS = null;
        SRAM = null;
        ROM = null;
        System.gc();
        return ret;
    }

    public final int writeSPSRAM(final byte[] sram) {
        int ret = -1;
        DataOutputStream out = null;
        DataInputStream in = null;
        boolean rewrite = false;
        //if (!compressDisable_ && compressMode_ > 0) {
        if (compressMode_ > 0) {
            SRAM = compressedData(sram);
        } else {
            SRAM = sram;
        }
        SIZE_SRAM_RAW = sram.length;
        SIZE_SRAM_COMPRESS = SRAM.length;

        if((SIZE_SP_BLOCK_HEADER + SIZE_SRAM_COMPRESS) > SIZE_SP_SRAM) {
            try {
                //ROM ǂ݂
                ROM = new byte[SIZE_SP_ROM];
                in = openSPInputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM, SIZE_SP_ROM);
                in.read(ROM);
                ret = 0;
            } catch (Exception e) {
                gui.alert("writeSP(SRAM): SPɕۑł܂B\n" + e);
                SRAM = null;
                ROM = null;
            } finally {
                try {
                    in.close();
                    in = null;
                } catch (Exception e) {
                    in = null;
                }
            }
            rewrite = true;
            SIZE_SP_SRAM = SIZE_SP_BLOCK_HEADER + SIZE_SRAM_COMPRESS;
        } else {
            ret = 0;
        }

        // 󂫗eʃ`FbN
        if ((START_SP_ROMCACHE
            + SIZE_SP_HEADER
            + SIZE_SP_BIOS
            + SIZE_SP_SRAM
            + SIZE_SP_ROM
            + SIZE_SP_SAVEx12) > SIZE_SP) {
            // eʕs
            noRomCache_ = true;
            ret = -1;
        }

        if (ret == 0) {
            try {
                int length = rewrite ? (SIZE_SP_SRAM + SIZE_SP_ROM) : SIZE_SP_SRAM;
                out = openSPOutputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS, length);
                out.writeInt(SIZE_SRAM_RAW);
                out.writeInt(SIZE_SRAM_COMPRESS);
                out.write(SRAM);
                if (rewrite) {
                    //ROM ď
                    out.write(ROM);
                    out.flush();
                    out.close();
                    //SIZE_SP_SRAM XV
                    out = openSPOutputStream(START_SP_ROMCACHE + 12 + 4, 4);
                    out.writeInt(SIZE_SP_SRAM);
                }
                out.flush();
            } catch (Exception e) {
                gui.alert("writeSP(SRAM): ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        SRAM = null;
        ROM = null;
        System.gc();
        return ret;
    }

    public final int writeSPROM(int rom_size, int vrom_size, int b6, int b7, byte[][] rom, byte[][] vrom) {
        int ret = -1;
        DataOutputStream out = null;
        // VROMTCYNemulatorLite̔{Ɋg(gƂĂ0xA10000xC1000)
        if(rom_size > 0x20 || vrom_size > 0x20) {
            //gui.infomsg("ROMTCY" + rom_size + " VROMTCY" + vrom_size);
            return ret;
        }

        boolean largeRom = false;
        // 傫ڂROM(500KBNX)͋IɈkĂ܂
        if (rom_size == 0x20 || vrom_size == 0x20) largeRom = true;
        ByteArrayOutputStream bin = new ByteArrayOutputStream();
        try {
            out = new DataOutputStream(bin);
            out.writeByte('N');
            out.writeByte('E');
            out.writeByte('S');
            out.writeByte(0x1A);
            out.writeByte(rom_size);
            out.writeByte(vrom_size);
            out.writeByte(b6);
            out.writeByte(b7);
            for(int i = 8; i < 0x10; i++) out.writeByte(0);
            for(int i = 0; i < (rom_size << 1); i++) //MAXSIZE:80000
                out.write(rom[i], 0, 0x2000);
            for(int i = 0; i < (vrom_size << 3); i++)//MAXSIZE:40000
                out.write(vrom[i], 0, 0x400);
            out.flush();
            ret = 0;
        } catch (Exception e) {
            gui.alert("err writeSP(ROM): " + e);
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        //if (!compressDisable_ && (compressMode_ > 0 || largeRom)) {
        if (compressMode_ > 0 || (largeRom && !compressDisable_)) {
            ROM = compressedData(bin.toByteArray());
        } else {
            ROM = bin.toByteArray();
        }
        SIZE_ROM_RAW = bin.toByteArray().length;
        SIZE_ROM_COMPRESS = ROM.length;
        SIZE_SP_ROM = SIZE_SP_BLOCK_HEADER + SIZE_ROM_COMPRESS;

        // 󂫗eʃ`FbN
        if ((START_SP_ROMCACHE
            + SIZE_SP_HEADER
            + SIZE_SP_BIOS
            + SIZE_SP_SRAM
            + SIZE_SP_ROM
            + SIZE_SP_SAVEx12) > SIZE_SP) {
            // eʕs
            noRomCache_ = true;
            ret = -1;
        }

        if (ret == 0) {
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM, SIZE_SP_ROM);
                out.writeInt(SIZE_ROM_RAW);
                out.writeInt(SIZE_ROM_COMPRESS);
                out.write(ROM);
                out.flush();
                out.close();
                //SIZE_SP_ROM XV
                out = openSPOutputStream(START_SP_ROMCACHE + 12 + 8, 4);
                out.writeInt(SIZE_SP_ROM);
                out.flush();
            } catch (Exception e) {
                gui.alert("writeSP(ROM): ݎs\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        ROM = null;
        System.gc();
        return ret;
    }

    public final int writeSPFD(int disk_banks, byte[][] disk) {
        int ret = -1;
        DataOutputStream out = null;
        ByteArrayOutputStream bin = new ByteArrayOutputStream();
        try {
            out = new DataOutputStream(bin);
            out.writeByte('F');
            out.writeByte('D');
            out.writeByte('S');
            out.writeByte(0x1A);
            out.writeByte(disk_banks);
            for (int i = 5; i < 0x10; i++) out.writeByte(0);
            for (int i = 0; i < disk_banks; i++) {
                out.write(disk[i], 0, 65500);
            }
            out.flush();
            ret = 0;
        } catch (Exception e) {
            gui.alert("err writeSP(FD): " + e);
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        //if (!compressDisable_ && compressMode_ > 0) {
        if (compressMode_ > 0) {
            ROM = compressedData(bin.toByteArray());
        } else {
            ROM = bin.toByteArray();
        }
        SIZE_ROM_RAW = bin.toByteArray().length;
        SIZE_ROM_COMPRESS = ROM.length;
        SIZE_SP_ROM = SIZE_SP_BLOCK_HEADER + SIZE_ROM_COMPRESS;

        // 󂫗eʃ`FbN
        if ((START_SP_ROMCACHE
            + SIZE_SP_HEADER
            + SIZE_SP_BIOS
            + SIZE_SP_SRAM
            + SIZE_SP_ROM
            + SIZE_SP_SAVEx12) > SIZE_SP) {
            // eʕs
            noRomCache_ = true;
            ret = -1;
        }

        if (ret == 0) {
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM, SIZE_SP_ROM);
                out.writeInt(SIZE_ROM_RAW);
                out.writeInt(SIZE_ROM_COMPRESS);
                out.write(ROM);
                out.flush();
                out.close();
                //SIZE_SP_ROM XV
                out = openSPOutputStream(START_SP_ROMCACHE + 12 + 8, 4);
                out.writeInt(SIZE_SP_ROM);
                out.flush();
            } catch (Exception e) {
                gui.alert("writeSP(FD): ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        ROM = null;
        System.gc();
        return ret;
    }

    public final DataInputStream openSPBIOSInputStream() {
        DataInputStream in = null;
        int pos = START_SP_ROMCACHE + SIZE_SP_HEADER;
        int length = SIZE_SP_BIOS;
        try {
            in = openSPInputStream(pos, length);
            SIZE_BIOS_RAW = in.readInt();
            SIZE_BIOS_COMPRESS = in.readInt();
        } catch (Exception e) {
        }
        if (SIZE_BIOS_RAW == SIZE_BIOS_COMPRESS && SIZE_BIOS_RAW > 0) {
            return in;
        } else if (SIZE_BIOS_COMPRESS > 0) {
            return openCompressInputStream(in);
        } else {
            //TCY0
            return in;
        }
    }

    public final DataInputStream openSPSRAMInputStream() {
        DataInputStream in = null;
        int pos = START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS;
        int length = SIZE_SP_SRAM;
        try {
            in = openSPInputStream(pos, length);
            SIZE_SRAM_RAW = in.readInt();
            SIZE_SRAM_COMPRESS = in.readInt();
        } catch (Exception e) {
        }
        if (SIZE_SRAM_RAW == SIZE_SRAM_COMPRESS && SIZE_SRAM_RAW > 0) {
            return in;
        } else if (SIZE_SRAM_COMPRESS > 0) {
            return openCompressInputStream(in);
        } else {
            //TCY0
            return in;
        }
    }

    public final DataInputStream openSPROMInputStream(){
        DataInputStream in = null;
        int pos = START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM;
        int length = SIZE_SP_ROM;
        try {
            in = openSPInputStream(pos, length);
            SIZE_ROM_RAW = in.readInt();
            SIZE_ROM_COMPRESS = in.readInt();
        } catch (Exception e) {
        }
        if (SIZE_ROM_RAW == SIZE_ROM_COMPRESS && SIZE_ROM_RAW > 0) {
            return in;
        } else if (SIZE_ROM_COMPRESS > 0) {
            return openCompressInputStream(in);
        } else {
            //TCY0
            return in;
        }
    }

    public final DataInputStream openSPSaveInputStream(final int slot){
        DataInputStream in = null;
        int pos, length;

        if (slot == 0) {
            pos = START_SP_SAVE_0;
            length = START_SP_ROMCACHE - START_SP_SAVE_0;
        } else {
            pos = SIZE_SP - SIZE_SP_SAVEx12;
            for (int i = 1; i < slot; i++) {
                pos += SIZE_SP_SAVE[i];
            }
            length = SIZE_SP_SAVE[slot];
        }

        try {
            in = openSPInputStream(pos, length);
            SIZE_SAVE_RAW[slot] = in.readInt();
            SIZE_SAVE_COMPRESS[slot] = in.readInt();
            SAVE_ROMID[slot] = in.readInt();
        } catch (Exception e) {
        }
        if (SIZE_SAVE_RAW[slot] == SIZE_SAVE_COMPRESS[slot] && SIZE_SAVE_RAW[slot] > 0) {
            return in;
        } else if (SIZE_SAVE_COMPRESS[slot] > 0) {
            return openCompressInputStream(in);
        } else {
            //TCY0
            return in;
        }
    }

    public final DataOutputStream openSPSaveOutputStream(final int slot, final int romid){
        if (slot >= 0 && slot <= 12) {
            SAVE_ROMID[slot] = romid;
            saveStream[slot] = new ByteArrayOutputStream();
            DataOutputStream out = new DataOutputStream(saveStream[slot]);
            onSaving = true;
            return out;
        } else {
            return null;
        }
    }

    public final boolean closeSPSaveOutputStream(final int s) {
        boolean ok = false;
        DataOutputStream out = null;
        DataInputStream in = null;
        boolean rewrite = false;
        boolean compress = !compressDisable_;

        if (s < 0 || s > 12) return false;

        byte[] save;
        if (saveStream[s] == null)
            save = null;
        else
            save = saveStream[s].toByteArray();
        try {
            saveStream[s].close();
            saveStream[s] = null;
        } catch (Exception e) {
            saveStream[s] = null;
        }

        SAVE[s] = save;
        SIZE_SAVE_RAW[s] = save.length;

        if (s == 0 && SIZE_SAVE_RAW[s] <= (SIZE_SP_SAVE_0 - SIZE_SP_SAVEBLOCK_HEADER)) {
                compress = false;
        }
        if (compress && compressMode_ >= 0) {
            SAVE[s] = compressedData(save);
        }
        SIZE_SAVE_COMPRESS[s] = SAVE[s].length;
        save = null;

        if (s == 0) {
            if (SIZE_SAVE_COMPRESS[0] > (SIZE_SP_SAVE_0 - SIZE_SP_SAVEBLOCK_HEADER)) {
                onSaving = false;
                return false;
            }
            ok = closeSPSaveOutputStream_0();
            onSaving = false;
            return ok;
        }

        if ((SIZE_SP_SAVEBLOCK_HEADER + SIZE_SAVE_COMPRESS[s]) > SIZE_SP_SAVE[s]) {
            if (s > 1) {
                try {
                    in = openSPInputStream(SIZE_SP - SIZE_SP_SAVEx12, SIZE_SP_SAVEx12);
                    for (int i = 1; i < s; i++) {
                        SAVE[i] = new byte[SIZE_SP_SAVE[i]];
                        in.read(SAVE[i]);
                    }
                    ok = true;
                } catch (Exception e) {
                    gui.alert("writeSP(SAVE): SPɕۑł܂B\n" + e);
                    for (int i = 1; i <=s; i++) {
                        SAVE[i] = null;
                    }
                } finally {
                    try {
                        in.close();
                        in = null;
                    } catch (Exception e) {
                        in = null;
                    }
                }
                rewrite = true;
            } else {
                ok = true;
            }
            SIZE_SP_SAVEx12 -= SIZE_SP_SAVE[s];
            SIZE_SP_SAVE[s] = SIZE_SP_SAVEBLOCK_HEADER + SIZE_SAVE_COMPRESS[s];
            SIZE_SP_SAVEx12 += SIZE_SP_SAVE[s];
        } else {
            ok = true;
        }

        // 󂫗eʃ`FbN
        if ((START_SP_ROMCACHE
            + SIZE_SP_HEADER
            + SIZE_SP_BIOS
            + SIZE_SP_SRAM
            + SIZE_SP_ROM
            + SIZE_SP_SAVEx12) > SIZE_SP) {
            // eʕs
            noRomCache_ = true;
            ok = false;
        }

        if (ok) {
            int pos = SIZE_SP - SIZE_SP_SAVEx12;
            int length = SIZE_SP_SAVE[s];
            for (int i = 1; i < s; i++) {
                if (rewrite) {
                    length += SIZE_SP_SAVE[i];
                } else {
                    pos += SIZE_SP_SAVE[i];
                }
            }
            try {
                out = openSPOutputStream(pos, length);
                if (rewrite) {
                    for (int i = 1; i < s; i++) {
                        out.write(SAVE[i]);
                    }
                }
                out.writeInt(SIZE_SAVE_RAW[s]);
                out.writeInt(SIZE_SAVE_COMPRESS[s]);
                out.writeInt(SAVE_ROMID[s]);
                out.write(SAVE[s]);
                out.flush();
                out.close();
                //SIZE_SP_SAVE[s] XV
                out = openSPOutputStream(SIZE_SP - SIZE_SP_SAVEx12_HEADER, SIZE_SP_SAVEx12_HEADER);
                for (int i = 1; i <= 12; i++) {
                    out.writeInt(SIZE_SP_SAVE[i]);
                }
                out.flush();
                out.close();
                //SIZE_SP_SAVEx12 XV
                out = openSPOutputStream(START_SP_ROMCACHE + 12 + 12, 4);
                out.writeInt(SIZE_SP_SAVEx12);
                out.flush();
            } catch (Exception e) {
                gui.alert("writeSP(SAVE): ݎsB\n" + e);
                ok = false;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        for (int i = 1; i <= s; i++) {
            SAVE[i] = null;
        }
        onSaving = false;
        System.gc();
        return ok;
    }

    // ܂Xe[gZ[u(ȈՃZ[u)
    public final boolean closeSPSaveOutputStream_0() {
        DataOutputStream out = null;
        boolean ok = true;
        try {
            out = openSPOutputStream(START_SP_SAVE_0, START_SP_ROMCACHE - START_SP_SAVE_0);
            out.writeInt(SIZE_SAVE_RAW[0]);
            out.writeInt(SIZE_SAVE_COMPRESS[0]);
            out.writeInt(SAVE_ROMID[0]);
            out.write(SAVE[0]);
        } catch (Exception e) {
            gui.alert("writeSP(SAVE(0)): ݎsB\n" + e);
            ok = false;
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        SAVE[0] = null;
        System.gc();
        return ok;
    }

    public final int compact() {
        int ret = -1;
        int i;
        DataInputStream in = null;
        DataOutputStream out = null;

        int OLD_SIZE_SP_BIOS = SIZE_SP_BIOS;
        int OLD_SIZE_SP_SRAM = SIZE_SP_SRAM;
        int OLD_SIZE_SP_ROM = SIZE_SP_ROM;
        int OLD_SIZE_SP_SAVEx12 = SIZE_SP_SAVEx12;
        int OLD_SIZE_SP_SAVE[] = new int[13];
        for (i = 1; i <= 12; i++) {
            OLD_SIZE_SP_SAVE[i] = SIZE_SP_SAVE[i];
        }

        int waste = SIZE_SP_BIOS + SIZE_SP_SRAM
            - (SIZE_SP_BLOCK_HEADER + SIZE_BIOS_COMPRESS + SIZE_SP_BLOCK_HEADER + SIZE_SRAM_COMPRESS);
        for (i = 1; i <= 12; i++) {
            waste += (SIZE_SP_SAVE[i] - (SIZE_SP_SAVEBLOCK_HEADER + SIZE_SAVE_COMPRESS[i]));
        }

        if (waste == 0) {
            gui.infomsg("œKKv͂܂B\n");
            return 0;
        }

        boolean rYes = true;
        rYes = gui.confirm("XNb`pbh̍œKs"
            + "\n" + waste + " oCg̋󂫗̈𑝂₹܂B"
            + "\nœKs܂H");
        if (!rYes)
            return 0;

        SIZE_SP_BIOS = SIZE_SP_BLOCK_HEADER + SIZE_BIOS_COMPRESS;
        SIZE_SP_SRAM = SIZE_SP_BLOCK_HEADER + SIZE_SRAM_COMPRESS;
        SIZE_SP_SAVEx12 = 0;
        for (i = 1; i <= 12; i++) {
            SIZE_SP_SAVE[i] = SIZE_SP_SAVEBLOCK_HEADER + SIZE_SAVE_COMPRESS[i];
            SIZE_SP_SAVEx12 += SIZE_SP_SAVE[i];
        }
        SIZE_SP_SAVEx12 += SIZE_SP_SAVEx12_HEADER;

        try {
            //SRAM  ROM ǂ݂B
            SRAM = new byte[SIZE_SP_SRAM];
            ROM = new byte[SIZE_SP_ROM];
            in = openSPInputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + OLD_SIZE_SP_BIOS, SIZE_SP_SRAM + SIZE_SP_ROM);
            in.read(SRAM);
            in.skipBytes(OLD_SIZE_SP_SRAM - SIZE_SP_SRAM);
            in.read(ROM);
            ret = 0;
        } catch (Exception e) {
            gui.alert("compact(): ǂݏosB\n" + e);
            SRAM = null;
            ROM = null;
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }

        if (ret == 0) {
            // wb_̏(sɔZ[uf[^͌Ŏ{)
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + 12, 8);
                out.writeInt(SIZE_SP_BIOS);
                out.writeInt(SIZE_SP_SRAM);
                out.flush();
            } catch (Exception e) {
                gui.alert("compact(): wb_ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }

        if (ret == 0) {
            // SRAM  ROM ̏݁iړj
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + SIZE_SP_HEADER + SIZE_SP_BIOS, SIZE_SP_SRAM + SIZE_SP_ROM);
                out.write(SRAM);
                out.write(ROM);
                out.flush();
            } catch (Exception e) {
                gui.alert("compact(): f[^ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }
        SRAM = null;
        ROM = null;
        System.gc();

        // s̒l߂Kv

        // ȍ~Z[uf[^֘A
        if (ret == 0) {
            // wb_(Z[uf[^)̏
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + 12 + 12, 4);
                out.writeInt(SIZE_SP_SAVEx12);
                out.flush();
            } catch (Exception e) {
                gui.alert("compact(): Z[uwb_ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }

        if (ret == 0) {
            // Z[uf[^̓ǂݍ
            try {
                in = openSPInputStream(SIZE_SP - OLD_SIZE_SP_SAVEx12, OLD_SIZE_SP_SAVEx12);
                for (i = 1; i <= 12; i++) {
                    SAVE[i] = new byte[SIZE_SP_SAVE[i]];
                    in.read(SAVE[i]);
                    in.skipBytes(OLD_SIZE_SP_SAVE[i] - SIZE_SP_SAVE[i]);
                }
                ret = 0;
            } catch (Exception e) {
                gui.alert("compact(): Z[uf[^ǂݏosB\n" + e);
                ret = -1;
            } finally {
                try {
                    in.close();
                    in = null;
                } catch (Exception e) {
                    in = null;
                }
            }
        }

        if (ret == 0) {
            // Z[uf[^̈ړۑ
            try {
                out = openSPOutputStream(SIZE_SP - SIZE_SP_SAVEx12, SIZE_SP_SAVEx12);
                for (i = 1; i <= 12; i++) {
                    out.write(SAVE[i]);
                }
                for (i = 1; i <= 12; i++) {
                    out.writeInt(SIZE_SP_SAVE[i]);
                }
            } catch (Exception e) {
                gui.alert("compact(): Z[uf[^ݎsB\n" + e);
                ret = -1;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }

        for (i = 1; i <= 12; i++) {
            SAVE[i] = null;
        }
        System.gc();

        if (ret == 0)
            gui.infomsg("œKłB"
                + "\n\nBIOSTCY : "
                + (OLD_SIZE_SP_BIOS - SIZE_SP_BLOCK_HEADER) + "  " + (SIZE_SP_BIOS - SIZE_SP_BLOCK_HEADER)
                +   "\nSRAMTCY : "
                + (OLD_SIZE_SP_SRAM - SIZE_SP_BLOCK_HEADER) + "  " + (SIZE_SP_SRAM - SIZE_SP_BLOCK_HEADER)
                +   "\n ROMTCY : "
                + (OLD_SIZE_SP_ROM - SIZE_SP_BLOCK_HEADER) + "  " + (SIZE_SP_ROM - SIZE_SP_BLOCK_HEADER)
                +   "\n  ް : "
                + (OLD_SIZE_SP_SAVEx12 - SIZE_SP_SAVEBLOCK_HEADER * 12 - SIZE_SP_SAVEx12_HEADER)
                    + "  " + (SIZE_SP_SAVEx12 - SIZE_SP_SAVEBLOCK_HEADER * 12 - SIZE_SP_SAVEx12_HEADER)
            );

        return ret;
    }

    public void status() {
        int bios = SIZE_SP_BIOS
                       - SIZE_SP_BLOCK_HEADER;
        int sram = SIZE_SP_SRAM
                       - SIZE_SP_BLOCK_HEADER;
        int rom  = SIZE_SP_ROM
                       - SIZE_SP_BLOCK_HEADER;
        int save = SIZE_SP_SAVEx12
                       - SIZE_SP_SAVEBLOCK_HEADER * 12
                       - SIZE_SP_SAVEx12_HEADER;
        int sum  = SIZE_SP_BIOS + SIZE_SP_SRAM + SIZE_SP_ROM + SIZE_SP_SAVEx12
                       - SIZE_SP_BLOCK_HEADER * 3
                       - SIZE_SP_SAVEBLOCK_HEADER * 12
                       - SIZE_SP_SAVEx12_HEADER;
        if (bios < 0) bios = 0;
        if (sram < 0) sram = 0;
        if (rom < 0) rom = 0;
        if (save < 0) save = 0;
        if (sum < 0) sum = 0;
        int free = (SIZE_SP - START_SP_ROMCACHE)
                       - (SIZE_SP_HEADER + SIZE_SP_BIOS + SIZE_SP_SRAM + SIZE_SP_ROM + SIZE_SP_SAVEx12);
        gui.infomsg("SP:ROMLbVXe[^X\n"
            + "\nBIOSTCY : " + bios
            + "\nSRAMTCY : " + sram
            + "\n ROMTCY : " + rom
            + "\nSAVETCY : " + save
            + "\n---------------------------"
            + "\nvTCY : " + sum
            + "\n  󂫗̈ : " + free
        );
    }

    public final boolean enableRomCache(boolean mode) {
        if (mode) {
            return noRomCache_;
        } else {
            return !mode;
        }
    }

    public final int setCompressMode(int mode) {
        compressMode_ = mode;
        return compressMode_;
    }

    public final boolean disableCompress(boolean mode) {
        compressDisable_ = mode;
        return compressDisable_;
    }

    public final int romid(final int slot) {
        if (slot < 0 || slot > 12) return 0;
        return SAVE_ROMID[slot];
    }

    public final int findSaveId(final int romid) {
        int slot = -1;
        for (int i = 1; i <= 12; i++) {
            if (SAVE_ROMID[i] == romid) {
                slot = i;
                // ŏɌްԂ
                // (Ōɂꍇ break RgAEg)
                break;
            }
        }
        return slot;
    }

    public final boolean hasBios() {
        return (SIZE_SP_BIOS > SIZE_SP_BLOCK_HEADER);
    }

    public final boolean hasSram() {
        return (SIZE_SP_SRAM > SIZE_SP_BLOCK_HEADER);
    }

    public final boolean hasRom() {
        return (SIZE_SP_ROM > SIZE_SP_BLOCK_HEADER);
    }

    public final void reset() {
        DataOutputStream out = null;
        try {
            out = openSPOutputStream(START_SP_ROMCACHE, 1);
            out.writeByte(0);
        } catch (Exception e) {
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
    }

    public final void deleteSaveData(final int slot) {
        if (slot < 1 || slot > 12) return;
        initSaveBlockHeader(slot);
    }

    public final boolean sd2cache(final int slot) {
        boolean ok = true;
        DataInputStream in = null;
        DataOutputStream out = null;
        SDbinding sd = new SDbinding();
        int start, end;
        if (slot == 0) {
            start = 1;
            end = 12;
        } else if (slot >= 1 && slot <= 12) {
            start = end = slot;
        } else {
            onMoving = false;
            return false;
        }

        for (int i = start; i <= end; i++) {
            saveStream[i] = new ByteArrayOutputStream();
            try {
                in = sd.openDataInputStream(gui.getSaveFile(i));
                out = new DataOutputStream(saveStream[i]);
                SAVE_ROMID[i] = -1;
                if (in != null) {
                    int b;
                    while ((b = in.read()) != -1) {
                        out.write(b);
                    }
                    out.flush();
                }
            }catch(Exception e){
                ok = false;
            }finally{
                try{
                    in.close();
                }catch(Exception e){}
                in = null;
                try {
                    out.close();
                } catch(Exception e) {}
                out = null;
                ok = closeSPSaveOutputStream(i);
            }
            if (!ok)
                break;
        }

        if(sd!=null) sd.close();
        sd = null;
        onMoving = false;
        return ok;
    }

    public final boolean cache2sd(final int slot) {
        boolean ok = true;
        DataInputStream in = null;
        DataOutputStream out = null;
        SDbinding sd = new SDbinding();
        int start, end;
        if (slot == 0) {
            start = 1;
            end = 12;
        } else if (slot >= 1 && slot <= 12) {
            start = end = slot;
        } else {
            onMoving = false;
            return false;
        }

        for (int i = start; i <= end; i++) {
            saveStream[i] = new ByteArrayOutputStream();
            try {
                in = openSPSaveInputStream(i);
                out = sd.openDataOutputStream(gui.getSaveFile(i));
                if (in != null) {
                    int b;
                    while ((b = in.read()) != -1) {
                        out.write(b);
                    }
                    out.flush();
                } else {
                    ok = false;
                }
            }catch(Exception e){
                ok = false;
            }finally{
                try{
                    in.close();
                }catch(Exception e){}
                in = null;
                try {
                    out.close();
                } catch(Exception e) {}
                out = null;
            }
            if (!ok)
                break;
        }

        if(sd!=null) sd.close();
        sd = null;

        if (ok) ok = initSaveBlockHeader(slot);

        onMoving = false;
        return ok;
    }


    /*
     * gp
     */
    private final void memoryStatus(String s) {
        if (showMemory) {
            com.nttdocomo.lang.MemoryManager M = com.nttdocomo.lang.MemoryManager.getMemoryManager();
            gui.infomsg(s +
                "\n[1]" +
                "\ng[^ = " + M.totalMemory()[0] +
                "\n  gp = " + (M.totalMemory()[0] - M.freeMemory()[0]) +
                "\n    󂫃 = " + M.freeMemory()[0] +
                "\nA󂫃 = " + M.maxContiguousMemory()[0]
            );
        }
    }

    private final boolean readBlockHeader() {
        boolean ret = false;
        DataInputStream in = null;
        try {
            in = openSPInputStream(START_SP_ROMCACHE + SIZE_SP_HEADER, SIZE_SP_BIOS + SIZE_SP_SRAM + SIZE_SP_ROM);
            SIZE_BIOS_RAW = in.readInt();
            SIZE_BIOS_COMPRESS = in.readInt();
            in.skipBytes(SIZE_SP_BIOS - SIZE_SP_BLOCK_HEADER);
            SIZE_SRAM_RAW = in.readInt();
            SIZE_SRAM_COMPRESS = in.readInt();
            in.skipBytes(SIZE_SP_SRAM - SIZE_SP_BLOCK_HEADER);
            SIZE_ROM_RAW = in.readInt();
            SIZE_ROM_COMPRESS = in.readInt();
            ret = true;
        } catch (Exception e) {
            gui.alert("err(readBlockHeader):" + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        return ret;
    }

    private final boolean readSaveHeader() {
        boolean ret = false;
        DataInputStream in = null;
        try {
            in = openSPInputStream(SIZE_SP - SIZE_SP_SAVEx12_HEADER, SIZE_SP_SAVEx12_HEADER);
            for (int i = 1; i <= 12; i++) {
                SIZE_SP_SAVE[i] = in.readInt();
            }
            ret = true;
        } catch (Exception e) {
            gui.alert("err(readSaveHeader):" + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        return ret;
    }

    private final boolean readSaveBlockHeader() {
        boolean ret = false;
        DataInputStream in = null;

        // XXX - ꂽ SIZE_SP_SAVEx12 pB񂩂sv
        SIZE_SP_SAVEx12 = 0;
        for (int i = 1; i <= 12; i++) {
            SIZE_SP_SAVEx12 += SIZE_SP_SAVE[i];
        }
        SIZE_SP_SAVEx12 += SIZE_SP_SAVEx12_HEADER;

        try {
            in = openSPInputStream(SIZE_SP - SIZE_SP_SAVEx12, SIZE_SP_SAVEx12 - SIZE_SP_SAVEx12_HEADER);
            for (int i = 1; i <= 12; i++) {
                SIZE_SAVE_RAW[i] = in.readInt();
                SIZE_SAVE_COMPRESS[i] = in.readInt();
                SAVE_ROMID[i] = in.readInt();
                in.skipBytes(SIZE_SP_SAVE[i] - SIZE_SP_SAVEBLOCK_HEADER);
            }
            ret = true;
        } catch (Exception e) {
            gui.alert("err(readSaveBlockHeader):" + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        return ret;
    }

    private final boolean readSave0Header() {
        boolean ret = false;
        DataInputStream in = null;
        try {
            in = openSPInputStream(START_SP_SAVE_0, SIZE_SP_SAVEBLOCK_HEADER);
            SIZE_SAVE_RAW[0] = in.readInt();
            SIZE_SAVE_COMPRESS[0] = in.readInt();
            SAVE_ROMID[0] = in.readInt();
            ret = true;
        } catch (Exception e) {
            gui.alert("err(readSave0Header):" + e);
        } finally {
            try {
                in.close();
                in = null;
            } catch (Exception e) {
                in = null;
            }
        }
        return ret;
    }

    private final boolean initSizeSP(boolean report) {
        DataInputStream in = null;
        boolean ok = false;
        int delta = 0x10000;
        SIZE_SP = END_SP_ROMCACHE;
        while (!ok) {
            ok = true;
            try {
                in = openSPInputStream(SIZE_SP - 1, 1);
            } catch (Exception e) {
                ok = false;
                SIZE_SP -= delta;
            } finally {
                try {
                    in.close();
                    in = null;
                } catch (Exception e) {
                    in = null;
                }
            }
            if ((START_SP_ROMCACHE * 2 - 1) > SIZE_SP) {
                SIZE_SP = START_SP_ROMCACHE * 2 - 1;
                ok = false;
                break;
            }
        }
        while (delta > 1) {
            delta >>= 1;
            if (ok) SIZE_SP += delta;
            else SIZE_SP -= delta;
            try {
                in = openSPInputStream(SIZE_SP - 1, 1);
                ok = true;
            } catch (Exception e) {
                ok = false;
            } finally {
                try {
                    in.close();
                    in = null;
                } catch (Exception e) {
                    in = null;
                }
            }
        }
        if (!ok) {
            SIZE_SP -= 1;
            ok = true;
        }

        if ((START_SP_ROMCACHE
             + SIZE_SP_HEADER
             + SIZE_SP_SAVEx12_HEADER
             + SIZE_SP_SAVEBLOCK_HEADER * 12) > SIZE_SP) {
            SIZE_SP = START_SP_ROMCACHE;
            ok = false;
        }

        if (report) {
            gui.infomsg("LSPTCY = " + SIZE_SP
                + "\n(ROMLbVgp = " + (SIZE_SP - START_SP_ROMCACHE) + ")");
        }
        return ok;
    }

    private final boolean initSaveBlockHeader(final int slot) {
        boolean ret = false;
        DataOutputStream out = null;
        int start, end;
        if (slot == 0) {
            start = 1;
            end = 12;
        } else if (slot >= 1 && slot <= 12) {
            // XXX - SɃ~XB
            //    Z[uf[^Ŝ̏ƂP̃Z[uf[^̍폜
            //    @łł킯ȂB
            //    ŏƍ폜͕邩̂EEE
            start = end = slot;
        } else {
            return ret;
        }
        try {
            int pos = SIZE_SP - SIZE_SP_SAVEx12;
            int length = SIZE_SP_SAVEx12;
            for (int i = 1; i < start; i++) {
                pos += SIZE_SP_SAVE[i];
                length -= SIZE_SP_SAVE[i];
            }
            out = openSPOutputStream(pos, length);
            for (int i = start; i <= end; i++) {
                SIZE_SAVE_RAW[i] = 0;
                SIZE_SAVE_COMPRESS[i] = 0;
                SAVE_ROMID[i] = -1;
                out.writeInt(SIZE_SAVE_RAW[i]);
                out.writeInt(SIZE_SAVE_COMPRESS[i]);
                out.writeInt(SAVE_ROMID[i]);
            }
            ret = true;
        } catch (Exception e) {
            gui.alert("err(initSaveBlockHeader):" + e);
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        if (start == end)
            return ret;

        try {
            out = openSPOutputStream(SIZE_SP - SIZE_SP_SAVEx12_HEADER, SIZE_SP_SAVEx12_HEADER);
            for (int i = start; i <= end; i++) {
                SIZE_SP_SAVE[i] = SIZE_SP_SAVEBLOCK_HEADER;
                out.writeInt(SIZE_SP_SAVE[i]);
            }
            ret = true;
        } catch (Exception e) {
            gui.alert("err(initSaveBlockHeader):" + e);
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        if (ret) {
            SIZE_SP_SAVEx12 = SIZE_SP_SAVEx12_HEADER + SIZE_SP_SAVEBLOCK_HEADER * 12;
            try {
                out = openSPOutputStream(START_SP_ROMCACHE + 12 + 12, 4);
                out.writeInt(SIZE_SP_SAVEx12);
                out.flush();
            } catch (Exception e) {
                ret = false;
            } finally {
                try {
                    out.close();
                    out = null;
                } catch (Exception e) {
                    out = null;
                }
            }
        }

        return ret;
    }

    private final boolean initSave0Header() {
        boolean ret = false;
        DataOutputStream out = null;
        try {
            out = openSPOutputStream(START_SP_SAVE_0, SIZE_SP_SAVEBLOCK_HEADER);
            SIZE_SAVE_RAW[0] = 0;
            SIZE_SAVE_COMPRESS[0] = 0;
            SAVE_ROMID[0] = -1;
            out.writeInt(SIZE_SAVE_RAW[0]);
            out.writeInt(SIZE_SAVE_COMPRESS[0]);
            out.writeInt(SAVE_ROMID[0]);
            ret = true;
        } catch (Exception e) {
            gui.alert("err(initSave0Header):" + e);
        } finally {
            try {
                out.close();
                out = null;
            } catch (Exception e) {
                out = null;
            }
        }
        return ret;
    }

}
